//
// This file contains functions related to database event nodes.
//

using namespace std;

#include <QDebug>
#include <QString>
#include <QStringList>
#include "library.h"
#include "fortran.h"
#include "eventnode.h"


// ***************************************************************
// This function returns a list of all wavefiles in an event file.
// ***************************************************************
void NodeGetWaveFiles(event_node_ *Node, QStringList *FileList)
{
   char line[81];

   // Clear the list.
   FileList->clear();

   // Create list of wavefiles.
   for (int i=0;i<Node->wavefiles.numlines;i++) {
      memcpy(line, Node->wavefiles.lines+i*80, 80);
      line[80]=0; FileList->append(line);
   }
}




// *********************************************************************
// This function copies all comment lines from a node to a QStringList
// structure. For each line in stringlist, the space in column one is
// removed, and so are all trailing spaces including the line number
// character (3) at column 79. Empty lines will be ignored.
// *********************************************************************
void NodeGetComments(event_node_ *Node, QStringList *CommentList)
{
   QString qsline;
   char line[81];

   // Copy all comment lines to 'CommentList'.
   for (int i=0;i<Node->comments.numlines;i++) {
      memcpy(line, Node->comments.lines+i*80, 80);
      line[80]=0; qsline = line; qsline[79] = 32;
      qsline.remove(0, 1); UnPadQString(&qsline);
      if (qsline.size()) CommentList->append(qsline);
   }
}




// *********************************************************************
// This function copies all comment lines from a QStringList structure
// to an event node. For each line, a space will be added to column one.
// Line will also be padded with spaces and the line number character
// (3) wil be added as well in column 79. The lines is assumed to no
// longer than 78 characters long.
// *********************************************************************
void NodeSetComments(event_node_ *Node, QStringList Comments)
{
   QString line;

   // Free all existing comment lines.
   if (Node->comments.numlines) {
      Node->comments.numlines = 0;
      delete [] Node->comments.lines;
      Node->comments.lines = (char*) 0;
   }

   // Add new comment lines if any.
   if (Comments.size()) {
      // Allocate memory for new comment lines.
      Node->comments.lines = new (nothrow) char [Comments.size()*80];
      if (!Node->comments.lines) return;
      // Add new comment lines to node.
      for (int i=0;i<Comments.size();i++) {
         line = Comments.at(i); line.insert(0, QChar(' '));
         PadQString(&line, 80); line[79] = QChar('3');
         memcpy(Node->comments.lines+i*80, line.toLocal8Bit().data(), 80);
      }
      Node->comments.numlines = Comments.size();
   }
}




// *********************************************************************
// This function is used to return a single property value
// for a node. The value returned is either a float or a string.
// *********************************************************************
void NodeGetProperty(event_node_ *node, QString property, prop_value *value)
{
   int count;
   phase_ *phase;
   QStringList StrList;

   // Initialize 'novalue'.
   value->novalue = false;

   // These properties return a numerical value.
   value->valueType = 0;
   if (property == "$LAT") {
      value->number = node->hypocenters.first->lat; return;
   }
   if (property == "$LON") {
      value->number = node->hypocenters.first->lon; return;
   }
   if (property == "$DEPTH") {
      value->number = node->hypocenters.first->depth; return;
   }
   if (property == "$RMS") {
      value->number = node->hypocenters.first->rms; return;
   }
   if (property == "$GAP") {
      value->number = node->hypocenters.first->gap; return;
   }
   if (property == "$NUMSTAT") {
      value->number = (float)node->hypocenters.first->nstat; return;
   }
   if (property == "$MAG") {
      if (!node->hypocenters.nmag) { value->number = -999; return; }
      value->number = node->hypocenters.mag_all[0].mag; return;
   }
   if (property == "$LATERR") {
      value->number = node->hypocenters.first->lat_err; return;
   }
   if (property == "$LONERR") {
      value->number = node->hypocenters.first->lon_err; return;
   }
   if (property == "$DEPTHERR") {
      value->number = node->hypocenters.first->depth_err; return;
   }
   if (property == "$NUMFPS") {
      value->number = node->faults.numlines; return;
   }
   if (property == "$NUMPOL") {
      count = 0;
      if (node->phases.nphase) {
         phase = node->phases.first;
         for (int i=0;i<node->phases.nphase;i++) {
            if (phase->polarity != 32) count++;
            phase = phase->next;
         }
      }
      value->number = count; return;
   }

   // These properties return a string value.
   value->valueType = 1;
   if (property == "$ACTION") {
      if (!strlen(node->action)) { value->string = ""; return; }
      value->string = QString(node->action).toUpper(); return;
   }
   if (property == "$MODEL") {
      if (node->hypocenters.first->model == 32) { value->string = ""; return; }
      value->string = QString(node->hypocenters.first->model); return;
   }
   if (property == "$AGENCY") {
      if (!strlen(node->hypocenters.first->agency)) { value->string = ""; return; }
      value->string = QString(node->hypocenters.first->agency).trimmed().toUpper(); return;
   }
   if (property == "$DISTIND") {
      if (node->hypocenters.first->dist_id == 32) { value->string = ""; return; }
      value->string = QString(node->hypocenters.first->dist_id).toUpper(); return;
   }
   if (property == "$EVENTIND") {
      if (node->hypocenters.first->type == 32) { value->string = ""; return; }
      value->string = QString(node->hypocenters.first->type).toUpper(); return;
   }
   if (property == "$LOCALITY") {
      if (!strlen(node->locality)) { value->string = ""; return; }
      value->string = QString(node->locality).toUpper(); return;
   }
   if (property == "$COMMENTS") {
      if (!node->comments.numlines) { value->string = ""; return; }
      value->string.clear(); NodeGetComments(node, &StrList);
      for (int i=0;i<StrList.size();i++) value->string.append(StrList.at(i) + " ");
      return;
   }
}




// *****************************************************************************
// This function is used to return a single property value for a pahse line in
// the node. The value returned in 'value' is either a float, or a string, or
// nothing if no value exists for the property.
// *****************************************************************************
void NodeGetStationProperty(phase_ *phase, QString property, prop_value *value)
{
   // Initialize 'novalue'.
   value->novalue = false;

   // Make property uppercase.
   property = property.toUpper();

   // These properties return a numerical value.
   value->valueType = 0;
   if (property == "$EPIDIST") { value->number = phase->dist; return; }
   if (property == "$AMPL") { value->number = phase->amp; return; }
   if (property == "$AMPLPERIOD") { value->number = phase->per; return; }
   if (property == "$BACKAZIMUT") { value->number = phase->baz_obs; return; }
   if (property == "$TRVLTMERES") { value->number = phase->res; return; }
   if (property == "$STRDROP") { value->number = phase->spectral.sdrop; return; }
   if (property == "$CORNFREQ") { value->number = phase->spectral.cornerf; return; }

   // These properties return a string value.
   value->valueType = 1;
   if (property == "$STACODE") {
      if (!strlen(phase->agency)) { value->string = ""; return; }
      value->string = QString(phase->agency).toUpper(); return;
   }
   if (property == "$PHASENAME") {
      if (!strlen(phase->phase)) { value->string = ""; return; }
      value->string = QString(phase->phase).toUpper(); return;
   }
   if (property == "$COMP") {
      if (!strlen(phase->co)) { value->string = ""; return; }
      value->string = QString(phase->co).toUpper(); return;
   }
}




// ***************************************************************
// This function is used to check if a node has a
// fault plane solution with a particular quality.
// ***************************************************************
bool NodeHasFpsQuality(event_node_ *Node, QString Quality)
{
   char line[81];
   QString NodeQuality;

   // Search all lines for the FPS quality.
   Quality = Quality.toUpper();
   for (int i=0;i<Node->faults.numlines;i++) {
      memcpy(line, Node->faults.lines+i*80, 80); line[80]=0;
      NodeQuality = line[77];
      NodeQuality = NodeQuality.trimmed().toUpper();
      if (NodeQuality == Quality) return true;
   }
   return false;
}




// ********************************************************************************
// This function returns a list of all fault plane solutions.
// Agency and quality are returned.
// ********************************************************************************
void NodeGetAllFPS(event_node_ *node, QList<agency_fps> *AgencyFPSList)
{
   agency_fps AgencyFps;
   char line[81], agency[4];

   // Create a list off all agencies that has a magnitude.
   for (int i=0;i<node->faults.numlines;i++) {
      // Get the line.
      memcpy(line, node->faults.lines+i*80, 80); line[80]=0;
      // Get the agency.
      memcpy(agency, line+66, 3); agency[3] = 0;
      AgencyFps.agency = agency;
      AgencyFps.agency = AgencyFps.agency.trimmed().toUpper();
      // Get the quality.
      AgencyFps.quality = line[77];
      AgencyFps.quality = AgencyFps.quality.trimmed().toUpper();
      // Add agency magnitudes to list.
      AgencyFPSList->append(AgencyFps);
   }
}




// **********************************************************************************
// This function is used to check if a node has a magnitude with a particular agency.
// **********************************************************************************
bool NodeHasMagAgency(event_node_ *node, QString Agency)
{
   QString NodeAgency;

   // Look for agency in all magnitudes.
   Agency = Agency.toUpper();
   for (int i=0;i<node->hypocenters.nmag;i++) {
      NodeAgency = node->hypocenters.mag_all[i].agency;
      NodeAgency = NodeAgency.trimmed().toUpper();
      if (Agency == NodeAgency) return true;
   }
   return false;
}




// **********************************************************************************
// This function is used to check if a node has a magnitude of a particular type.
// **********************************************************************************
bool NodeHasMagType(event_node_ *node, QString Type)
{
   QString NodeType;

   // Look for type in all magnitudes.
   for (int i=0;i<node->hypocenters.nmag;i++) {
      NodeType = node->hypocenters.mag_all[i].type;
      if (Type == NodeType) return true;
   }
   return false;
}




// ********************************************************************************
// This function returns a list of agencies and their
// corresponding magnitudes from all hypocenters.
// ********************************************************************************
void NodeGetAllHypocenterMags(event_node_ *node, QList<agency_mags> *AgencyMagsList)
{
   float Mag;
   QString Agency, Type;
   agency_mags AgencyMags;
   QStringList AgencyList;

   // Create a list off all agencies that has a magnitude.
   for (int i=0;i<node->hypocenters.nmag;i++) {
      Agency = QString(node->hypocenters.mag_all[i].agency).toUpper().trimmed();
      if (!AgencyList.contains(Agency)) AgencyList.append(Agency);
   }

   // Create a table listing all agencies and their corresponding magnitudes.
   AgencyMags.MB = AgencyMags.Mb = AgencyMags.MC = AgencyMags.ML =
   AgencyMags.MS = AgencyMags.Ms = AgencyMags.MW = -999;
   for (int i=0;i<AgencyList.size();i++) {
      AgencyMags.agency = AgencyList.at(i);
      for (int j=0;j<node->hypocenters.nmag;j++) {
         Type = QString(node->hypocenters.mag_all[j].type);
         Agency = QString(node->hypocenters.mag_all[j].agency).toUpper().trimmed();
         if (Agency == AgencyMags.agency) {
            Mag = node->hypocenters.mag_all[j].mag;
            if (Type == "B") AgencyMags.MB = Mag;
            if (Type == "b") AgencyMags.Mb = Mag;
            if (Type == "S") AgencyMags.MS = Mag;
            if (Type == "s") AgencyMags.Ms = Mag;
            if (Type == "C") AgencyMags.MC = Mag;
            if (Type == "L") AgencyMags.ML = Mag;
            if (Type == "W") AgencyMags.MW = Mag;
         }
      }
      // Add agency magnitudes to list.
      AgencyMagsList->append(AgencyMags);
   }
}




// ********************************************************************************
// This function returns a list of agencies and their
// corresponding magnitudes from the primary hypocenter.
// ********************************************************************************
void NodeGetPriHypocenterMags(event_node_ *node, QList<agency_mags> *AgencyMagsList)
{
   QString Agency;
   hypocenter_ *hyp;
   agency_mags AgencyMags;
   QStringList AgencyList;

   // Create a pointer to the primary hypocenter.
   hyp = node->hypocenters.first;

   // Create a list off all agencies that has a magnitude.
   for (int i=0;i<hyp->magnitudes.MB.nmag;i++) {
      Agency = QString(hyp->magnitudes.MB.mag_agency[i]).toUpper().trimmed();
      if (!AgencyList.contains(Agency)) AgencyList.append(Agency);
   }
   for (int i=0;i<hyp->magnitudes.Mb.nmag;i++) {
      Agency = QString(hyp->magnitudes.Mb.mag_agency[i]).toUpper().trimmed();
      if (!AgencyList.contains(Agency)) AgencyList.append(Agency);
   }
   for (int i=0;i<hyp->magnitudes.MC.nmag;i++) {
      Agency = QString(hyp->magnitudes.MC.mag_agency[i]).toUpper().trimmed();
      if (!AgencyList.contains(Agency)) AgencyList.append(Agency);
   }
   for (int i=0;i<hyp->magnitudes.ML.nmag;i++) {
      Agency = QString(hyp->magnitudes.ML.mag_agency[i]).toUpper().trimmed();
      if (!AgencyList.contains(Agency)) AgencyList.append(Agency);
   }
   for (int i=0;i<hyp->magnitudes.MS.nmag;i++) {
      Agency = QString(hyp->magnitudes.MS.mag_agency[i]).toUpper().trimmed();
      if (!AgencyList.contains(Agency)) AgencyList.append(Agency);
   }
   for (int i=0;i<hyp->magnitudes.Ms.nmag;i++) {
      Agency = QString(hyp->magnitudes.Ms.mag_agency[i]).toUpper().trimmed();
      if (!AgencyList.contains(Agency)) AgencyList.append(Agency);
   }
   for (int i=0;i<hyp->magnitudes.MW.nmag;i++) {
      Agency = QString(hyp->magnitudes.MW.mag_agency[i]).toUpper().trimmed();
      if (!AgencyList.contains(Agency)) AgencyList.append(Agency);
   }

   // Create a table listing all agencies and their corresponding magnitudes.
   AgencyMags.MB = AgencyMags.Mb = AgencyMags.MC = AgencyMags.ML =
   AgencyMags.MS = AgencyMags.Ms = AgencyMags.MW = -999;
   for (int i=0;i<AgencyList.size();i++) {
      AgencyMags.agency = AgencyList.at(i);
      for (int j=0;j<hyp->magnitudes.MB.nmag;j++) {
         Agency = QString(hyp->magnitudes.MB.mag_agency[j]).toUpper().trimmed();
         if (Agency == AgencyMags.agency) AgencyMags.MB = hyp->magnitudes.MB.mag[j];
      }
      for (int j=0;j<hyp->magnitudes.Mb.nmag;j++) {
         Agency = QString(hyp->magnitudes.Mb.mag_agency[j]).toUpper().trimmed();
         if (Agency == AgencyMags.agency) AgencyMags.Mb = hyp->magnitudes.Mb.mag[j];
      }
      for (int j=0;j<hyp->magnitudes.MC.nmag;j++) {
         Agency = QString(hyp->magnitudes.MC.mag_agency[j]).toUpper().trimmed();
         if (Agency == AgencyMags.agency) AgencyMags.MC = hyp->magnitudes.MC.mag[j];
      }
      for (int j=0;j<hyp->magnitudes.ML.nmag;j++) {
         Agency = QString(hyp->magnitudes.ML.mag_agency[j]).toUpper().trimmed();
         if (Agency == AgencyMags.agency) AgencyMags.ML = hyp->magnitudes.ML.mag[j];
      }
      for (int j=0;j<hyp->magnitudes.MS.nmag;j++) {
         Agency = QString(hyp->magnitudes.MS.mag_agency[j]).toUpper().trimmed();
         if (Agency == AgencyMags.agency) AgencyMags.MS = hyp->magnitudes.MS.mag[j];
      }
      for (int j=0;j<hyp->magnitudes.Ms.nmag;j++) {
         Agency = QString(hyp->magnitudes.Ms.mag_agency[j]).toUpper().trimmed();
         if (Agency == AgencyMags.agency) AgencyMags.Ms = hyp->magnitudes.Ms.mag[j];
      }
      for (int j=0;j<hyp->magnitudes.MW.nmag;j++) {
         Agency = QString(hyp->magnitudes.MW.mag_agency[j]).toUpper().trimmed();
         if (Agency == AgencyMags.agency) AgencyMags.MW = hyp->magnitudes.MW.mag[j];
      }

      // Add agency magnitudes to list.
      AgencyMagsList->append(AgencyMags);
   }
}




// *********************************************************************
// This function is used by the database. It reads a single S-file into
// a database node structure. Memory for the node is allocated here.
// Function returns a pointer to the new node. One of the following
// status values are also returned in 'status':
// 0 - No error or warning reading S-file into memory.
// 1 - Unable to open S-file.
// 2 - Function rea_event_in() returned an error.
// 3 - Memory allocation failed, out of memory.
// 4 - Function rea_event_in() returned a non-fatal error.
// 5 - S-file has an invalid date.
// *********************************************************************
event_node_* ReadSfile(char file[], char *status)
{
   QFile Sfile;
   QDate fDate;
   int index, i, i2;
   qint64 LineLength;
   event_node_ *Node;
   bool OutOfMem, Done;
   sint unit, read_code, all;
   phase_ *PhaseData, *LastPhase = 0;
   hypocenter_ *HypData, *LastHypo = 0;
   char line[192], data[MAX_SFILE_LINES][80], str4[5];

   // Create new pointers to fortran common blocks.
   // This is done to simplify debugging in Qt Creator.
   comblck_hyp1* hyp1 = &hyp1_; comblck_hyp4* hyp4 = &hyp4_;
   comblck_hyp5* hyp5 = &hyp5_; comblck_rea1* rea1 = &rea1_;
   comblck_rea2* rea2 = &rea2_; comblck_rea3* rea3 = &rea3_;
   comblck_rea4* rea4 = &rea4_; comblck_rea5* rea5 = &rea5_;
   comblck_rea8* rea8 = &rea8_;

   // Open the S-file. Abort if file can not be opened.
   *status = 0;
   Sfile.setFileName(file);
   if (!Sfile.open(QIODevice::ReadOnly | QIODevice::Text)) { *status = 1; return 0; }

   // Read S-file into the 'data' array.
   Done = false;
   rea4_.rea_nrecord = rea4->rea_nhead = 0;
   while(rea4->rea_nrecord<MAX_SFILE_LINES && !Done) {
      LineLength = Sfile.readLine(line, sizeof(line));
      if (LineLength >=3 && !IsDataLineEmpty(line, LineLength)) {
         memset((void*)data[rea4->rea_nrecord], 32, 80);
         memcpy((void*)data[rea4->rea_nrecord], (void*)line, LineLength - 1);
         if (LineLength == 81 && line[79] != 32 && line[79] != '4') rea4->rea_nhead++;
         rea4->rea_nrecord++;
      }
      Done = Sfile.atEnd();
   }

   // Close the S-file.
   Sfile.close();

   // Read S-file from 'data' array into common blocks.
   rea_event_in_(&(unit=0), &(all=1), (char*)data, &read_code, sizeof(data));

   // Handle errors and warnings from 'rea_event_in'.
   if (read_code) {
      if (read_code == 1) { *status = 2; return 0; }
      if (read_code == 2) *status = 4;
      if (read_code == 3) { *status = 2; return 0; }
   }

   // Abort if no header lines have been read.
   if (!rea4->rea_nhead) { *status = 2; return 0; }

   // Abort if no hypocenters have been read.
   if (!rea4->rea_nhyp) { *status = 2; return 0; }

   // Make sure first hypocenter at least has a year.
   if (hyp4->hyp_year[0] == -999) { *status = 2; return 0; }

   // Allocate memory for a new db node.
   Node = new (nothrow) event_node_;
   if (!Node) { *status = 3; return 0; }

   // Clear memory for the new db node.
   memset(Node, 0, sizeof(event_node_));

   // Copy all information from common blocks to new db node.
   // Set name of S-file.
   strcpy(Node->metadata.sfile, file);

   // Copy number of header lines.
   Node->nhead = rea4->rea_nhead;

   // Copy number of lines.
   Node->nrecord = rea4->rea_nrecord;

   // Copy number of stations.
   Node->nstat = rea4->rea_nstat;

   // Copy number of spectra.
   Node->nspec = rea4->rea_nspec;

   // Copy action parameter.
   memcpy((void*)Node->action, (void*)rea3->rea_action, 3);
   UnPadString(Node->action, sizeof(Node->action));

   // Copy locality.
   memcpy((void*)Node->locality, (void*)rea4->rea_locality, 68);
   UnPadString(Node->locality, sizeof(Node->locality));

   // Copy macro data (type 2 lines).
   OutOfMem = false;
   Node->macros.numlines = rea4->rea_nmacro;
   if (Node->macros.numlines) {
      if ((Node->macros.lines = new (nothrow) char [rea4->rea_nmacro*80])) {
         for (int i=0;i<rea4->rea_nmacro;i ++) {
            memcpy((Node->macros.lines+i*80), rea4->rea_macro[i], 80);
         }
      } else {
         OutOfMem = true;
      }
   }

   // Copy wave filenames (type 6 lines).
   Node->wavefiles.numlines = rea4->rea_nwav;
   if (Node->wavefiles.numlines && !OutOfMem) {
      if ((Node->wavefiles.lines = new (nothrow) char [rea4->rea_nwav*80])) {
         for (int i=0;i<rea4->rea_nwav;i ++) {
            memcpy((Node->wavefiles.lines+i*80), rea4->rea_wav[i], 80);
         }
      } else {
         OutOfMem = true;
      }
   }

   // Copy faults (type F lines).
   Node->faults.numlines = rea4->rea_nfault;
   if (Node->faults.numlines && !OutOfMem) {
      if ((Node->faults.lines = new (nothrow) char [rea4->rea_nfault*80])) {
         for (int i=0;i<rea4->rea_nfault;i ++) {
            memcpy((Node->faults.lines+i*80), rea4->rea_fault[i], 80);
         }
      } else {
         OutOfMem = true;
      }
   }

   // Copy comment lines (type 3 lines).
   Node->comments.numlines = rea4->rea_ncomment;
   if (Node->comments.numlines && !OutOfMem) {
      if ((Node->comments.lines = new (nothrow) char [rea4->rea_ncomment*80])) {
         for (int i=0;i<rea4->rea_ncomment;i ++) {
            memcpy((Node->comments.lines+i*80), rea4->rea_comment[i], 80);
         }
      } else {
         OutOfMem = true;
      }
   }

   // Copy ID-line.
   memcpy((void*)Node->id_line, (void*)rea4->rea_id_line, 80);

   // Copy phase data to event node.
   LastPhase = 0;
   Node->phases.nphase = rea4->rea_nphase;
   for (index=0;index<Node->phases.nphase;index ++) {
      // Copy data for phase pointed to by 'index'.
      if (!OutOfMem && (PhaseData = new (nothrow) phase_)) {
         memset(PhaseData, 0, sizeof(phase_));
         memcpy(PhaseData->agency, rea5->rea_stat[index], 5);
         UnPadString(PhaseData->agency, 6);
         memcpy(PhaseData->comp, rea4->rea_comp[index], 4);
         memcpy(PhaseData->co, rea2->rea_co[index], 2);
         memcpy(PhaseData->phase, rea8->rea_phase[index], 8);
         PhaseData->onset = rea1->rea_onset[index];
         PhaseData->weight_in = rea1->rea_weight_in[index];
         memcpy(PhaseData->weight_out, rea2->rea_weight_out[index], 2);
         PhaseData->polarity = rea1->rea_polarity[index];
         PhaseData->year = rea4->rea_year[index];
         PhaseData->month = rea4->rea_month[index];
         PhaseData->day = rea4->rea_day[index];
         PhaseData->hour = rea4->rea_hour[index];
         PhaseData->min = rea4->rea_min[index];
         PhaseData->sec = rea4->rea_sec[index];
         PhaseData->abs_time = rea8->rea_abs_time[index];
         PhaseData->coda = rea4->rea_coda[index];
         PhaseData->amp = rea4->rea_amp[index];
         PhaseData->per = rea4->rea_per[index];
         PhaseData->baz_obs = rea4->rea_baz_obs[index];
         PhaseData->baz_cal = rea4->rea_baz_cal[index];
         PhaseData->vel = rea4->rea_vel[index];
         PhaseData->ain = rea4->rea_ain[index];
         PhaseData->baz_res = rea4->rea_baz_res[index];
         PhaseData->res = rea4->rea_res[index];
         PhaseData->dist = rea4->rea_dist[index];
         PhaseData->az = rea4->rea_az[index];
         memcpy(PhaseData->autoproc, rea4->rea_auto[index], 20);
         UnPadString(PhaseData->autoproc, 21);
         PhaseData->spectral.moment = rea4->rea_moment[index];
         PhaseData->spectral.sdrop = rea4->rea_sdrop[index];
         PhaseData->spectral.omega0 = rea4->rea_omega0[index];
         PhaseData->spectral.cornerf = rea4->rea_cornerf[index];
         PhaseData->spectral.radius = rea4->rea_radius[index];
         PhaseData->spectral.swin = rea4->rea_swin[index];
         PhaseData->spectral.vs = rea4->rea_vs[index];
         PhaseData->spectral.vp = rea4->rea_vp[index];
         PhaseData->spectral.q0 = rea4->rea_q0[index];
         PhaseData->spectral.qalpha = rea4->rea_qalpha[index];
         PhaseData->spectral.kappa = rea4->rea_kappa[index];
         PhaseData->spectral.density = rea4->rea_density[index];
         PhaseData->spectral.slope = rea4->rea_slope[index];
         PhaseData->spectral.geo_dist = rea4->rea_geo_dist[index];
         PhaseData->magnitudes.mc = rea4->rea_mc[index];
         PhaseData->magnitudes.ml = rea4->rea_ml[index];
         PhaseData->magnitudes.mb = rea4->rea_mb[index];
         PhaseData->magnitudes.ms = rea4->rea_ms[index];
         PhaseData->magnitudes.mw = rea4->rea_mw[index];
         PhaseData->next = (phase_*)0;
         // Add phase to event node.
         if (!index) {
            Node->phases.first = PhaseData;
            LastPhase = PhaseData;
         } else {
            LastPhase->next = PhaseData;
            LastPhase = PhaseData;
         }
      } else {
         OutOfMem = true;
      }
   }

   // Copy spectral averages to event node.
   Node->spectral_avg.av_moment = rea4->rea_av_moment;
   Node->spectral_avg.av_sdrop = rea4->rea_av_sdrop;
   Node->spectral_avg.av_omega0 = rea4->rea_av_omega0;
   Node->spectral_avg.av_cornerf = rea4->rea_av_cornerf;
   Node->spectral_avg.av_radius = rea4->rea_av_radius;
   Node->spectral_avg.av_swin = rea4->rea_av_swin;
   Node->spectral_avg.av_mw = rea4->rea_av_mw;
   Node->spectral_avg.av_slope = rea4->rea_av_slope;

   // Copy hypocenter data to event node.
   Node->hypocenters.nmag = rea4->rea_nmag;
   if (rea4->rea_nmag) {
      if (!OutOfMem && (Node->hypocenters.mag_all = new (nothrow) mag_info [rea4->rea_nmag])) {
         memset(Node->hypocenters.mag_all, 0, sizeof(mag_info)*rea4->rea_nmag);
         for (int i=0;i<Node->hypocenters.nmag;i ++) {
            Node->hypocenters.mag_all[i].mag = hyp4->hyp_mag_all[i];
            Node->hypocenters.mag_all[i].type = hyp1->hyp_mag_type_all[i];
            memcpy(Node->hypocenters.mag_all[i].agency, hyp5->hyp_mag_agency_all[i], 5);
         }
      } else {
         OutOfMem = true;
      }
   }
   Node->hypocenters.nhyp = rea4->rea_nhyp;
   for (index=0;index<Node->hypocenters.nhyp;index ++) {
      // Copy data for hypocenter pointed to by 'index'.
      if (!OutOfMem && (HypData = new (nothrow) hypocenter_)) {
         memset(HypData, 0, sizeof(hypocenter_));
         HypData->time.year = hyp4->hyp_year[index];
         HypData->time.month = hyp4->hyp_month[index];
         HypData->time.day = hyp4->hyp_day[index];
         HypData->time.hour = hyp4->hyp_hour[index];
         HypData->time.minute = hyp4->hyp_min[index];
         HypData->time.second = hyp4->hyp_sec[index];
         HypData->model = hyp1->hyp_model[index];
         HypData->dist_id = hyp1->hyp_dist_id[index];
         HypData->type = hyp1->hyp_type[index];
         HypData->fix_org = hyp1->hyp_fix_org[index];
         HypData->lat = hyp4->hyp_lat[index];
         HypData->lon = hyp4->hyp_lon[index];
         HypData->depth = hyp4->hyp_depth[index];
         HypData->depth_flag = hyp1->hyp_depth_flag[index];
         HypData->epi_flag = hyp1->hyp_epi_flag[index];
         memcpy(HypData->agency, hyp5->hyp_agency[index], 5);
         UnPadString(HypData->agency, 6);
         HypData->nstat = hyp4->hyp_nstat[index];
         HypData->rms = hyp4->hyp_rms[index];
         HypData->high_accuracy = hyp1->hyp_high_accuracy[index];
         HypData->error = hyp1->hyp_error[index];
         HypData->gap = hyp4->hyp_gap[index];
         HypData->sec_err = hyp4->hyp_sec_err[index];
         HypData->lat_err = hyp4->hyp_lat_err[index];
         HypData->lon_err = hyp4->hyp_lon_err[index];
         HypData->depth_err = hyp4->hyp_depth_err[index];
         HypData->cov[0] = hyp4->hyp_cov[index][0];
         HypData->cov[1] = hyp4->hyp_cov[index][1];
         HypData->cov[2] = hyp4->hyp_cov[index][2];
         memcpy(HypData->autoproc, hyp4->hyp_auto[index], 20);
         UnPadString(HypData->autoproc, 21);
         // Copy magnitude data.
         for (i=0;i<6;i++) {
            if (hyp1->hyp_mag_type[index][i] == 'W') {
               i2 = HypData->magnitudes.MW.nmag++;
               HypData->magnitudes.MW.mag[i2] = hyp4->hyp_mag[index][i];
               memcpy(HypData->magnitudes.MW.mag_agency[i2], hyp5->hyp_mag_agency[index][i], 5);
            }
            if (hyp1->hyp_mag_type[index][i] == 'L') {
               i2 = HypData->magnitudes.ML.nmag++;
               HypData->magnitudes.ML.mag[i2] = hyp4->hyp_mag[index][i];
               memcpy(HypData->magnitudes.ML.mag_agency[i2], hyp5->hyp_mag_agency[index][i], 5);
            }
            if (hyp1->hyp_mag_type[index][i] == 'C') {
               i2 = HypData->magnitudes.MC.nmag++;
               HypData->magnitudes.MC.mag[i2] = hyp4->hyp_mag[index][i];
               memcpy(HypData->magnitudes.MC.mag_agency[i2], hyp5->hyp_mag_agency[index][i], 5);
            }
            if (hyp1->hyp_mag_type[index][i] == 'b') {
               i2 = HypData->magnitudes.Mb.nmag++;
               HypData->magnitudes.Mb.mag[i2] = hyp4->hyp_mag[index][i];
               memcpy(HypData->magnitudes.Mb.mag_agency[i2], hyp5->hyp_mag_agency[index][i], 5);
            }
            if (hyp1->hyp_mag_type[index][i] == 'B') {
               i2 = HypData->magnitudes.MB.nmag++;
               HypData->magnitudes.MB.mag[i2] = hyp4->hyp_mag[index][i];
               memcpy(HypData->magnitudes.MB.mag_agency[i2], hyp5->hyp_mag_agency[index][i], 5);
            }
            if (hyp1->hyp_mag_type[index][i] == 's') {
               i2 = HypData->magnitudes.Ms.nmag++;
               HypData->magnitudes.Ms.mag[i2] = hyp4->hyp_mag[index][i];
               memcpy(HypData->magnitudes.Ms.mag_agency[i2], hyp5->hyp_mag_agency[index][i], 5);
            }
            if (hyp1->hyp_mag_type[index][i] == 'S') {
               i2 = HypData->magnitudes.MS.nmag++;
               HypData->magnitudes.MS.mag[i2] = hyp4->hyp_mag[index][i];
               memcpy(HypData->magnitudes.MS.mag_agency[i2], hyp5->hyp_mag_agency[index][i], 5);
            }
         }
         HypData->next = (hypocenter_*)0;

         // Add hypocenter to event node.
         if (!index) {
            Node->hypocenters.first = HypData;
            LastHypo = HypData;
         } else {
            LastHypo->next = HypData;
            LastHypo = HypData;
         }
      } else {
         OutOfMem = true;
      }
   }

   // Create a date/time string in the metadata structure.
   HypData = Node->hypocenters.first;
   sprintf(str4, "%04u", HypData->time.year);
   strcpy(Node->metadata.datetime, str4);
   sprintf(str4, "%02u", Node->hypocenters.first->time.month);
   if (HypData->time.month == -999) strcpy(str4, "??");
   strcat(Node->metadata.datetime, str4);
   sprintf(str4, "%02u", Node->hypocenters.first->time.day);
   if (HypData->time.day == -999) strcpy(str4, "??");
   strcat(Node->metadata.datetime, str4);
   sprintf(str4, "%02u", Node->hypocenters.first->time.hour);
   if (HypData->time.hour == -999) strcpy(str4, "??");
   strcat(Node->metadata.datetime, str4);
   sprintf(str4, "%02u", Node->hypocenters.first->time.minute);
   if (HypData->time.minute == -999) strcpy(str4, "??");
   strcat(Node->metadata.datetime, str4);

   // If we have run out of memory, free up all allocated node memory.
   if (OutOfMem) { FreeNode(Node); *status = 3; return 0; }

   // Check if file has an invalid date.
   fDate.setDate(HypData->time.year, HypData->time.month, HypData->time.day);
   if (!fDate.isValid()) { FreeNode(Node); *status = 5; return 0; }

   // Return pointer to new node.
   return Node;
}




// *************************************************************************
// This function writes an S-file from an event node. Any existing file will
// be overwritten. Function returns 'false' if file could not be written.
//
// One of the following status values are also returned in 'status':
// 0 - No error occurred.
// 1 - Function 'rea_event_out' returned an error.
// 2 - Unable to open S-file.
// 3 - Unable to write S-file.
// *************************************************************************
bool WriteSfile(event_node_ *Node, char *status)
{
   QFile Sfile;
   QByteArray line;
   int index, i, i2;
   bool bResult;
   signed int unit, write_code, all;
   char data[MAX_SFILE_LINES][80];
   phase_ *PhaseData;
   hypocenter_ *HypData;

   // Create new pointers to fortran common blocks.
   // This is done to simplify debugging in QtCreator.
   comblck_hyp1* hyp1 = &hyp1_; comblck_hyp4* hyp4 = &hyp4_;
   comblck_hyp5* hyp5 = &hyp5_; comblck_rea1* rea1 = &rea1_;
   comblck_rea2* rea2 = &rea2_; comblck_rea3* rea3 = &rea3_;
   comblck_rea4* rea4 = &rea4_; comblck_rea5* rea5 = &rea5_;
   comblck_rea8* rea8 = &rea8_;

   // Clear the common blocks.
   ClearCommonBlocks();

   // Copy all information from db node to common blocks.
   // Copy number of stations.
   rea4->rea_nstat = Node->nstat;

   // Copy number of spectra.
   rea4->rea_nspec = Node->nspec;

   // Copy action parameter.
   fstrcpy(rea3->rea_action, Node->action, 3);

   // Copy locality.
   fstrcpy(rea4->rea_locality, Node->locality, 68);

   // Copy macro data.
   rea4->rea_nmacro = Node->macros.numlines;
   for (int i=0;i<rea4->rea_nmacro;i++) {
      memcpy(rea4->rea_macro[i], Node->macros.lines+i*80, 80);
   }

   // Copy wave filenames.
   rea4->rea_nwav = Node->wavefiles.numlines;
   for (int i=0;i<rea4->rea_nwav;i++) {
      memcpy(rea4->rea_wav[i], Node->wavefiles.lines+i*80, 80);
   }

   // Copy faults.
   rea4->rea_nfault = Node->faults.numlines;
   for (int i=0;i<rea4->rea_nfault;i++) {
      memcpy(rea4->rea_fault[i], Node->faults.lines+i*80, 80);
   }

   // Copy comment lines.
   rea4->rea_ncomment = Node->comments.numlines;
   for (int i=0;i<rea4->rea_ncomment;i++) {
      memcpy(rea4->rea_comment[i], Node->comments.lines+i*80, 80);
   }

   // Copy ID-line.
   fstrcpy(rea4->rea_id_line, Node->id_line, 80);

   // Copy phases from event node.
   rea4->rea_nphase = Node->phases.nphase;
   if (rea4->rea_nphase) {
      PhaseData = Node->phases.first;
      for (index=0;index<Node->phases.nphase;index++) {
         // Copy data for one phase.
         fstrcpy(rea5->rea_stat[index], PhaseData->agency, 5);
         fstrcpy(rea4->rea_comp[index], PhaseData->comp, 4);
         fstrcpy(rea2->rea_co[index], PhaseData->co, 2);
         fstrcpy(rea8->rea_phase[index], PhaseData->phase, 8);
         rea1->rea_onset[index] = PhaseData->onset;
         rea1->rea_weight_in[index] = PhaseData->weight_in;
         fstrcpy(rea2->rea_weight_out[index], PhaseData->weight_out, 2);
         rea1->rea_polarity[index] = PhaseData->polarity;
         rea4->rea_year[index] = PhaseData->year;
         rea4->rea_month[index] = PhaseData->month;
         rea4->rea_day[index] = PhaseData->day;
         rea4->rea_hour[index] = PhaseData->hour;
         rea4->rea_min[index] = PhaseData->min;
         rea4->rea_sec[index] = PhaseData->sec;
         rea8->rea_abs_time[index] = PhaseData->abs_time;
         rea4->rea_coda[index] = PhaseData->coda;
         rea4->rea_amp[index] = PhaseData->amp;
         rea4->rea_per[index] = PhaseData->per;
         rea4->rea_baz_obs[index] = PhaseData->baz_obs;
         rea4->rea_baz_cal[index] = PhaseData->baz_cal;
         rea4->rea_vel[index] = PhaseData->vel;
         rea4->rea_ain[index] = PhaseData->ain;
         rea4->rea_baz_res[index] = PhaseData->baz_res;
         rea4->rea_res[index] = PhaseData->res;
         rea4->rea_dist[index] = PhaseData->dist;
         rea4->rea_az[index] = PhaseData->az;
         fstrcpy(rea4->rea_auto[index], PhaseData->autoproc, 20);
         rea4->rea_moment[index] = PhaseData->spectral.moment;
         rea4->rea_sdrop[index] = PhaseData->spectral.sdrop;
         rea4->rea_omega0[index] = PhaseData->spectral.omega0;
         rea4->rea_cornerf[index] = PhaseData->spectral.cornerf;
         rea4->rea_radius[index] = PhaseData->spectral.radius;
         rea4->rea_swin[index] = PhaseData->spectral.swin;
         rea4->rea_vs[index] = PhaseData->spectral.vs;
         rea4->rea_vp[index] = PhaseData->spectral.vp;
         rea4->rea_q0[index] = PhaseData->spectral.q0;
         rea4->rea_qalpha[index] = PhaseData->spectral.qalpha;
         rea4->rea_kappa[index] = PhaseData->spectral.kappa;
         rea4->rea_density[index] = PhaseData->spectral.density;
         rea4->rea_slope[index] = PhaseData->spectral.slope;
         rea4->rea_geo_dist[index] = PhaseData->spectral.geo_dist;
         rea4->rea_mc[index] = PhaseData->magnitudes.mc;
         rea4->rea_ml[index] = PhaseData->magnitudes.ml;
         rea4->rea_mb[index] = PhaseData->magnitudes.mb;
         rea4->rea_ms[index] = PhaseData->magnitudes.ms;
         rea4->rea_mw[index] = PhaseData->magnitudes.mw;
         // Move to next phase.
         PhaseData = PhaseData->next;
      }
   }

   // Copy spectral averages.
   rea4->rea_av_moment = Node->spectral_avg.av_moment;
   rea4->rea_av_sdrop = Node->spectral_avg.av_sdrop;
   rea4->rea_av_omega0 = Node->spectral_avg.av_omega0;
   rea4->rea_av_cornerf = Node->spectral_avg.av_cornerf;
   rea4->rea_av_radius = Node->spectral_avg.av_radius;
   rea4->rea_av_swin = Node->spectral_avg.av_swin;
   rea4->rea_av_mw = Node->spectral_avg.av_mw;
   rea4->rea_av_slope = Node->spectral_avg.av_slope;

   // Copy hypocenter data.
   rea4->rea_nmag = Node->hypocenters.nmag;
   for (int i=0;i<rea4->rea_nmag;i++) {
      hyp4->hyp_mag_all[i] = Node->hypocenters.mag_all[i].mag;
      hyp1->hyp_mag_type_all[i] = Node->hypocenters.mag_all[i].type;
      fstrcpy(hyp5->hyp_mag_agency_all[i], Node->hypocenters.mag_all[i].agency, 5);
   }
   rea4->rea_nhyp = Node->hypocenters.nhyp;
   if (rea4->rea_nhyp) {
      HypData = Node->hypocenters.first;
      for (index=0;index<Node->hypocenters.nhyp;index ++) {
         hyp4->hyp_year[index] = HypData->time.year;
         hyp4->hyp_month[index] = HypData->time.month;
         hyp4->hyp_day[index] = HypData->time.day;
         hyp4->hyp_hour[index] = HypData->time.hour;
         hyp4->hyp_min[index] = HypData->time.minute;
         hyp4->hyp_sec[index] = HypData->time.second;
         hyp1->hyp_model[index] = HypData->model;
         hyp1->hyp_dist_id[index] = HypData->dist_id;
         hyp1->hyp_type[index] = HypData->type;
         hyp1->hyp_fix_org[index] = HypData->fix_org;
         hyp4->hyp_lat[index] = HypData->lat;
         hyp4->hyp_lon[index] = HypData->lon;
         hyp4->hyp_depth[index] = HypData->depth;
         hyp1->hyp_depth_flag[index] = HypData->depth_flag;
         hyp1->hyp_epi_flag[index] = HypData->epi_flag;
         fstrcpy(hyp5->hyp_agency[index], HypData->agency, 5);
         hyp4->hyp_nstat[index] = HypData->nstat;
         hyp4->hyp_rms[index] = HypData->rms;
         hyp1->hyp_high_accuracy[index] = HypData->high_accuracy;
         hyp1->hyp_error[index] = HypData->error;
         hyp4->hyp_gap[index] = HypData->gap;
         hyp4->hyp_sec_err[index] = HypData->sec_err;
         hyp4->hyp_lat_err[index] = HypData->lat_err;
         hyp4->hyp_lon_err[index] = HypData->lon_err;
         hyp4->hyp_depth_err[index] = HypData->depth_err;
         hyp4->hyp_cov[index][0] = HypData->cov[0];
         hyp4->hyp_cov[index][1] = HypData->cov[1];
         hyp4->hyp_cov[index][2] = HypData->cov[2];
         fstrcpy(hyp4->hyp_auto[index], HypData->autoproc, 20);
         // Copy magnitude data (MW).
         i2 = 0;
         for (i=0;i<HypData->magnitudes.MW.nmag;i++) {
            hyp1->hyp_mag_type[index][i2] = 'W';
            hyp4->hyp_mag[index][i2] = HypData->magnitudes.MW.mag[i];
            fstrcpy(hyp5->hyp_mag_agency[index][i2], HypData->magnitudes.MW.mag_agency[i], 5);
            i2++;
         }
         for (i=0;i<HypData->magnitudes.ML.nmag;i++) {
            hyp1->hyp_mag_type[index][i2] = 'L';
            hyp4->hyp_mag[index][i2] = HypData->magnitudes.ML.mag[i];
            fstrcpy(hyp5->hyp_mag_agency[index][i2], HypData->magnitudes.ML.mag_agency[i], 5);
            i2++;
         }
         for (i=0;i<HypData->magnitudes.MC.nmag;i++) {
            hyp1->hyp_mag_type[index][i2] = 'C';
            hyp4->hyp_mag[index][i2] = HypData->magnitudes.MC.mag[i];
            fstrcpy(hyp5->hyp_mag_agency[index][i2], HypData->magnitudes.MC.mag_agency[i], 5);
            i2++;
         }
         for (i=0;i<HypData->magnitudes.Mb.nmag;i++) {
            hyp1->hyp_mag_type[index][i2] = 'b';
            hyp4->hyp_mag[index][i2] = HypData->magnitudes.Mb.mag[i];
            fstrcpy(hyp5->hyp_mag_agency[index][i2], HypData->magnitudes.Mb.mag_agency[i], 5);
            i2++;
         }
         for (i=0;i<HypData->magnitudes.MB.nmag;i++) {
            hyp1->hyp_mag_type[index][i2] = 'B';
            hyp4->hyp_mag[index][i2] = HypData->magnitudes.MB.mag[i];
            fstrcpy(hyp5->hyp_mag_agency[index][i2], HypData->magnitudes.MB.mag_agency[i], 5);
            i2++;
         }
         for (i=0;i<HypData->magnitudes.Ms.nmag;i++) {
            hyp1->hyp_mag_type[index][i2] = 's';
            hyp4->hyp_mag[index][i2] = HypData->magnitudes.Ms.mag[i];
            fstrcpy(hyp5->hyp_mag_agency[index][i2], HypData->magnitudes.Ms.mag_agency[i], 5);
            i2++;
         }
         for (i=0;i<HypData->magnitudes.MS.nmag;i++) {
            hyp1->hyp_mag_type[index][i2] = 'S';
            hyp4->hyp_mag[index][i2] = HypData->magnitudes.MS.mag[i];
            fstrcpy(hyp5->hyp_mag_agency[index][i2], HypData->magnitudes.MS.mag_agency[i], 5);
            i2++;
         }

         // Move to next hypocenter.
         HypData = HypData->next;
      }
   }

   // Create the S-file contents from the data in the common blocks.
   // Function rea_event_out will write S-file lines into 'data' array.
   rea4_.rea_err_unit = 1;
   memset((void*)data, 32, sizeof(data));
   rea_event_out_(&(unit=0), &(all=1), (char*)data, &write_code, sizeof(data));
   if (write_code) { *status = 2; return 0; }

   // Create the S-file. File will be overwritten if it already exists.
   Sfile.setFileName(QString(Node->metadata.sfile));
   bResult = Sfile.open(QIODevice::WriteOnly|QIODevice::Text|QIODevice::Truncate);
   if (!bResult) { *status = 2; return false; }

   // Write the new S-file from the 'data' array. Abort if write a error occur.
   for (index=0;index<rea4_.rea_nrecord;index++) {
      line.clear(); line.append(data[index], 80); line.append('\n');
      if (Sfile.write(line, 81) == -1) { Sfile.close(); *status = 3; return false; }
   }

   // S-file was created sucessfully. Close file and return.
   Sfile.close(); *status = 0;
   return true;
}




// *****************************************************
// This function is used by the database to
// free up all memory used by a database node.
// *****************************************************
void FreeNode(event_node_ *Node)
{
   // Check for nullpointer to node.
   if (!Node) return;

   // Free macrolines.
   if (Node->macros.numlines) delete [] Node->macros.lines;

   // Free wave file names.
   if (Node->wavefiles.numlines) delete [] Node->wavefiles.lines;

   // Free fault lines.
   if (Node->faults.numlines) delete [] Node->faults.lines;

   // Free comment lines.
   if (Node->comments.numlines) delete [] Node->comments.lines;

   // Free all phase structures.
   struct phase_ *Phase, *NextPhase;
   if (Node->phases.nphase) {
      Phase = Node->phases.first;
      while (Phase) {
         NextPhase = Phase->next;
         delete Phase;
         Phase = NextPhase;
      }
   }

   // Free all hypocenter structures.
   struct hypocenter_ *Hyp, *NextHyp;
   if (Node->hypocenters.nhyp) {
      delete [] Node->hypocenters.mag_all;
      Hyp = Node->hypocenters.first;
      while (Hyp) {
         NextHyp = Hyp->next;
         delete Hyp;
         Hyp = NextHyp;
      }
   }

   // Free up the node itself.
   delete Node;
}




// *****************************************************************
// This function returns the distance indicator for an event node.
// ******************************************************************
QChar GetNodeDistanseIndicator(event_node_* Node)
{
   return QChar(Node->hypocenters.first->dist_id);
}




// *****************************************************************
// This function sets the distance indicator for an event node.
// ******************************************************************
void SetNodeDistanseIndicator(event_node_* Node, QChar DistInd)
{
   // Change the event file name.
   int len = strlen(Node->metadata.sfile);
   Node->metadata.sfile[len-9] = DistInd.toAscii();

   // Change the first hypocenter line.
   Node->hypocenters.first->dist_id = DistInd.toAscii();
}




// *****************************************************************
// This function compares two event nodes for sorting purposes.
// The currently selected sort order is taken into consideration.
// Return values: -1=node1<node2 ,0=nodes are equal , 1=node1>node2.
// A node with missing data is larger (>) then a node with data.
// ******************************************************************
char NodeCmp(event_node_ *node1, event_node_ *node2, sortCol SortColumn)
{
   int iValue1, iValue2;
   float fValue1, fValue2;

   switch (SortColumn) {
      case Action:
      // A 3 character string.
      // If first string character is 0, then data is missing.
      iValue1 = (int)node1->action[0];
      iValue2 = (int)node2->action[0];
      if (!iValue1 && !iValue2) return 0;
      if (iValue1 && iValue2) {
         return (char)strcmp(node1->action, node2->action);
      } else {
         if (!iValue1) { return 1; } else { return -1; }
      }

      case DateTime:
      if (node1->hypocenters.first->time.year < node2->hypocenters.first->time.year) return -1;
      if (node1->hypocenters.first->time.year > node2->hypocenters.first->time.year) return 1;
      if (node1->hypocenters.first->time.month < node2->hypocenters.first->time.month) return -1;
      if (node1->hypocenters.first->time.month > node2->hypocenters.first->time.month) return 1;
      if (node1->hypocenters.first->time.day < node2->hypocenters.first->time.day) return -1;
      if (node1->hypocenters.first->time.day > node2->hypocenters.first->time.day) return 1;
      if (node1->hypocenters.first->time.hour < node2->hypocenters.first->time.hour) return -1;
      if (node1->hypocenters.first->time.hour > node2->hypocenters.first->time.hour) return 1;
      if (node1->hypocenters.first->time.minute < node2->hypocenters.first->time.minute) return -1;
      if (node1->hypocenters.first->time.minute > node2->hypocenters.first->time.minute) return 1;
      if (node1->hypocenters.first->time.second < node2->hypocenters.first->time.second) return -1;
      if (node1->hypocenters.first->time.second > node2->hypocenters.first->time.second) return 1;
      return 0;

      case Latitude:
      fValue1 = node1->hypocenters.first->lat;
      fValue2 = node2->hypocenters.first->lat;
      if (fValue1 == fValue2) return 0;
      if (fValue1 != -999.0 && fValue2 != -999.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == -999.0) { return 1; } else { return -1; };
      }

      case Longitude:
      fValue1 = node1->hypocenters.first->lon;
      fValue2 = node2->hypocenters.first->lon;
      if (fValue1 == fValue2) return 0;
      if (fValue1 != -999.0 && fValue2 != -999.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == -999.0) { return 1; } else { return -1; };
      }

      case Depth:
      fValue1 = node1->hypocenters.first->depth;
      fValue2 = node2->hypocenters.first->depth;
      if (fValue1 == fValue2) return 0;
      if (fValue1 != -999.0 && fValue2 != -999.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == -999.0) { return 1; } else { return -1; };
      }

      case Model:
      // One character. A value of '32' means data missing.
      iValue1 = (int)node1->hypocenters.first->model;
      iValue2 = (int)node2->hypocenters.first->model;
      if (iValue1 == iValue2) return 0;
      if (iValue1 != 32 && iValue2 != 32) {
         if (iValue1 > iValue2) { return 1; } else { return -1; }
      } else {
         if (iValue1 == 32) { return 1; } else { return -1; }
      }

      case Agency:
      // A 3 character string.
      iValue1 = (int)node1->hypocenters.first->agency[0];
      iValue2 = (int)node2->hypocenters.first->agency[0];
      if (iValue1 == 32 && iValue2 == 32) return 0;
      if (iValue1 != 32 && iValue2 != 32) {
         return (char)strcmp(node1->hypocenters.first->agency, node2->hypocenters.first->agency);
      } else {
         if (iValue1 == 32)  { return 1; } else { return -1; }
      }

      case RMS:
      fValue1 = node1->hypocenters.first->rms;
      fValue2 = node2->hypocenters.first->rms;
      if (fValue1 == fValue2) return 0;
      if (fValue1 != -999.0 && fValue2 != -999.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == -999.0) { return 1; } else { return -1; };
      }

      case Gap:
      fValue1 = node1->hypocenters.first->gap;
      fValue2 = node2->hypocenters.first->gap;
      if (fValue1 == fValue2) return 0;
      if (fValue1 != -999.0 && fValue2 != -999.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == -999.0) { return 1; } else { return -1; };
      }

      case ErrLat:
      fValue1 = node1->hypocenters.first->lat_err;
      fValue2 = node2->hypocenters.first->lat_err;
      if (fValue1 == fValue2) return 0;
      if (fValue1 != -999.0 && fValue2 != -999.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == -999.0) { return 1; } else { return -1; };
      }

      case ErrLon:
      fValue1 = node1->hypocenters.first->lon_err;
      fValue2 = node2->hypocenters.first->lon_err;
      if (fValue1 == fValue2) return 0;
      if (fValue1 != -999.0 && fValue2 != -999.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == -999.0) { return 1; } else { return -1; };
      }

      case ErrDep:
      fValue1 = node1->hypocenters.first->depth_err;
      fValue2 = node2->hypocenters.first->depth_err;
      if (fValue1 == fValue2) return 0;
      if (fValue1 != -999.0 && fValue2 != -999.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == -999.0) { return 1; } else { return -1; };
      }

      case Dist:
      // One character. A value of '32' means data missing.
      iValue1 = (int)node1->hypocenters.first->dist_id;
      iValue2 = (int)node2->hypocenters.first->dist_id;
      if (iValue1 == iValue2) return 0;
      if (iValue1 != 32 && iValue2 != 32) {
         if (iValue1 > iValue2) { return 1; } else { return -1; }
      } else {
         if (iValue1 == 32) { return 1; } else { return -1; }
      }

      case Type:
      // One character. A value of '32' means data missing.
      iValue1 = (int)node1->hypocenters.first->type;
      iValue2 = (int)node2->hypocenters.first->type;
      if (iValue1 == iValue2) return 0;
      if (iValue1 != 32 && iValue2 != 32) {
         if (iValue1 > iValue2) { return 1; } else { return -1; }
      } else {
         if (iValue1 == 32) { return 1; } else { return -1; }
      }

      case NSta:
      iValue1 = node1->hypocenters.first->nstat;
      iValue2 = node2->hypocenters.first->nstat;
      if (iValue1 == iValue2) return 0;
      if (iValue1 != -999 && iValue2 != -999) {
         if (iValue1 > iValue2) { return 1; } else { return -1; }
      } else {
         if (iValue1 == -999) { return 1; } else { return -1; };
      }

      case M:
      fValue1 = fValue2 = 0.0;
      if (node1->hypocenters.nmag) fValue1 = node1->hypocenters.mag_all[0].mag;
      if (node2->hypocenters.nmag) fValue2 = node2->hypocenters.mag_all[0].mag;
      if (fValue1 == fValue2) return 0;
      if (fValue1 != 0.0 && fValue2 != 0.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == 0.0) { return 1; } else { return -1; };
      }

      case MW:
      fValue1 = node1->hypocenters.first->magnitudes.MW.mag[0];
      fValue2 = node2->hypocenters.first->magnitudes.MW.mag[0];
      if (fValue1 == fValue2) return 0;
      if (fValue1 != 0.0 && fValue2 != 0.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == 0.0) { return 1; } else { return -1; };
      }

      case ML:
      fValue1 = node1->hypocenters.first->magnitudes.ML.mag[0];
      fValue2 = node2->hypocenters.first->magnitudes.ML.mag[0];
      if (fValue1 == fValue2) return 0;
      if (fValue1 != 0.0 && fValue2 != 0.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == 0.0) { return 1; } else { return -1; };
      }

      case MC:
      fValue1 = node1->hypocenters.first->magnitudes.MC.mag[0];
      fValue2 = node2->hypocenters.first->magnitudes.MC.mag[0];
      if (fValue1 == fValue2) return 0;
      if (fValue1 != 0.0 && fValue2 != 0.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == 0.0) { return 1; } else { return -1; };
      }

      case Mb:
      fValue1 = node1->hypocenters.first->magnitudes.Mb.mag[0];
      fValue2 = node2->hypocenters.first->magnitudes.Mb.mag[0];
      if (fValue1 == fValue2) return 0;
      if (fValue1 != 0.0 && fValue2 != 0.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == 0.0) { return 1; } else { return -1; };
      }

      case MB:
      fValue1 = node1->hypocenters.first->magnitudes.MB.mag[0];
      fValue2 = node2->hypocenters.first->magnitudes.MB.mag[0];
      if (fValue1 == fValue2) return 0;
      if (fValue1 != 0.0 && fValue2 != 0.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == 0.0) { return 1; } else { return -1; };
      }

      case Ms:
      fValue1 = node1->hypocenters.first->magnitudes.Ms.mag[0];
      fValue2 = node2->hypocenters.first->magnitudes.Ms.mag[0];
      if (fValue1 == fValue2) return 0;
      if (fValue1 != 0.0 && fValue2 != 0.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == 0.0) { return 1; } else { return -1; };
      }

      case MS:
      fValue1 = node1->hypocenters.first->magnitudes.MS.mag[0];
      fValue2 = node2->hypocenters.first->magnitudes.MS.mag[0];
      if (fValue1 == fValue2) return 0;
      if (fValue1 != 0.0 && fValue2 != 0.0) {
         if (fValue1 > fValue2) { return 1; } else { return -1; }
      } else {
         if (fValue1 == 0.0) { return 1; } else { return -1; };
      }

      case Locality:
      // A 68 character string.
      // If first string character is 0, then data is missing.
      iValue1 = (int)node1->locality[0];
      iValue2 = (int)node2->locality[0];
      if (!iValue1 && !iValue2) return 0;
      if (iValue1 && iValue2) {
         return (char)strcmp(node1->locality, node2->locality);
      } else {
         if (!iValue1) { return 1; } else { return -1; }
      }

      case SFile:
      return (char)strcmp(node1->metadata.sfile, node2->metadata.sfile);

      default:
      return 2;
   }
}




// ****************************************************************************
// The LessThan functions below compares two event nodes. They
// always returns true if node1 < node2, else they return false.
// ****************************************************************************
bool LessThan_DateTime(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, DateTime) == -1);
}

bool LessThan_Action(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, Action) == -1);
}

bool LessThan_Latitude(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, Latitude) == -1);
}

bool LessThan_Longitude(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, Longitude) == -1);
}

bool LessThan_Depth(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, Depth) == -1);
}

bool LessThan_Model(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, Model) == -1);
}

bool LessThan_Agency(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, Agency) == -1);
}

bool LessThan_RMS(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, RMS) == -1);
}

bool LessThan_Gap(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, Gap) == -1);
}

bool LessThan_ErrLat(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, ErrLat) == -1);
}

bool LessThan_ErrLon(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, ErrLon) == -1);
}

bool LessThan_ErrDep(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, ErrDep) == -1);
}

bool LessThan_Dist(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, Dist) == -1);
}

bool LessThan_Type(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, Type) == -1);
}

bool LessThan_NSta(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, NSta) == -1);
}

bool LessThan_M(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, M) == -1);
}

bool LessThan_MW(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, MW) == -1);
}

bool LessThan_ML(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, ML) == -1);
}

bool LessThan_MC(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, MC) == -1);
}

bool LessThan_Mb(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, Mb) == -1);
}

bool LessThan_MB(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, MB) == -1);
}

bool LessThan_Ms(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, Ms) == -1);
}

bool LessThan_MS(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, MS) == -1);
}

bool LessThan_Locality(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, Locality) == -1);
}

bool LessThan_SFile(event_node_ *node1, event_node_ *node2)
{
   return (NodeCmp(node1, node2, SFile) == -1);
}
