#include "fileeditordlg.h"
#include "ui_fileeditordlg.h"
#include <QMessageBox>
#include <QFileDialog>


// ************************************
// Constructor function for this class.
// ************************************
FileEditorDlg::FileEditorDlg(QWidget *parent) : QDialog(parent), ui(new Ui::FileEditorDlg)
{
    ui->setupUi(this);

    // Use monospaced font.
    QFont font("Monospace");
    font.setStyleHint(QFont::TypeWriter);
    ui->TextEdit->setFont(font);
}




// ************************************
// Destructor function for this class.
// ************************************
FileEditorDlg::~FileEditorDlg()
{
    delete ui;
}



// *******************************************************
// This function is used to set the full path to the file.
// Also the application name is set for the dialog boxes.
// This function must be called before window is shown.
// *******************************************************
bool FileEditorDlg::InitDialog(QString FilePath, QString AppName)
{
    // Save application name.
    appname = AppName;
    filepath = FilePath;

    // Set name of file in dialog box.
    string = "File: "; string.append(filepath);

    // Use elided text if filename is too long.
    QFontMetrics fm(ui->labelFile->font());
    string = fm.elidedText(string, Qt::ElideLeft, ui->labelFile->width());

    // Set label text.
    ui->labelFile->setText(string);

    // Open file.
    file.setFileName(filepath);
    if (!file.open(QIODevice::ReadOnly|QIODevice::Text)) return false;

    // Read and close file.
    buffer = file.readAll();
    file.close();

    // Copy file contents to edit box.
    ui->TextEdit->setPlainText((QString)buffer);

    // Set Close button as default button.
    ui->buttonBox->activateWindow();
    ui->buttonBox->button(QDialogButtonBox::Close)->setDefault(true);

    return true;
}




// ***********************************
// User has pressed the 'Save' button.
// ***********************************
void FileEditorDlg::accept()
{
   // Change mouse pointer to wait pointer.
   setCursor(Qt::WaitCursor);

   // Copy file contents from edit box.
   string = ui->TextEdit->toPlainText();

   // Save the file. File will be overwritten.
   file.setFileName(filepath);
   if (!file.open(QIODevice::ReadWrite|QIODevice::Text)) {
      setCursor(Qt::ArrowCursor);
      message = "Unable to open file for writing!";
      QMessageBox::critical(this, appname, message);
      return;
   }

   // Write and close file.
   buffer.clear(); buffer.append(string);
   file.resize(0); file.write(buffer); file.close();

   // Restore mouse pointer.
   setCursor(Qt::ArrowCursor);

   // Close the dialog and return 'Accepted' signal.
   QDialog::done(QDialog::Accepted);
}




// **************************************
// User has pressed the 'Close' button.
// **************************************
void FileEditorDlg::reject()
{
   // Close the dialog and return 'Rejected' signal.
   QDialog::done(QDialog::Rejected);
}




// **************************************
// User has pressed the 'Save As' button.
// **************************************
void FileEditorDlg::save_as()
{
    QStringList FileNames;

    // Bring up file selection dialog box. Return if user cancels the
    // operation. User must enter the name of a new or existing file.
    QFileDialog dialog(this);
    dialog.setFileMode(QFileDialog::AnyFile);
    message = "Please select a destination directory and a file name:";
    dialog.setWindowTitle(message);
    if (dialog.exec()) {
        // User have selected a file.
        FileNames = dialog.selectedFiles();
        // If file already exist then ask for permission to overvrite it.
        if (QFile(FileNames.at(0)).exists()) {
            QMessageBox::StandardButton Answer;
            message = "File already exists. Do you want to overwrite the file?";
            Answer = QMessageBox::question(this, appname, message, QMessageBox::Yes | QMessageBox::No,
                                           QMessageBox::No);
            if (Answer == QMessageBox::No) return;
        }
    } else {
        // Abort. User have pressed 'Cancel'.
        return;
    }

    // Change mouse pointer to wait pointer.
    setCursor(Qt::WaitCursor);

    // Save the file. File will be overwritten.
    file.setFileName(FileNames.at(0));
    if (!file.open(QIODevice::ReadWrite|QIODevice::Text))
    {
        setCursor(Qt::ArrowCursor);
        message = "Unable to open selected file for writing!";
        QMessageBox::critical(this, appname, message);
        return;
    }

    // Copy file contents from edit box.
    buffer.clear(); buffer.append(ui->TextEdit->toPlainText());

    // Write and close file.
    file.resize(0); file.write(buffer); file.close();

    // Restore mouse pointer.
    setCursor(Qt::ArrowCursor);
}




// ***********************************************
// This function is signalled from QPlainTextEdit
// when user moves the cursor to a different line.
// ***********************************************
void FileEditorDlg::CursorPositionChanged()
{
    // Cursor position has changed. Update column label.
    QTextCursor cursor = ui->TextEdit->textCursor();
    string = "Column: ";
    string.append(QString::number(cursor.columnNumber()+1));
    ui->labelColumn->setText(string);
}
