#include <cmath>
#include <stdio.h>
#include <stdlib.h>
#include <QDir>
#include <QChar>
#include <QFile>
#include <QDebug>
#include <QString>
#include <QIODevice>
#include <QByteArray>
#include <QMainWindow>
#include <QMessageBox>
#include <QStringList>
#include "fortran.h"
#include "library.h"
#include "database.h"
#include "mainwindow.h"


// *****************************************************
// Test if all requirements for running the program is
// satisfied. Displays error messages and warnings.
// Returns 'false' if application cannot be started.
// *****************************************************
bool IsPreflightCheckOK()
{
   QByteArray string;
   QString message;
   QDir directory;
   int size, index;

   // First, check for fatal conditions that will abort application statup.
   // Check that environmental variable 'SEISAN_TOP' has been set.
   string = getenv("SEISAN_TOP");
   if (string.isEmpty()) {
      message = "<b> Error: </b>";
      message.append("Environment variable SEISAN_TOP has not been set.");
      QMessageBox::critical(NULL, "Seisan Explorer", message);
      return false;
   }

   // Check that the directory pointed to by 'SEISAN_TOP' exists.
   directory.setPath(string);
   if (!directory.exists()) {
      message = "<b> Error: </b>";
      message.append("Environment variable SEISAN_TOP points to an invalid directory (");
      message.append(directory.path() + ").");
      QMessageBox::critical(NULL, "Seisan Explorer", message);
      return false;
   }

   // Next, check for non-fatal conditions that will only give warnings.
   // Check if environmental variable 'DEF_BASE' has been set.
   string = qgetenv("DEF_BASE");
   if (string.isEmpty()) {
      message = "<b> Warning: </b>";
      message.append("Environment variable DEF_BASE has not been set.");
      QMessageBox::warning(NULL, "Seisan Explorer", message);
   } else {
      // Make sure DEF_BASE is 5 chars long. Underscores will be added if needed.
      size = string.size();
      if (size < 5) for (index=size;index<5;index++) string.append("_");
      qputenv("DEF_BASE", string);
   }

   return true;
}



// *******************************************
// Check if 'FileName' is a valid S-file name.
// *******************************************
bool IsSfileNameOk(QString FileName)
{
   QChar Character;

   // Check length of filename.
   if (FileName.size() != 19) return false;

   // Check all numbers in the name.
   if (!FileName.at(0).isDigit()) return false;
   if (!FileName.at(1).isDigit()) return false;
   if (!FileName.at(3).isDigit()) return false;
   if (!FileName.at(4).isDigit()) return false;
   if (!FileName.at(5).isDigit()) return false;
   if (!FileName.at(6).isDigit()) return false;
   if (!FileName.at(8).isDigit()) return false;
   if (!FileName.at(9).isDigit()) return false;
   if (!FileName.at(13).isDigit()) return false;
   if (!FileName.at(14).isDigit()) return false;
   if (!FileName.at(15).isDigit()) return false;
   if (!FileName.at(16).isDigit()) return false;
   if (!FileName.at(17).isDigit()) return false;
   if (!FileName.at(18).isDigit()) return false;

   // Check all fixed characters in the name.
   if (FileName.at(2) != '-') return false;
   if (FileName.at(7) != '-') return false;
   if (FileName.at(11)!= '.') return false;
   if (FileName.at(12)!= 'S') return false;

   // Check the distance id in the name.
   if (FileName.at(10) == 'L') return true;
   if (FileName.at(10) == 'R') return true;
   if (FileName.at(10) == 'D') return true;

   return false;
}



// *************************************************************
// This functions checks for problems with an operator identity.
// *************************************************************
bool IsOperatorIdentityOK(QString Id)
{
   // Check identity.
   if (Id.isEmpty()) return false;
   if (Id.size() > 4) return false;
   if (Id.contains(QRegExp("\\W"))) return false;

   // Identity is OK.
   return true;
}




// *************************************************
// This functions copies a QString to a char array.
// *************************************************
void CopyQstrToChar(QString qstr, char* cstr)
{
   strcpy(cstr, qstr.toLatin1().data());
}




// *********************************************************
// This function converts a char string to a fortran string
// by filling free space in a string with spaces. The string
// terminator (\0) is also overwritten with a space.
// *********************************************************
void PadString(char *string, int strsize)
{
   for (int i=strlen(string);i<=strsize-1;i++) string[i]=32;
}




// *********************************************************
// This function extends the length of the string up to
// 'strsize' by adding a number of spaces to the string.
// *********************************************************
void PadQString(QString *string, int strsize)
{
   while (string->size() < strsize) string->append(" ");
}




// ********************************************************
// This functions converts a fortran string to a C++ string
// by adding a \0 terminator after the last used character
// in the string
// ********************************************************
void UnPadString(char *string, int strsize)
{
   // Start at end of string and find first
   // character that is not a space. Replace
   // the last space with a NULL terminator.
   bool done = false;
   int index = strsize-1;

   while (!done) {
      if (string[index] != 32 && string[index] != 0) {
         if (index == strsize-1) {
            string[index] = 0;
         } else {
            string[index+1] = 0;
         }
         done = true;
      }
      index --;
      // Check if we have processed the whole
      // string and the string is empty.
      if (index == 0 && !done) {
         string[index] = 0;
         done = true;
      }
   }
}




// *********************************************************
// This function removes all spaces at the end of a QString.
// *********************************************************
void UnPadQString(QString *string)
{
   int StrLength;
   while (true) {
      StrLength = string->size();
      if (!StrLength) return;
      if (string->at(StrLength-1) == QChar(32)) {
         string->remove(StrLength-1, 1);
      } else {
         return;
      }
   }
}




// **********************************************************
// Deletes a charater from a string.
// Delets the character pointed to by the zero based index
// pointer 'strptr'. Characters to the right of the deleted
// character are shifted left.
// **********************************************************
void StrDelete(char *string, int strptr)
{
   int i, strlngth;

   // Get the length of the string.
   strlngth=(int)strlen(string);

   // Return if string is empty.
   if (!strlngth) return;

   // Return if strptr is larger then length of string.
   if (strptr>=strlngth) return;

   // Delete character pointed to by strptr.
   for (i=strptr;i<=strlngth-1;i++) string[i]=string[i+1];

   return;
}




// ************************************************
// Removes all spaces in the start of a string.
// ************************************************
void StrRmvTrailingSpaces(char *string)
{
   while (string[0]==' ') StrDelete(string, 0);
}



// ************************************************
// Removes all double, trple etc spaces from string
// leaving only a single space.
// ************************************************
void TrimSpaces(QString *String)
{
   while (String->contains("  ")) *String = String->replace("  ", " ");
}




// ************************************************
// Returns true is string only has spaces.
// ************************************************
bool IsQStringAllSpace(QString String)
{
   // Get the length of the string.
   int Length = String.size();

   // Return true if string is empty.
   if (!Length) return true;

   // Search for characters other than space in string.
   for (int Index=0;Index<Length;Index++) {
      if(String.at(Index) != QChar(32)) return false;
   }

   // Only space characters was found.
   return true;
}




// ********************************************************
// This function copies a source string into a dest string,
// and then converts the dest string to a fortran string.
// ********************************************************
void fstrcpy(char dest[], char *source, int strsize_dest)
{
   int length = strlen(source);
   memset((void*) dest, 0, (size_t) strsize_dest);
   if (length > strsize_dest) length = strsize_dest;
   if (length) strncpy(dest, source, length);
   if (length < strsize_dest) PadString(dest, strsize_dest);
}




// ********************************************************
// This function takes a valid path and removes the last
// directory from the path. Function does nothing if path
// is already at root level.
// ********************************************************
void PathCdUp(QString *Path)
{
   QString NewPath;

   // Return if path is empty.
   if (!Path->size()) return;

   // Make sure the path is clean.
   *Path = QDir::cleanPath(*Path);

   // Return if path is already at root.
   if (Path->size() == 1 && Path->at(0) == QChar(47)) return;
   if (Path->size() == 1 && Path->at(0) == QChar(92)) return;
   if (Path->size() == 3 && Path->contains(":\\")) return;

   // Remove any slash at end of path.
   if (Path->endsWith(QChar(47))) Path->remove(Path->size()-1, 1);
   if (Path->endsWith(QChar(92))) Path->remove(Path->size()-1, 1);

   // Find the last slash in path.
   bool found = false; int index;
   while (!found) {
      index = Path->size() - 1;
      if (Path->at(index) == QChar(47) || Path->at(index) == QChar(92)) found = true;
      Path->remove(index, 1);
   }

   // If path is empty then add a slash.
   if (Path->size() == 0) Path->append(47);
   if (Path->size() == 2 && Path->at(1) == QChar(':')) Path->append(47);
}




// ********************************************************
// This function creates a time string suitable for
// use with the event list view. Invalid time/date values
// are replaced with question marks in the string.
// ********************************************************
void CreateElvTimeStr(db_time *time, QString *string)
{
   char str10[10];

   // Add year to string.
   sprintf(str10, "%04u-", time->year);
   if (time->year <= 0) strcpy(str10, "????-");
   string->append(str10);

   // Add month to string.
   sprintf(str10, "%02u-", time->month);
   if (time->month <= 0) strcpy(str10, "??-");
   string->append(str10);

   // Add day to string.
   sprintf(str10, "%02u ", time->day);
   if (time->day <= 0) strcpy(str10, "?? ");
   string->append(str10);

   // Add hour to string.
   sprintf(str10, "%02u:", time->hour);
   if (time->hour < 0) strcpy(str10, "??:");
   string->append(str10);

   // Add minute to string.
   sprintf(str10, "%02u:", time->minute);
   if (time->minute < 0) strcpy(str10, "??:");
   string->append(str10);

   // Add second to string.
   sprintf(str10, "%05.2f", time->second);
   if (time->second < 0.0) strcpy(str10, "?????");
   string->append(str10);
}




// *******************************************************************************
// Set program's current directory.
// *******************************************************************************
void SetCurrentDirectory(app_config *Config)
{
   bool DirSet = false;
   if (Config->UseAltWorkDir) {
      DirSet = QDir::setCurrent(Config->AltWorkDir);
      if (!DirSet) Config->UseAltWorkDir = false;
   }
   if (!DirSet) DirSet = QDir::setCurrent(Config->DefWorkDir);
   if (!DirSet) QDir::setCurrent(Config->AppLaunchDir);
}




// ************************************
// Get a file's modification date/time.
// ************************************
void GetFileModTime(QString FileName, QDateTime *ModTime)
{
    QFileInfo File(FileName);
    *ModTime = File.lastModified();
}




// ****************************************
// Extracts the file name from a full path.
// ****************************************
QString GetNameFromFilePath(QString Path)
{
   QFileInfo file(Path);
   return file.fileName();
}




// *****************************************************************
// This function is used clear all information in the common blocks.
// *****************************************************************
void ClearCommonBlocks()
{
   // Initialize single int/float single values.
   rea4_.rea_nwav = 0;
   rea4_.rea_nmag = 0;
   rea4_.rea_nhyp = 0;
   rea4_.rea_nstat = 0;
   rea4_.rea_nspec = 0;
   rea4_.rea_nhead = 0;
   rea4_.rea_nmacro = 0;
   rea4_.rea_nfault = 0;
   rea4_.rea_nphase = 0;
   rea4_.rea_nrecord = 0;
   rea4_.rea_ncomment = 0;
   rea4_.rea_av_moment = -999.0;
   rea4_.rea_av_sdrop = -999.0;
   rea4_.rea_av_omega0 = -999.0;
   rea4_.rea_av_cornerf = -999.0;
   rea4_.rea_av_radius = -999.0;
   rea4_.rea_av_swin = -999.0;
   rea4_.rea_av_mw = -999.0;
   rea4_.rea_av_slope = -999.0;
   rea4_.rea_err_unit = 0;
   rea4_.rea_read_err = 0;
   rea4_.rea_write_err = 0;

   // Fill character variables zeros.
   memset(rea1_.rea_onset, 0, MAX_DATA/2);
   memset(rea1_.rea_polarity, 0, MAX_DATA/2);
   memset(rea1_.rea_weight_in, 0, MAX_DATA/2);
   memset(rea2_.rea_co, 0, MAX_DATA/2*2);
   memset(rea2_.rea_weight_out, 0, MAX_DATA/2*2);
   memset(rea3_.rea_action, 0, 3);
   memset(rea4_.rea_locality, 0, 68);
   memset(rea4_.rea_id_line, 0, 80);
   memset(rea4_.rea_macro, 0, 100*80);
   memset(rea4_.rea_macro, 0, 100*80);
   memset(rea4_.rea_wav, 0, 100*80);
   memset(rea4_.rea_fault, 0, 100*80);
   memset(rea4_.rea_comment, 0, 100*80);
   memset(rea4_.rea_comment, 0, 100*80);
   memset(rea4_.rea_comp, 0, MAX_DATA/2*4);
   memset(rea4_.rea_auto, 0, MAX_DATA/2*20);
   memset(rea5_.rea_stat, 0, MAX_DATA/2*5);
   memset(rea8_.rea_phase, 0, MAX_DATA/2*8);
   memset(hyp1_.hyp_mag_type_all, 0, 200);
   memset(hyp1_.hyp_model, 0, 100);
   memset(hyp1_.hyp_dist_id, 0, 100);
   memset(hyp1_.hyp_type, 0, 100);
   memset(hyp1_.hyp_fix_org, 0, 100);
   memset(hyp1_.hyp_depth_flag, 0, 100);
   memset(hyp1_.hyp_epi_flag, 0, 100);
   memset(hyp1_.hyp_mag_type, 0, 100*6);
   memset(hyp4_.hyp_auto, 0, 100*20);
   memset(hyp5_.hyp_mag_agency_all, 0, 200*5);
   memset(hyp5_.hyp_agency, 0, 100*5);
   memset(hyp5_.hyp_mag_agency, 0, 100*6*5);

   // Initialize int/float arrays.
   for (int index=0;index<MAX_DATA/2;index++) {
      rea4_.rea_year[index] = -999;
      rea4_.rea_month[index] = -999;
      rea4_.rea_day[index] = -999;
      rea4_.rea_hour[index] = -999;
      rea4_.rea_min[index] = -999;
      rea4_.rea_sec[index] = -999.0;
      rea4_.rea_coda[index] = -999.0;
      rea4_.rea_amp[index] = -999.0;
      rea4_.rea_per[index] = -999.0;
      rea4_.rea_baz_obs[index] = -999.0;
      rea4_.rea_baz_cal[index] = -999.0;
      rea4_.rea_vel[index] = -999.0;
      rea4_.rea_ain[index] = -999.0;
      rea4_.rea_baz_res[index] = -999.0;
      rea4_.rea_dist[index] = -999.0;
      rea4_.rea_az[index] = -999.0;
      rea4_.rea_res[index] = -999.0;
      rea4_.rea_moment[index] = -999.0;
      rea4_.rea_sdrop[index] = -999.0;
      rea4_.rea_omega0[index] = -999.0;
      rea4_.rea_cornerf[index] = -999.0;
      rea4_.rea_radius[index] = -999.0;
      rea4_.rea_swin[index] = -999.0;
      rea4_.rea_vs[index] = -999.0;
      rea4_.rea_vp[index] = -999.0;
      rea4_.rea_q0[index] = -999.0;
      rea4_.rea_qalpha[index] = -999.0;
      rea4_.rea_kappa[index] = -999.0;
      rea4_.rea_density[index] = -999.0;
      rea4_.rea_slope[index] = -999.0;
      rea4_.rea_mc[index] = -999.0;
      rea4_.rea_ml[index] = -999.0;
      rea4_.rea_mb[index] = -999.0;
      rea4_.rea_ms[index] = -999.0;
      rea4_.rea_mw[index] = -999.0;
      rea4_.rea_geo_dist[index] = -999.0;
      rea8_.rea_abs_time[index] = -999.0;
   }
   for (int index=0;index<100;index++) {
      hyp1_.hyp_high_accuracy[index] = 0;
      hyp1_.hyp_error[index] = 0;
      hyp4_.hyp_year[index] = -999;
      hyp4_.hyp_month[index] = -999;
      hyp4_.hyp_day[index] = -999;
      hyp4_.hyp_hour[index] = -999;
      hyp4_.hyp_min[index] = -999;
      hyp4_.hyp_nstat[index] = -999;
      hyp4_.hyp_sec[index] = -999.0;
      hyp4_.hyp_lat[index] = -999.0;
      hyp4_.hyp_lon[index] = -999.0;
      hyp4_.hyp_depth[index] = -999.0;
      hyp4_.hyp_rms[index] = -999.0;
      hyp4_.hyp_gap[index] = -999.0;
      hyp4_.hyp_sec_err[index] = -999.0;
      hyp4_.hyp_lat_err[index] = -999.0;
      hyp4_.hyp_lon_err[index] = -999.0;
      hyp4_.hyp_depth_err[index] = -999.0;
      hyp4_.hyp_cov[index][0] = -9.9e10;
      hyp4_.hyp_cov[index][1] = -9.9e10;
      hyp4_.hyp_cov[index][2] = -9.9e10;
      hyp4_.hyp_mag[index][0] = -999.0;
      hyp4_.hyp_mag[index][1] = -999.0;
      hyp4_.hyp_mag[index][2] = -999.0;
      hyp4_.hyp_mag[index][3] = -999.0;
      hyp4_.hyp_mag[index][4] = -999.0;
      hyp4_.hyp_mag[index][5] = -999.0;
   }
   for (int index=0;index<200;index++) {
      hyp4_.hyp_mag_all[index] = -999.0;
   }
}




// *****************************************************************
// This function merges a number of files into a new file.
// The new file is placed in the current working directory.
// ******************************************************************
bool MergeFiles(QStringList FileList, QString OutFile)
{
   int NumFiles;
   QByteArray text;
   QFile infile, outfile;

   // Get number of files in file list Return if list is empty.
   if (!(NumFiles = FileList.size())) return false;

   // Open out file. Existing file will be overwritten. Abort if 'open' failes.
   outfile.setFileName(OutFile);
   if (!outfile.open(QIODevice::WriteOnly | QIODevice::Truncate | QIODevice::Text)) return false;

   // Copy all files in file list to out file.
   bool error = false;
   for (int index=0;index<NumFiles;index++) {
      infile.setFileName(FileList.at(index));
      if (infile.open(QIODevice::ReadOnly | QIODevice::Text)) {
         text = infile.readAll(); infile.close();
         if (outfile.write(text) == -1) error = true;
      } else {
         // The file either do not exist, or
         // it cannot be openened for reading.
         error = true;
      }
   }

   // Close the out file.
   outfile.close();

   // Return true if fuction succeded, or false otherwise.
   return !error;
}




// *****************************************************************
// This function checks if a line is empty. An empty line
// is here defined as a being a line with only spaces.
// ******************************************************************
bool IsDataLineEmpty(char Line[], qint64 LineLength)
{
   for (int i=0;i<LineLength-1;i++) if (Line[i]!=32) return false;
   return true;
}



// *****************************************************************
// This function checks if a file is in the path.
// ******************************************************************
bool IsFileInPath(QString file)
{
   QStringList PathList;
   QString Path, PathToFile;

   // Get the current path.
   Path = getenv("PATH");
   if (!Path.size()) return false;

   // Create a list of directories.
   #ifdef Q_OS_WIN32
   QChar SplitChar = ';';
   #endif
   #ifdef Q_OS_LINUX
   QChar SplitChar = ':';
   #endif
   PathList = Path.split(SplitChar, QString::SkipEmptyParts);

   // Enumerate directories to search for file.
   for (int i=0;i<PathList.size();i++) {
      PathToFile = PathList.at(i);
      PathToFile.append("/");
      PathToFile.append(file);
      PathToFile = QDir::cleanPath(PathToFile);
      PathToFile = QDir::toNativeSeparators(PathToFile);
      if (QFile::exists(PathToFile)) return true;
   }

   return false;
}





// ***************************************************************
// Calculates a Julian Day number.
// ***************************************************************
ulong JulianDay(db_time *Time)
{
   long a,b,c,e,f;
   ushort Year, Month;

   Year=Time->year;
   Month=Time->month;
   if (Month<=2) { Year=Year-1; Month=Month+12; }
   a=Year/100; b=a/4; c=2-a+b;
   e=(long)(double)(365.25*(Year+4716));
   f=(long)(double)(30.6001*(Month+1));
   return (c+Time->day+e+f-1525);
}




// *************************************************************
// Compares two time structures. The return value for this
// function indicates the relation between Time1 and Time2:
// -1 = Time1 is earlier (older) than Time2
//  0 = Time1 is identical to Time2
//  1 = Time1 is later than Time2
// *************************************************************
char TimeCmp(db_time *Time1, db_time *Time2)
{
   if (Time1->year<Time2->year) return -1;
   if (Time1->year>Time2->year) return 1;
   if (Time1->month<Time2->month) return -1;
   if (Time1->month>Time2->month) return 1;
   if (Time1->day<Time2->day) return -1;
   if (Time1->day>Time2->day) return 1;
   if (Time1->hour<Time2->hour) return -1;
   if (Time1->hour>Time2->hour) return 1;
   if (Time1->minute<Time2->minute) return -1;
   if (Time1->minute>Time2->minute) return 1;
   if (Time1->second<Time2->second) return -1;
   if (Time1->second>Time2->second) return 1;

   return 0;
}




// ****************************************************
// Calculates the timedifference in seconds between
// two time structures and return the absolute result.
// ****************************************************
double TimeDiff(db_time *Time1, db_time *Time2)
{
   double Diff, DaySeconds1, DaySeconds2;
   char CmpResult;
   long WholeDayDiff;

   // Compare the two times.
   CmpResult=TimeCmp(Time1, Time2);

   // Check if the times are equal.
   if (!CmpResult) return 0.0;

   // Calculate number of whole days elased between the two times.
   WholeDayDiff = (long)(JulianDay(Time1)-JulianDay(Time2));

   // Calculate number of seconds elapsed since the start of the day.
   DaySeconds1=Time1->hour*3600+Time1->minute*60+Time1->second;
   DaySeconds2=Time2->hour*3600+Time2->minute*60+Time2->second;

   // Calculate the total time-difference in seconds.
   Diff=fabs(WholeDayDiff*86400.0);
   if (CmpResult==-1 && DaySeconds1< DaySeconds2) Diff=Diff+DaySeconds2-DaySeconds1;
   if (CmpResult==-1 && DaySeconds1>=DaySeconds2) Diff=Diff+DaySeconds2-DaySeconds1;
   if (CmpResult== 1 && DaySeconds1< DaySeconds2) Diff=Diff+DaySeconds1-DaySeconds2;
   if (CmpResult== 1 && DaySeconds1>=DaySeconds2) Diff=Diff+DaySeconds1-DaySeconds2;

   return Diff;
}




// *********************************************************
// Read a line from a text file. Line number is needed.
// Returns a status value:
// 0=OK, 1=file not found, 2=open failed, 3=line not found.
// *********************************************************
int ReadFileLine(QString FileName, int LineNr, QString *Line)
{
   QFile file;
   QString line;

   // Check that the file exists.
   if (!QFile::exists(FileName)) return 1;

   // Open file.
   file.setFileName(FileName);
   if (!file.open(QIODevice::ReadOnly|QIODevice::Text)) return 2;

   // File was opened successfully. Read lines until
   // the correct line are found or we reach end of file.
   int linecount = 0;
   QTextStream ts(&file);
   while (!ts.atEnd()) {
      linecount++;
      line = ts.readLine();
      if (linecount == LineNr) {
         line.replace(QChar(0), QChar(32));
         *Line = line;
         file.close();
         return 0;
      }
   }

   // The line was not found.
   file.close();
   return 3;
}




// *********************************************************
// Read a text file into a QStringList.
// Returns a status value:
// 0=OK, 1=file not found, 2=open failed.
// *********************************************************
int ReadTextFile(QString FileName, QStringList *Lines)
{
   // Check that the file exists.
   if (!QFile::exists(FileName)) return 1;

   // Open file.
   QFile file(FileName);
   if (!file.open(QFile::ReadOnly|QFile::Text)) return 2;

   // File was opened successfully. Read all lines.
   QTextStream in(&file);
   while (!in.atEnd()) Lines->append(in.readLine());

   // Close file.
   file.close();
   return 0;
}





// *****************************************************
// Writes a number of lines to a text file.
// Returns false if something goes wrong.
// *****************************************************
bool WriteTextFile(QString FileName, QStringList Lines)
{
   // Open file.
   QFile file(FileName);
   if (!file.open(QFile::WriteOnly|QFile::Text)) return false;

   // Write to file.
   QTextStream out(&file);
   for (int i=0;i<Lines.size();i++) out << Lines.at(i) << "\n";

   // Close file.
   file.close();
   return true;
}




// ***************************************************************
// This function creates a lock file in current working directory.
// ***************************************************************
void CreateLockFile()
{
   QFile LockFile(WORKDIR_LOCK_FILE);
   if (!LockFile.exists()) {
      LockFile.open(QIODevice::ReadWrite);
      LockFile.close();
   }
}
