#include <QProcess>
#include <QWebView>
#include <QCloseEvent>
#include <QFileDialog>
#include <QInputDialog>

#include "mainwindow.h"
#include "ui_mainwindow.h"

#include "google.h"
#include "fortran.h"
#include "library.h"
#include "aboutdlg.h"
#include "logmodel.h"
#include "seisandb.h"
#include "mergedlg.h"
#include "gotorowdlg.h"
#include "poissondlg.h"
#include "weichertdlg.h"
#include "magvsmagdlg.h"
#include "fileeditordlg.h"
#include "gettextlinesdlg.h"
#include "eventsperyeardlg.h"
#include "sample_function_a.h"
#include "sample_function_b.h"
#include "timeofdaydistrdlg.h"
#include "gutenbergrichterdlg.h"
#include "completenesscheckdlg.h"
#include "seteventindicatordlg.h"
#include "setmodelindicatordlg.h"
#include "driftofhypocentersdlg.h"


// *****************************************
// Constructor for this class.
// *****************************************
MainWindow::MainWindow(QWidget *parent) : QMainWindow(parent), ui(new Ui::MainWindow)
{
   // Register user defined metatypes.
   qRegisterMetaType<db::dbStatus>("db::dbStatus");
   qRegisterMetaType<db::dbError>("db::dbError");
   qRegisterMetaType<logmodel::msg_type>("logmodel::msg_type");

   // Set program name, version, and build date.
   AppName = "Seisan Explorer";
   AppVersion = "2.2.2";
   AppBuildDate = __DATE__;

   // Store organisation and application name.
   // This information is used by QSession.
   QCoreApplication::setOrganizationName("UiB");
   QCoreApplication::setApplicationName(AppName);
   QCoreApplication::setApplicationVersion(AppVersion);

   // Set up the user interface.
   ui->setupUi(this);
   ui->EventListView->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);

   // Modify colors for the event list view. When view is inactive, selected
   // rows should use the same colors as with an active view. This is done to
   // make 'Speed Search' operation in the view more visible for the operator.
   Palette = ui->EventListView->palette();
   Color = Palette.color(QPalette::Active, QPalette::Highlight);
   Palette.setColor(QPalette::Inactive, QPalette::Highlight, Color);
   Color = Palette.color(QPalette::Active, QPalette::HighlightedText);
   Palette.setColor(QPalette::Inactive, QPalette::HighlightedText, Color);
   ui->EventListView->setPalette(Palette);

   // Add labels to the status bar
   statusBar()->addPermanentWidget(&LblInterval);  // Used to show current time interval.
   statusBar()->addPermanentWidget(&LblOperator);  // Used to show operator's identity.

   // Create connection from the tab widget to the mainwindow.
   connect(ui->tabWidget, SIGNAL(currentChanged(int)),
                  this, SLOT(CurrentTabChanged(int)));

   // Create connection from the operator class to the mainwindow.
   connect(&Operator, SIGNAL(SetStatusBarOperator()),
                  this, SLOT(SetStatusBarOperator()));

   // Create connections from the database to the mainwindow.
   connect(&DB, SIGNAL(DbOpenProgress(db::dbStatus, int)),
           this, SLOT(DbOpenProgress(db::dbStatus, int)));
   connect(&DB, SIGNAL(DbOpenResult(db::dbStatus, db::dbError)),
           this, SLOT(DbOpenResult(db::dbStatus, db::dbError)));
   connect(&DB, SIGNAL(SetWindowTitle()), this, SLOT(SetWindowTitle()));
   connect(&DB, SIGNAL(DbFilterProgress(char)), this, SLOT(DbFilterProgress(char)));

   // Create and set a model for the log view.
   LogModel = new logmodel(this); ui->LogView->setModel(LogModel);

   // Give log view model access to the log view.
   LogModel->SetView(ui->LogView);

   // Create connection from the mainwindow to the log view model.
   connect(this, SIGNAL(WriteLog(logmodel::msg_type, QString)),
           LogModel, SLOT(WriteLog(logmodel::msg_type, QString)));

   // Create connection from the operator class to the log view model.
   connect(&Operator, SIGNAL(WriteLog(logmodel::msg_type, QString)),
           LogModel, SLOT(WriteLog(logmodel::msg_type, QString)));

   // Create connections from the database to the log view model.
   connect(&DB, SIGNAL(WriteLog(logmodel::msg_type, QString)),
           LogModel, SLOT(WriteLog(logmodel::msg_type, QString)));
   connect(&DB.ReadSfiles, SIGNAL(WriteLog(logmodel::msg_type, QString, QString)),
           LogModel, SLOT(WriteLog(logmodel::msg_type, QString, QString)));

   // Load program configurations into 'Config' structure.
   LoadProgramSettings();

   // Enable the context menu for the log view.
   ui->LogView->setContextMenuPolicy(Qt::CustomContextMenu);
   connect(ui->LogView, SIGNAL(customContextMenuRequested(const QPoint&)),
           this, SLOT(CCMR_LogView(const QPoint&)));

   // Create and set a model for the event list view.
   ELVModel = new eventlistmodel(&DB); ui->EventListView->setModel(ELVModel);

   // Give event list model access to the event list view.
   ELVModel->SetView(ui->EventListView);

   // Create connection from the database to the event list model.
   connect(&DB, SIGNAL(SpeedSearchReply(int)), ELVModel, SLOT(SpeedSearchReply(int)));

   // Show/hide user selectable columns and rezise columns to fit headers.
   SetEventListViewHeaders(); ui->EventListView->resizeColumnsToContents();

   // Save pointer to event list view's header object.
   ELVHeader = ui->EventListView->horizontalHeader();

   // Set sort indicator on Date/Time column.
   ELVHeader->setSortIndicator(ColDateTime, Qt::AscendingOrder);

   // Create connection from event list view's header to mainwindow.
   connect(ELVHeader, SIGNAL(sortIndicatorChanged(int, Qt::SortOrder)),
           this, SLOT(ELV_SortIndicatorChanged(int, Qt::SortOrder)));

   // Enable the context menu for the event list view.
   ui->EventListView->setContextMenuPolicy(Qt::CustomContextMenu);
   connect(ui->EventListView, SIGNAL(customContextMenuRequested(const QPoint&)),
           this, SLOT(CCMR_EventListView(const QPoint&)));

   // Create and install event filter for the event list view.
   ELVFilter = new eventlistviewfilter(this); ui->EventListView->installEventFilter(ELVFilter);

   // Create connection from event list view filter to event list model.
   connect(ELVFilter, SIGNAL(ActivateSpeedSearch(char)), ELVModel, SLOT(ActivateSpeedSearch(char)));
   connect(ELVFilter, SIGNAL(ScrollToNextSelection()), ELVModel, SLOT(ScrollToNextSelection()));
   connect(ELVFilter, SIGNAL(ScrollToPrevSelection()), ELVModel, SLOT(ScrollToPrevSelection()));

   // Create connections from the event list view filter to mainwindow.
   connect(ELVFilter, SIGNAL(ELV_Duplicate_Event()), this, SLOT(ELV_Duplicate_Event()));
   connect(ELVFilter, SIGNAL(ELV_Edit_With_Text_Editor()), this, SLOT(ELV_Edit_With_Text_Editor()));
   connect(ELVFilter, SIGNAL(ELV_Edit_StationHyp()), this, SLOT(ELV_Edit_StationHyp()));
   connect(ELVFilter, SIGNAL(ELV_Edit_Comment_Lines()), this, SLOT(ELV_Edit_Comment_Lines()));
   connect(ELVFilter, SIGNAL(ELV_Locate_Event()), this, SLOT(ELV_Locate_Event()));
   connect(ELVFilter, SIGNAL(ELV_Merge_Events()), this, SLOT(ELV_Merge_Events()));
   connect(ELVFilter, SIGNAL(ELV_Copy_Events()), this, SLOT(ELV_Copy_Events()));
   connect(ELVFilter, SIGNAL(ELV_Plot_With_Mulplot()), this, SLOT(ELV_Plot_With_Mulplot()));
   connect(ELVFilter, SIGNAL(ELV_Plot_With_Mulplot_Menu()), this, SLOT(ELV_Plot_With_Mulplot_Menu()));
   connect(ELVFilter, SIGNAL(SetTimeInterval()), this, SLOT(SetTimeInterval()));
   connect(ELVFilter, SIGNAL(ELV_Set_Distance_Indicator()), this, SLOT(ELV_Set_Distance_Indicator()));
   connect(ELVFilter, SIGNAL(ELV_Associate_Events()), this, SLOT(ELV_Associate_Events()));
   connect(ELVFilter, SIGNAL(ELV_Delete_Event()), this, SLOT(ELV_Delete_Event()));
   connect(ELVFilter, SIGNAL(ELV_Refresh_View()), this, SLOT(ELV_Refresh_View()));
   connect(ELVFilter, SIGNAL(ELV_Select_All()), this, SLOT(ELV_Select_All()));
   connect(ELVFilter, SIGNAL(ELV_Launch_EEV()), this, SLOT(ELV_Launch_EEV()));
   connect(ELVFilter, SIGNAL(ELV_Show_Events_With_GE()), this, SLOT(ELV_Show_Events_With_GE()));
   connect(ELVFilter, SIGNAL(ELV_Show_Events_With_GM()), this, SLOT(ELV_Show_Events_With_GM()));
   connect(ELVFilter, SIGNAL(ELV_Show_Events_With_SV()), this, SLOT(ELV_Show_Events_With_SV()));
   connect(ELVFilter, SIGNAL(ELV_GoToNextNewEvent()), this, SLOT(ELV_GoToNextNewEvent()));
   connect(ELVFilter, SIGNAL(ELV_GoToRowNr()), this, SLOT(ELV_GoToRowNr()));
   connect(ELVFilter, SIGNAL(ELV_Register_Event()), this, SLOT(ELV_Register_Event()));
   connect(ELVFilter, SIGNAL(ELV_Load_Event()), this, SLOT(ELV_Load_Event()));
   connect(ELVFilter, SIGNAL(ELV_Unload_Event()), this, SLOT(ELV_Unload_Event()));
   connect(ELVFilter, SIGNAL(SetFilter()), this, SLOT(SetFilter()));

   // Create and install event filter for the log view.
   LVFilter = new logviewfilter(this); ui->LogView->installEventFilter(LVFilter);

   // Create connection from the log view filter to mainwindow.
   connect(LVFilter, SIGNAL(SetTimeInterval()), this, SLOT(SetTimeInterval()));

   // Initialize some variables.
   AccTimeInt = 200;
   OpenLastUsed = false;
   NoCloseQuery = false;
   NoTimeIntQuery = false;

   // Read environment variables 'SEISAN_TOP' and 'DEF_BASE'.
   SEISAN_TOP = getenv("SEISAN_TOP"); DEF_BASE = getenv("DEF_BASE");

   // Check and handle issues with 'DEF_BASE'.
   if (DEF_BASE.isEmpty()) {
      // Disable the 'File|Open|Default Database' menu item.
      ui->actionDefaultDatabase->setDisabled(true);
      // Inform user that DEF_BASE is not properly set.
      Message = "Variable 'DEF_BASE' is not defined. ";
      Message.append("Program will continue with reduced functionality.");
      emit WriteLog(logmodel::warning, Message);
   }

   // No database or index file has been opened yet.
   // Set application status to "database closed".
   AppStatus = db_closed;

   // Set and show default time interval.
   Config.StartDate = QDate::currentDate().addDays(-Config.TimeInterval);
   Config.EndDate = QDate::currentDate();
   Config.EndDateType = 1; SetStatusBarInterval();

   // Save application's current directory. This is the
   // directory from where the application is launched.
   Config.AppLaunchDir = QDir::currentPath();

   // Create and save path to the default work directory.
   Config.DefWorkDir = QDir::toNativeSeparators(SEISAN_TOP+"/WOR");

   // Create and save path to the Seisan data directory.
   Config.DatDir = SEISAN_TOP+"/DAT";
   Config.DatDir = QDir::toNativeSeparators(Config.DatDir);

   // Set the current directory. This will either be
   // a user spesified folder, or the default folder.
   SetCurrentDirectory(&Config);

   // Open a console file that functions in the seisan library can
   // write to as replacement for the missing console in Explorer.
   fstrcpy(ConsFile, (char*)CONSOLE_FILE, sizeof(ConsFile));
   openfile_unit_six_(ConsFile, sizeof(ConsFile));

   // Google Earth support.
   // Create a KML file named 'Seisan Explorer.kml'.
   // File is created in the current working directory.
   CreatePriKmlFile(Config);

   // Create the map view dialog window.
   // The window will not be shown yet.
   MapView = new MapViewDlg(&Config);

   // Activate the Log View.
   ui->tabWidget->setCurrentIndex(1); CurrentTabChanged(1);

   // Write startup messages to the program log.
   Message = "Seisan Explorer version ";
   Message.append(AppVersion).append(" started. ");
   if (Operator.IsSet()) Message.append("Operator is '" + Operator.GetCurrentId() + "'.");
   emit WriteLog(logmodel::info, Message);
   Message = "Seisan home directory (SEISAN_TOP) is '" + SEISAN_TOP + "'.";
   emit WriteLog(logmodel::info, Message);
   Message = "Current work directory is '";
   Message.append(QDir::toNativeSeparators(QDir::currentPath())+ "'. ");
   emit WriteLog(logmodel::info, Message);

   // Check if another copy of Explorer is using the work directory.
   if (!QFile::exists(WORKDIR_LOCK_FILE)) {
      // Create the lock file.
      CreateLockFile(); LockFileCreated = true;
   } else {
      // The current work directory may be in use. Warn the user.
      // User will get the opportunity to remove the lock file at
      // program exit.
      QMessageBox msgBox;
      msgBox.setIcon(QMessageBox::Warning);
      msgBox.setWindowTitle("Lock file found.");
      msgBox.setTextFormat(Qt::RichText);
      Message = "<b> Warning:</b> The lock file '";
      Message.append(WORKDIR_LOCK_FILE);
      Message.append("' exists in work directory.<br>");
      Message.append("Directory may already be in use by another Seisan Explorer user.");
      msgBox.setText(Message);
      msgBox.addButton(" Continue ", QMessageBox::AcceptRole);
      msgBox.addButton(" Remove lock file ", QMessageBox::RejectRole);
      LockFileCreated = (msgBox.exec() == QMessageBox::RejectRole);
   }

   // Load last used database if requested.
   if (Config.AutoLoad && Config.LastUsedDatabaseType != no_database) Open_LastUsedDatabase();
}




// *****************************************
// Destructor for this class.
// *****************************************
MainWindow::~MainWindow()
{
   // Save program settings to registry or disk.
   SaveProgramSettings();

   // Delete mainwindow class.
   delete ui;
}




// *******************************************************
// Opens and read data from the default database folder.
// A database folder contains year/month subfolders.
// *******************************************************
void MainWindow::Open_DefaultDatabase()
{
   // Check if a database is already open.
   if (AppStatus == db_open) {
      Message = "<b>A database is already open.</b>";
      Message.append("<br>Please close current database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Check if a database is loading.
   if (AppStatus == db_loading) {
      Message = "<b>A database is currently loading.</b>";
      Message.append("<br>To stop loading, please close database.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Check that the database folder exist.
   QString path = SEISAN_TOP + "/REA/" + DEF_BASE;
   if (!QDir(path).exists()) {
      Message = "<b>Database defined in environment variable 'DEF_BASE' do not exist.</b>";
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Save database path and database name.
   DatabasePath = QDir::toNativeSeparators(path);
   DatabaseName = DEF_BASE;

   // Request time interval from user.
   if (!NoTimeIntQuery && !SetTimeInterval()) return;

   // Set application status and type of DB we are trying to open.
   AppStatus = db_loading; CurrentDbType = default_database;

   // Write message to program log.
   Message = "Database load started for default database (" + DEF_BASE + ").";
   emit WriteLog(logmodel::info, Message);

   // Set sort indicator on DateTime column.
   ELVHeader->setSortIndicator(ColDateTime, Qt::AscendingOrder);

   // Load database into 'DB' class. This operation may take some time.
   DB.Open(default_database, Config.StartDate, Config.EndDate, DatabasePath, DatabaseName);
}




// ******************************************************
// Opens and read data from a normal database folder.
// Database folder must contain year/month subfolders.
// ******************************************************
void MainWindow::Open_Database()
{
   QString WorkDir;

   // Check if a database is already open.
   if (AppStatus == db_open) {
      Message = "<b>A database is already open.</b>";
      Message.append("<br>Please close current database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Check if a database is loading.
   if (AppStatus == db_loading) {
      Message = "<b>A database is currently loading.</b>";
      Message.append("<br>To stop loading, please close database.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   if (!OpenLastUsed) {
      // Set working directory for the folder selection dialg box.
      if (Config.LastUsedDatabaseType == normal_database && QDir(Config.LastUsedDatabasePath).exists()) {
         WorkDir = Config.LastUsedDatabasePath;
      } else {
         WorkDir = SEISAN_TOP + "/REA/";
         if (!QDir(WorkDir).exists()) WorkDir = QDir::currentPath();
      }

      // Bring up folder selection dialog box. Return if user cancels the
      // operation. User must point to a folder with year/month subfolders.
      Message = "Please select a folder containing a Seisan database:";
      DatabasePath = QFileDialog::getExistingDirectory(this, Message, WorkDir, QFileDialog::ShowDirsOnly);

      // Return if user did not select a folder.
      if (DatabasePath.isNull()) return;

      // Use native separators in the path.
      DatabasePath = QDir::toNativeSeparators(DatabasePath);

      // Extract name of database from database folderpath.
      QDir DbPath(DatabasePath);
      DatabaseName = DbPath.dirName();
   } else {
      // Open last used database.
      DatabaseName = Config.LastUsedDatabaseName;
      DatabasePath = Config.LastUsedDatabasePath;
      if (!QDir(DatabasePath).exists()) {
         Message = "<b>Database path for last used database do not exist.</b>";
         QMessageBox::critical(NULL, AppName, Message);
         return;
      }
   }

   // Request time interval from user.
   if (!NoTimeIntQuery && !SetTimeInterval()) return;

   // Set application status and type of databse we are trying to open.
   AppStatus = db_loading;
   CurrentDbType = normal_database;

   // Write message to log.
   Message = QString("Database load started (").append(DatabaseName).append(").");
   emit WriteLog(logmodel::info, Message);

   // Set sort indicator on DateTime column.
   ELVHeader->setSortIndicator(ColDateTime, Qt::AscendingOrder);

   // Start loading database into 'DB' class.
   DB.Open(normal_database, Config.StartDate, Config.EndDate, DatabasePath, DatabaseName);
}




// **************************************************
// Opens and read data from a folder with S-files.
// **************************************************
void MainWindow::Open_LocalDatabase()
{
   QString WorkDir;

   // Check if a database is already open.
   if (AppStatus == db_open) {
      Message = "<b>A database is already open.</b>";
      Message.append("<br>Please close current database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Check if a database is loading.
   if (AppStatus == db_loading) {
      Message = "<b>A database is currently loading.</b>";
      Message.append("<br>To stop loading, please close database.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   if (!OpenLastUsed) {
      // Set working directory for the folder selection dialg box.
      WorkDir = QDir::currentPath();
      if (Config.LastUsedDatabaseType == local_database && QDir(Config.LastUsedDatabasePath).exists()) {
         WorkDir = Config.LastUsedDatabasePath;
      }

      // Bring up file selection dialog box. User must point to a folder
      // with one or more S-files. Return if user cancels the operation.
      Message = "Please select a folder with S-files:";
      DatabasePath = QFileDialog::getExistingDirectory(this, Message, WorkDir, QFileDialog::ShowDirsOnly);

      // Return if user did not select a folder.
      if (DatabasePath.isNull()) return;

      // Use native separators in the path.
      DatabasePath = QDir::toNativeSeparators(DatabasePath);

   } else {
      // Open last used database.
      DatabasePath = Config.LastUsedDatabasePath;
      if (!QDir(DatabasePath).exists()) {
         Message = "<b>Path to last used database do not exist.</b>";
         QMessageBox::critical(NULL, AppName, Message);
         return;
      }
   }

   // Set database name to ",,";
   DatabaseName = ",,";

   // Set application status and type of DB we are trying to open.
   AppStatus = db_loading;
   CurrentDbType = local_database;

   // Write message to log.
   Message = QString("Database load started for local database (").append(DatabasePath).append(").");
   emit WriteLog(logmodel::info, Message);

   // Set sort indicator on DateTime column.
   ELVHeader->setSortIndicator(ColDateTime, Qt::AscendingOrder);

   // Set current working directory to the database directory.
   QDir::setCurrent(DatabasePath);

   // Load database into 'DB' class.
   DB.Open(local_database, Config.StartDate, Config.EndDate, DatabasePath, DatabaseName);
}




// **************************************************
// Opens and read from an index file.
// **************************************************
void MainWindow::Open_IndexFile()
{
   QString WorkDir;

   // Check if a database is already open.
   if (AppStatus == db_open) {
      Message = "<b>A database is already open.</b>";
      Message.append("<br>Please close current database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Check if a database is loading.
   if (AppStatus == db_loading) {
      Message = "<b>A database is currently loading.</b>";
      Message.append("<br>To stop loading, please close database.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   if (!OpenLastUsed) {
       // Set working directory for the file selection dialg box.
       if (Config.LastUsedDatabaseType == index_file && QFile(Config.LastUsedDatabaseName).exists()) {
          WorkDir = Config.LastUsedDatabaseName;
       } else {
          WorkDir = QDir::currentPath();
       }

       // Bring up file selection dialog box. Return if user cancels
       // the operation. User must point to an index file.
       Message = "Please select an index file:";
       DatabaseName = QFileDialog::getOpenFileName(this, Message, WorkDir);

       // Return if user did not select a file.
       if (DatabaseName.isNull()) return;

   } else {
       // Open last used index file.
       DatabaseName = Config.LastUsedDatabaseName;
       if (!QFile(DatabaseName).exists()) {
          Message = "<b>Last used index file do not exist.</b>";
          QMessageBox::critical(NULL, AppName, Message);
          return;
       }
   }

   // Set application status and type of DB we are trying to open.
   AppStatus = db_loading;
   CurrentDbType = index_file;

   // Write message to log.
   Message = QString("Database load started with index file (").append(DatabaseName).append(").");
   emit WriteLog(logmodel::info, Message);

   // Set sort indicator on DateTime column.
   ELVHeader->setSortIndicator(ColDateTime, Qt::AscendingOrder);

   // Load events in index file into 'DB' class.
   DB.Open(index_file, Config.StartDate, Config.EndDate, DatabasePath, DatabaseName);
}




// **************************************************
// Open and read the last used database.
// **************************************************
void MainWindow::Open_LastUsedDatabase()
{
   // Check if a database is already open.
   if (AppStatus == db_open) {
      Message = "<b>A database is already open.</b>";
      Message.append("<br>Please close current database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Check if a database is loading.
   if (AppStatus == db_loading) {
      Message = "<b>A database is currently loading.</b>";
      Message.append("<br>To stop loading, please close database.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Open last used database.
   if (Config.LastUsedDatabaseType != no_database) {
      OpenLastUsed = true;
      if (Config.LastUsedDatabaseType == default_database) Open_DefaultDatabase();
      if (Config.LastUsedDatabaseType == normal_database) Open_Database();
      if (Config.LastUsedDatabaseType == local_database) Open_LocalDatabase();
      if (Config.LastUsedDatabaseType == index_file) Open_IndexFile();
      OpenLastUsed = false;
   } else {
      Message = "<b>No database has previously been used.</b>";
      QMessageBox::critical(NULL, AppName, Message);
   }
}




// **************************************************
// This slot is called by the database to report
// progress status when opening a database.
// **************************************************
void MainWindow::DbOpenProgress(db::dbStatus status, int progress)
{
   // Explorer is scanning for S-files.
   if (status == db::enum_sfiles) {
      DoOnceAtLoadStart = true;
      // Display number of S-files found.
      sprintf(str11, "%i", progress);
      Message = "Loading database, enumerating eventfiles... (";
      Message.append(QString(str11) + " files found.)");
      statusBar()->showMessage(Message, 0);
   }

   // Explorer is reading S-files.
   if (status == db::read_sfiles) {
      // Display read progress in percents.
      sprintf(str11, "%i", progress);
      Message = "Loading database, reading eventfiles... (";
      Message.append(QString(str11)+ "% done.)");
      statusBar()->showMessage(Message, 0);

      // Things we do only once at start.
      // Wait until 200 events are read.
      if (DoOnceAtLoadStart && DB.GetNumberOfEvents() >= 200) {
         // Reset flag.
         DoOnceAtLoadStart = false;
         // Reset the Event List View model and select first row in Event List View.
         ELVModel->Reset(); ELVModel->SelectRow(0, true, true, false);
         // Add new data to the Event List View model.
         ELVModel->FetchMoreData();
         // Fit column widths automatically.
         ui->EventListView->resizeColumnsToContents();
         // Switch to Event List View.
         ui->tabWidget->setCurrentIndex(0);
      } else {
         // Add new data to the Event List View model.
         if (!DoOnceAtLoadStart) ELVModel->FetchMoreData();
      }
   }
}




// **************************************************
// This slot is called by the database to report
// progress status when filtering a database.
// **************************************************
void MainWindow::DbFilterProgress(char progress)
{
   // Display filtering progress in percents.
   if (progress < 100) {
      sprintf(str11, "%i", progress);
      Message = "Filtering database, ";
      Message.append(QString(str11) + "% done.");
      statusBar()->showMessage(Message, 0);
   } else {
      Message = "Database filtering done.";
      statusBar()->showMessage(Message, 3000);
   }
}




// *********************************************************
// This slot is called by the database when db.open() has
// finished. The result is returned in 'status' and 'error'.
// *********************************************************
void MainWindow::DbOpenResult(db::dbStatus status, db::dbError error)
{
   //
   // Database could not be opened.
   //
   if (status == db::closed) {
      // Restore current working directory if needed.
      if (CurrentDbType == local_database) {
         if (!Config.UseAltWorkDir) QDir::setCurrent(Config.DefWorkDir);
         if ( Config.UseAltWorkDir) QDir::setCurrent(Config.AltWorkDir);
      }
      // Change application status.
      AppStatus = db_closed;
      // Reset the event list view model.
      ELVModel->Reset();
      // Display error message in program log or otherwise.
      statusBar()->clearMessage();
      if (error == db::cancelled) {
         statusBar()->showMessage(QString("Load cancelled."), 10000);
         Message = "Database load was cancelled by user.";
         emit WriteLog(logmodel::info, Message);
         ui->tabWidget->setCurrentIndex(1);
      }
      if (error == db::out_of_memory) {
         statusBar()->showMessage(QString("Database closed."), 10000);
         Message = "Out of memory. Unable to open database.";
         emit WriteLog(logmodel::error, Message);
         ui->tabWidget->setCurrentIndex(1);
      }
      if (error == db::corrpupt_files) {
         statusBar()->showMessage(QString("Database closed."), 10000);
         Message = "All files in database are corrupted. Unable to open database.";
         emit WriteLog(logmodel::error, Message);
         ui->tabWidget->setCurrentIndex(1);
      }
      if (error == db::no_files_found) {
         statusBar()->clearMessage();
         if (CurrentDbType == local_database || CurrentDbType == index_file) {
            QMessageBox::information(NULL, AppName, "No event files could be found.");
         } else {
            QMessageBox::information(NULL, AppName, "No events was found. Check current time interval.");
         }
      }
      return;
   }

   //
   // Database has been opened (with or without errors).
   //
   if (status == db::open) {
      // This code is also needed here in case
      // less than 200 events have been opened.
      if (DoOnceAtLoadStart) {
         // Reset flag.
         DoOnceAtLoadStart = false;
         // Reset the Event List View model.
         ELVModel->Reset(); ELVModel->SelectRow(0, true, true, false);
         // Add new data to the Event List View model.
         ELVModel->FetchMoreData();
      } else {
         // Make sure all data is added to the Event List View model.
         ELVModel->FetchMoreData();
      }

      // Fit column widths automatically.
      ui->EventListView->resizeColumnsToContents();

      // Write number of events in database into str11.
      sprintf(str11, "%i", DB.GetNumberOfEvents());

      // Change application status.
      AppStatus = db_open;

      // Set new window title, and save information
      // about the current database to configuration.
      if (CurrentDbType == default_database) {
         // Set new window title.
         SetWindowTitle();
         // Save 'last database used' information to configuration.
         Config.LastUsedDatabaseType = default_database;
         Config.LastUsedDatabaseName.clear();
         Config.LastUsedDatabasePath.clear();
         // Write message to log.
         Message = "Default database opened with ";
         Message.append(QString(str11)).append(" events.");
         emit WriteLog(logmodel::info, Message);
      }
      if (CurrentDbType == normal_database) {
         // Set new window title.
         SetWindowTitle();
         // Save 'last database used' information to configuration.
         Config.LastUsedDatabaseType = normal_database;
         Config.LastUsedDatabaseName = DatabaseName;
         Config.LastUsedDatabasePath = DatabasePath;
         // Write message to log.
         Message = "Database opened with ";
         Message.append(QString(str11)).append(" events.");
         emit WriteLog(logmodel::info, Message);
      }
      if (CurrentDbType == local_database) {
         // Set new window title.
         SetWindowTitle();
         // Save 'last database used' information to configuration.
         Config.LastUsedDatabaseType = local_database;
         Config.LastUsedDatabaseName.clear();
         Config.LastUsedDatabasePath = DatabasePath;
         // Write message to log.
         Message = "Local database opened with ";
         Message.append(QString(str11)).append(" events.");
         emit WriteLog(logmodel::info, Message);
      }
      if (CurrentDbType == index_file) {
         // Set new window title.
         SetWindowTitle();
         // Save 'last database used' information to configuration.
         Config.LastUsedDatabaseType = index_file;
         Config.LastUsedDatabaseName = DatabaseName;
         Config.LastUsedDatabasePath.clear();
         // Write message to log.
         Message = "Index file opened with ";
         Message.append(QString(str11)).append(" events.");
         emit WriteLog(logmodel::info, Message);
      }

      // Show message on statusbar.
      statusBar()->showMessage(QString("Database opened."), 10000);

      // If database is of type local database or index file,
      // then the user given time interval is neglected. Instead
      // we have to find (and show on the status bar) the actual
      // time interval that has been loaded into the database.
      if (CurrentDbType == local_database || CurrentDbType == index_file) {
         hypocenter_ *Hyp = DB.GetEventByIndex(0)->hypocenters.first;
         Config.StartDate.setDate(Hyp->time.year, Hyp->time.month, Hyp->time.day);
         Hyp = DB.GetEventByIndex(DB.GetNumberOfEvents()-1)->hypocenters.first;
         Config.EndDate.setDate(Hyp->time.year, Hyp->time.month, Hyp->time.day);
         SetStatusBarInterval();
      }

      // Inform user of any database error.
      if (error != db::no_error) {
         if (error == db::out_of_memory) {
            // We run out of memory while loading eventfiles.
            Message = "Out of memory while reading from database. ";
            Message.append("Seisan database has only been partly loaded.");
            emit WriteLog(logmodel::warning, Message);
         }
         // Switch to Log View to show messages.
         ui->tabWidget->setCurrentIndex(1);
      } else {
         // Switch to Event List View to show events.
         ui->tabWidget->setCurrentIndex(0);
      }
   }
}




// **************************************************
// Close the database.
// **************************************************
void MainWindow::Close_Database()
{
   bool close = false;
   switch (AppStatus) {
   case db_open:
      // Close open database.
      if (NoCloseQuery) {
         close = true;
      } else {
         Message = "Do you want to close the database?";
         close = (QMessageBox::question(this, AppName, Message, QMB_YES|QMB_NO, QMB_YES) == QMB_YES);
      }
      if (close) {
         // User have chosen to close the database.
         if (CurrentDbType == local_database) {
            if (!Config.UseAltWorkDir) QDir::setCurrent(Config.DefWorkDir);
            if ( Config.UseAltWorkDir) QDir::setCurrent(Config.AltWorkDir);
         }
         statusBar()->showMessage((char*)"Closing database, please wait...", 0);
         DB.Close();
         statusBar()->showMessage((char*)"Database closed.", 5000);
         emit WriteLog(logmodel::info, QString("Database closed."));
         AppStatus = db_closed;
         // Update the window title.
         SetWindowTitle();
         // Switch to the Log View.
         ui->tabWidget->setCurrentIndex(1);
         // Reset the event list view model.
         ELVModel->Reset();
      }
      break;

   case db_loading:
      // Database is currently loading. Loading will be terminated.
      DB.Close();
      Message = "Closing database, please wait...";
      statusBar()->showMessage(Message, 0);
      break;

   case db_closed:
      // Database is already closed.
      Message = "No database is open.";
      statusBar()->showMessage(Message, 5000);
      break;
   }
}




// *********************************************************
// Handles File|Exit and close button in upper right corner.
// *********************************************************
void MainWindow::closeEvent(QCloseEvent *event)
{
   // If a database is currently open, then ask
   // user for permission to close the program.
   if (AppStatus == db_open) {
      Message = "Do you want to close the database and exit?";
      if (QMessageBox::question(this, AppName, Message, QMB_YES | QMB_NO, QMB_YES) == QMB_YES) {
         // User have chosen to close the program, close the database.
         statusBar()->showMessage((char*)"Closing database, please wait...", 0);
         DB.Close();
         CleanUpOnExit();
         // Exit the program.
         event->accept();
      } else {
         event->ignore();
      }
      return;
   }

   // A database is now being loaded. User must manually
   // close the database before he can close the program.
   if (AppStatus == db_loading) {
       Message = "<b>Error:</b> Busy loading database. To stop loading, please close database.";
       QMessageBox::critical(NULL, AppName, Message);
       event->ignore();
       return;
   }

   // No database is open. Close the program without user permission.
   if (AppStatus == db_closed) {
      // Remove temporary files from the work directory.
      CleanUpOnExit();
      // Exit the program.
      event->accept();
   }
}




// **************************************************
// Let the user edit the SEISAN.DEF file.
// **************************************************
void MainWindow::Edit_SeisanDef()
{
   EditParamFile(1, 0);
}




// **************************************************
// Let the user edit the MULPLT.DEF file.
// **************************************************
void MainWindow::Edit_MulpltDef()
{
   EditParamFile(2, 0);
}




// ***********************************************************
// Brings up the 'Apply filter to Event List view' dialog box.
// ***********************************************************
void MainWindow::SetFilter()
{
   if (AppStatus == db_open) {
      // Create and initialize the dialog.
      FilterDlg = new FilterViewDlg(this);
      FilterDlg->setAttribute(Qt::WA_DeleteOnClose, true);
      FilterDlg->InitDialog(&DbFilter, AppName);

      // Open the dialog box.
      if (FilterDlg->exec() == QDialog::Accepted) {
         if (DbFilter.InclQuery.size()) {
            // Apply the filter.
            DB.ApplyFilter(DbFilter);
            ELVModel->Reset();
            ELVModel->SelectRow(0, false, true, false);
         } else {
            // Remove any existing filter.
            if (DB.isFilterApplied()) {
               DB.RemoveFilter();
               ELVModel->Reset();
               ELVModel->SelectRow(0, false, true, false);
            }
         }
      }
   } else {
      if (AppStatus == db_closed) Message = "No database is open.";
      if (AppStatus == db_loading) Message = "Database is busy loading.";
      statusBar()->showMessage(Message, 3000);
   }
}




// ***********************************************************
// Remove a filter from the Event List view.
// ***********************************************************
void MainWindow::RemoveFilter()
{
   if (AppStatus == db_open) {
      // Remove any existing filter.
      if (DB.isFilterApplied()) {
         DB.RemoveFilter();
         ELVModel->Reset();
         ELVModel->SelectRow(0, false, true, false);
      }
   } else {
      if (AppStatus == db_closed) Message = "No database is open.";
      if (AppStatus == db_loading) Message = "Database is busy loading.";
      statusBar()->showMessage(Message, 3000);
   }
}




// **************************************************
// Brings up the 'Set Time Interval' dialog box.
// **************************************************
bool MainWindow::SetTimeInterval()
{
   // Save current start and stop dates.
   QDate StartDate = Config.StartDate;
   QDate EndDate = Config.EndDate;

   // Create and initialize the dialog.
   TimeIntDlg = new TimeIntervalDlg(this);
   TimeIntDlg->InitDialog(&Config);

   // Open the dialog box.
   int Status = TimeIntDlg->exec(); delete TimeIntDlg;
   if (Status == QDialog::Accepted)     {
      // Write a message in the Log View.
      Message = "Time interval set from '" + Config.StartDate.toString("yyyy.MM.dd");
      Message.append("' to '" + Config.EndDate.toString("yyyy.MM.dd") + "'.");
      emit WriteLog(logmodel::info, QString(Message));

      // Update status bar text.
      SetStatusBarInterval();

      // If a database is currently open and the user has changed
      // the timeinterval, then reopen the database again with new
      // time interval. User will have to confirm the action first.
      // If a filter is currently active them it will be reapplied.
      if (AppStatus == db_open) {
         if (StartDate != Config.StartDate || EndDate != Config.EndDate) {
            Message = "Time interval has changed. Do you want to reload the database?";
            if (QMessageBox::question(this, AppName, Message, QMB_YES | QMB_NO, QMB_YES) == QMB_YES) {
               // Close current datatabase. Do not ask for confirmation
               NoCloseQuery = true;
               Close_Database();
               NoCloseQuery = false;
               // Open last used datatabase. Do not ask for time interval.
               NoTimeIntQuery = true;
               Open_LastUsedDatabase();
               NoTimeIntQuery = false;
            }
         }
      }
      return true;
   }

   return false;
}




// ***********************************************
// Brings up the 'View|Select Columns' dialog box.
// ***********************************************
void MainWindow::SelectColumns()
{
   // Create the 'Select Columns' dialog.
   SelectColDlg = new SelectColumnsDlg(this);

   // Set intial state of the checkboxes. Note that the
   // dialog box class will return the new checkbox states
   // when the user closes the dialg box with the 'OK' button.
   SelectColDlg->InitDialog(&Config);

   // Open the dialog box.
   if (SelectColDlg->exec() == QDialog::Accepted) {
      // User closed the dialog box by pressing the 'OK' button.
      // New checkbox states has been returned to 'Config.EvtListColStat'.
      // Update the event list widget.
      SetEventListViewHeaders();
      ui->EventListView->resizeColumnsToContents();
   }

   // Destroy the dialog.
   delete SelectColDlg;
}




// ***********************************************
// Brings up the 'File|Configure' dialog box.
// ***********************************************
void MainWindow::SetConfiguration()
{
   QString OldCWD, NewCWD;

   // Create the 'Configure' dialog.
   ConfigDlg = new ConfigureDlg(this);
   ConfigDlg->setAttribute(Qt::WA_DeleteOnClose, true);

   // Set intial state of dialog. Note that the dialog
   // box class will return the new state when the
   // user closes the dialg box with the 'OK' button.
   ConfigDlg->InitDialog(&Config, AppStatus);

   // Save Current working directory.
   OldCWD = QDir::currentPath();

   // Open the dialog box.
   if (ConfigDlg->exec() != QDialog::Accepted) return;

   // Check if current working directory has changed.
   SetCurrentDirectory(&Config); // Try setting CWD from 'Config'.
   NewCWD = QDir::currentPath(); // Read the current CWD.
   QDir::setCurrent(OldCWD);     // Restore back to last CWD.

   // Handle configuration changes.
   if (NewCWD != OldCWD) {
      // Remove the work directory lock
      // file if it has been created.
      if (LockFileCreated) {
         LockFileCreated = false;
         QFile::remove(WORKDIR_LOCK_FILE);
      }
      // Change to the new work directory.
      SetCurrentDirectory(&Config);
      // Create Google Earth KML file in new work directory.
      CreatePriKmlFile(Config);
      // Write message about new work directory to the log.
      Message = "Work directory has been set to '";
      Message.append(QDir::toNativeSeparators(NewCWD) + "'.");
      emit WriteLog(logmodel::info, Message);
      // Create a new lock file in the new work directory.
      if (!QFile::exists(WORKDIR_LOCK_FILE)) {
         // Create a new lock file.
         CreateLockFile(); LockFileCreated = true;
      } else {
         // This directory already have a lock file. Warn
         // user that work directory may already be in use.
         QMessageBox msgBox;
         msgBox.setIcon(QMessageBox::Warning);
         msgBox.setWindowTitle("Lock file found.");
         msgBox.setTextFormat(Qt::RichText);
         Message = "<b> Warning:</b> The lock file '";
         Message.append(WORKDIR_LOCK_FILE);
         Message.append("' exists in work directory.<br>");
         Message.append("Directory may already be in use by another Seisan Explorer user.");
         msgBox.setText(Message);
         msgBox.addButton(" Continue ", QMessageBox::AcceptRole);
         msgBox.addButton(" Remove lock file ", QMessageBox::RejectRole);
         LockFileCreated = (msgBox.exec() == QMessageBox::RejectRole);
      }
   } else {
      // Handle Google Earth support.
      CreatePriKmlFile(Config);
   }
}




// ******************************************************
// Brings up the 'Poisson Distribution' dialog box.
// ******************************************************
void MainWindow::PoissonDistribution()
{
   // Check that database is open.
   if (AppStatus != db_open) {
      Message = "<b>No database is open.</b>";
      Message.append("<br>Please open a database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Bring up the dialog box.
   poissondlg *dialog = new poissondlg(&DB);
   dialog->exec(); delete dialog;
}




// ******************************************************
// Brings up the 'Gutenberg-Richter Relation' dialog box.
// ******************************************************
void MainWindow::GutenbergRichterRelation()
{
   // Check that database is open.
   if (AppStatus != db_open) {
      Message = "<b>No database is open.</b>";
      Message.append("<br>Please open a database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Bring up the dialog box.
   GutenbergRichterDlg *dialog = new GutenbergRichterDlg(&DB);
   dialog->exec(); delete dialog;
}




// ******************************************************
// Brings up the 'Completeness' dialog box.
// ******************************************************
void MainWindow::CompletenessCheck()
{
   // Check that database is open.
   if (AppStatus != db_open) {
      Message = "<b>No database is open.</b>";
      Message.append("<br>Please open a database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Bring up the dialog box.
   CompletenessCheckDlg *dialog = new CompletenessCheckDlg(&DB);
   dialog->exec(); delete dialog;
}




// ******************************************************
// Brings up the 'Weichert Method' dialog box.
// ******************************************************
void MainWindow::WeichertMethod()
{
   // Check that the file se-completenss.out exists.
   QFile inputFile("se-completenss.out");
   if (!inputFile.open(QIODevice::ReadOnly))    {
      Message = "<b>There is no se-completeness.out file in WOR.</b>";
      Message.append("<br>Please create the file e.g. with the function Completeness check.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Bring up the dialog box.
   WeichertDlg *dialog = new WeichertDlg(&DB);
   dialog->exec(); delete dialog;
}




// ******************************************************
// Brings up the 'Set event indicator' dialog box.
// ******************************************************
void MainWindow::SetEventIndicator()
{
   // Check that database is open.
   if (AppStatus != db_open) {
      Message = "<b>No database is open.</b>";
      Message.append("<br>Please open a database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Bring up the dialog box.
   SetEventIndicatorDlg *dialog = new SetEventIndicatorDlg(&DB);
   dialog->exec(); delete dialog;
}




// ******************************************************
// Brings up the 'Set model indicator' dialog box.
// ******************************************************
void MainWindow::SetModelIndicator()
{
   // Check that database is open.
   if (AppStatus != db_open) {
      Message = "<b>No database is open.</b>";
      Message.append("<br>Please open a database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Bring up the dialog box.
   SetModelIndicatorDlg *dialog = new SetModelIndicatorDlg(&DB);
   dialog->exec(); delete dialog;
}




// ******************************************************
// Brings up the 'Magnitude vs Magnitude' dialog box.
// ******************************************************
void MainWindow::MagnitudeVsMagnitude()
{
   // Check that database is open.
   if (AppStatus != db_open) {
      Message = "<b>No database is open.</b>";
      Message.append("<br>Please open a database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Bring up the dialog box.
   MagVsMagDlg *dialog = new MagVsMagDlg(&DB);
   dialog->exec(); delete dialog;
}




// ******************************************************
// Brings up the 'Events per year' dialog box.
// ******************************************************
void MainWindow::EventsPerYear()
{
   // Check that database is open.
   if (AppStatus != db_open) {
      Message = "<b>No database is open.</b>";
      Message.append("<br>Please open a database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Bring up the dialog box.
   EventsPerYearDlg *dialog = new EventsPerYearDlg(&DB);
   dialog->exec(); delete dialog;
}




// ******************************************************
// Brings up the 'Time of day distribution' dialog box.
// ******************************************************
void MainWindow::TimeOfDayDistribution()
{
   // Check that database is open.
   if (AppStatus != db_open) {
      Message = "<b>No database is open.</b>";
      Message.append("<br>Please open a database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Bring up the dialog box.
   TimeOfDayDistrDlg *dialog = new TimeOfDayDistrDlg(&DB);
   dialog->exec(); delete dialog;
}




// ******************************************************
// Brings up the 'Drift of hypocenters' dialog box.
// ******************************************************
void MainWindow::DriftOfHypocenters()
{
   // Check that database is open.
   if (AppStatus != db_open) {
      Message = "<b>No database is open.</b>";
      Message.append("<br>Please open a database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Bring up the dialog box.
   DriftOfHypocentersDlg *dialog = new DriftOfHypocentersDlg(&DB);
   dialog->exec(); delete dialog;
}




// ******************************************************
// Brings up the 'Sample function A' dialog box.
// ******************************************************
void MainWindow::Sample_a()
{
   // Check that database is open.
   if (AppStatus != db_open) {
      Message = "<b>No database is open.</b>";
      Message.append("<br>Please open a database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Bring up the dialog box.
   sample_function_a *dialog = new sample_function_a(&DB);
   dialog->exec(); delete dialog;
}




// ******************************************************
// Brings up the 'Sample function B' dialog box.
// ******************************************************
void MainWindow::Sample_b()
{
   // Check that database is open.
   if (AppStatus != db_open) {
      Message = "<b>No database is open.</b>";
      Message.append("<br>Please open a database first.");
      QMessageBox::critical(NULL, AppName, Message);
      return;
   }

   // Bring up the dialog box.
   sample_function_b *dialog = new sample_function_b(&DB);
   dialog->exec(); delete dialog;
}




// *************************************
// Brings up the 'Help|About' dialog box.
// *************************************
void MainWindow::AboutSeisanExplorer()
{
   aboutdlg *dialog = new aboutdlg(this);
   dialog->exec(); delete dialog;
}




// ********************************************************************
// Brings up the Event List View popup meny. This function is signalled
// from EventListView when user right-clicks inside the vidget view.
// ********************************************************************
void MainWindow::CCMR_EventListView(const QPoint &pos)
{
   QAction* menuItem;
   event_node_ *EventPtr;
   QModelIndexList ModelIndexes;

   // Get the list of selected rows.
   ModelIndexes = ui->EventListView->selectionModel()->selectedRows();
   int NumSelectedRows = ModelIndexes.count();

   // Exit if no rows are selected. This should not be possible.
   if (!NumSelectedRows) return;

   // If one events are selected, get pointer to the first event.
   EventPtr = DB.GetEventByIndex(ModelIndexes.at(0).row());

   // Create a up the popup menu adapted to current selection.
   QMenu *popupMenu = new QMenu(this);

   // Add 'Associate' item to menu.
   menuItem = new QAction(tr("Associate"), this);
   menuItem->setShortcut(QKeySequence("a"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Associate_Events()));
   if (AppStatus != db_open) menuItem->setDisabled(true);
   if (NumSelectedRows > 1) menuItem->setDisabled(true);

   // Add 'Copy to file' item to menu.
   menuItem = new QAction(tr("Copy to file"), this);
   menuItem->setShortcut(QKeySequence("c"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Copy_Events()));

   // Add 'Delete' item to menu.
   menuItem = new QAction(tr("Delete"), this);
   menuItem->setShortcut(QKeySequence("d"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Delete_Event()));
   if (CurrentDbType == index_file) menuItem->setDisabled(true);
   if (AppStatus != db_open) menuItem->setDisabled(true);
   if (NumSelectedRows > 10) menuItem->setDisabled(true);

   // Add 'Duplicate' item to menu.
   menuItem = new QAction(tr("Duplicate"), this);
   menuItem->setShortcut(QKeySequence("shift+d"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Duplicate_Event()));
   if (CurrentDbType == index_file) menuItem->setDisabled(true);
   if (AppStatus != db_open) menuItem->setDisabled(true);
   if (NumSelectedRows > 1) menuItem->setDisabled(true);

   // Add 'Edit comment lines' item to menu.
   menuItem = new QAction(tr("Edit comment lines"), this);
   menuItem->setShortcut(QKeySequence("shift+c"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Edit_Comment_Lines()));
   if (NumSelectedRows > 1) menuItem->setDisabled(true);

   // Add 'Edit STATIONx.HYP file' item to menu.
   menuItem = new QAction(tr("Edit STATIONx.HYP file"), this);
   menuItem->setShortcut(QKeySequence("shift+s"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Edit_StationHyp()));
   if (NumSelectedRows > 1) menuItem->setDisabled(true);

  // Add 'Edit with text editor' item to menu.
   menuItem = new QAction(tr("Edit with text editor"), this);
   menuItem->setShortcut(QKeySequence("e"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Edit_With_Text_Editor()));
   if (AppStatus != db_open) menuItem->setDisabled(true);
   if (NumSelectedRows > 1 ) menuItem->setDisabled(true);

   // Add 'EEV' item to menu.
   menuItem = new QAction(tr("EEV"), this);
   menuItem->setShortcut(QKeySequence("<"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Launch_EEV()));
   if (NumSelectedRows > 1 ) menuItem->setDisabled(true);

   // Add 'Locate' item to menu.
   menuItem = new QAction(tr("Locate"), this);
   menuItem->setShortcut(QKeySequence("l"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Locate_Event()));
   if (NumSelectedRows > 1) menuItem->setDisabled(true);

   // Add 'Merge' item to menu.
   menuItem = new QAction(tr("Merge"), this);
   menuItem->setShortcut(QKeySequence("m"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Merge_Events()));
   if (AppStatus != db_open) menuItem->setDisabled(true);
   if (NumSelectedRows < 2) menuItem->setDisabled(true);

   // Add 'Plot with mulplot' item to menu.
   menuItem = new QAction(tr("Plot with Mulplot"), this);
   menuItem->setShortcut(QKeySequence("p"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Plot_With_Mulplot()));
   if (AppStatus != db_open) menuItem->setDisabled(true);
   if (NumSelectedRows > 1) menuItem->setDisabled(true);
   if (NumSelectedRows == 1 && !EventPtr->wavefiles.numlines) menuItem->setDisabled(true);

   // Add 'Plot with mulplot (show plot menu)' item to menu.
   menuItem = new QAction(tr("Plot with Mulplot (show plot menu)"), this);
   menuItem->setShortcut(QKeySequence("shift+p"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Plot_With_Mulplot_Menu()));
   if (AppStatus != db_open) menuItem->setDisabled(true);
   if (NumSelectedRows > 1) menuItem->setDisabled(true);
   if (NumSelectedRows == 1 && !EventPtr->wavefiles.numlines) menuItem->setDisabled(true);

   // Add 'Register' item to menu.
   menuItem = new QAction(tr("Register"), this);
   menuItem->setShortcut(QKeySequence("r"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Register_Event()));
   if (AppStatus != db_open) menuItem->setDisabled(true);
   if (NumSelectedRows > 1) menuItem->setDisabled(true);

   // Add 'Set distance indicator' item to menu.
   menuItem = new QAction(tr("Set distance indicator"), this);
   menuItem->setShortcut(QKeySequence("shift+r"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Set_Distance_Indicator()));
   if (NumSelectedRows > 1) menuItem->setDisabled(true);

   // Add 'Show with Google Map' item to menu.
   menuItem = new QAction(tr("Show with Google Map"), this);
   menuItem->setShortcut(QKeySequence("g"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Show_Events_With_GM()));

   // Add 'Show with Google Earth' item to menu.
   menuItem = new QAction(tr("Show with Google Earth"), this);
   menuItem->setShortcut(QKeySequence("shift+g"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Show_Events_With_GE()));

   // Add 'Show with Seismicity Viewer' item to menu.
   menuItem = new QAction(tr("Show with Seismicity Viewer"), this);
   menuItem->setShortcut(QKeySequence("v"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Show_Events_With_SV()));

   // Add separator to menu.
   popupMenu->addSeparator();

   // Add 'Select all events' item to menu.
   menuItem = new QAction(tr("Select all events"), this);
   menuItem->setShortcut(QKeySequence("ctrl+a"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Select_All()));
   if (AppStatus != db_open) menuItem->setDisabled(true);

   // Add 'Set Filter' item to menu.
   menuItem = new QAction(tr("Set Filter ..."), this);
   menuItem->setShortcut(QKeySequence("ctrl+f"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(SetFilter()));
   if (AppStatus != db_open) menuItem->setDisabled(true);

   // Add 'Load event file into Explorer' item to menu.
   menuItem = new QAction(tr("Load event file into Explorer"), this);
   menuItem->setShortcut(QKeySequence("f3"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Load_Event()));
   if (AppStatus != db_open) menuItem->setDisabled(true);
   if (NumSelectedRows > 1 ) menuItem->setDisabled(true);

   // Add 'Unoad event file from Explorer' item to menu.
   menuItem = new QAction(tr("Unload event file from Explorer"), this);
   menuItem->setShortcut(QKeySequence("f4"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Unload_Event()));
   if (AppStatus != db_open) menuItem->setDisabled(true);
   if (NumSelectedRows > 1 ) menuItem->setDisabled(true);

   // Add 'Refresh view' item to menu.
   menuItem = new QAction(tr("Refresh view"), this);
   menuItem->setShortcut(QKeySequence("f5"));
   popupMenu->addAction(menuItem);
   connect(menuItem, SIGNAL(triggered()), this, SLOT(ELV_Refresh_View()));

   // Show popup menu at mouse position.
   popupMenu->exec(QCursor::pos());
}



// ********************************************************************
// This function edits the STATIONx.HYP file for the selected event.
// This function is signaled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Edit_StationHyp()
{
    char Model;
    event_node_ *Node;

    // Check that database is open.
    if (AppStatus == db_closed) return;

    // Get the list of selected rows.
    ELVModel->GetSelection(&ELVSelection);

    // Exit if no rows are selected.
    if (!ELVSelection.size()) return;

    // Only one row can be selected.
    if (ELVSelection.size() > 1) {
       Message = "Multiple selections are not allowed for this operation.";
       statusBar()->showMessage(Message, 3000);
       return;
    }

    // Get pointer to slelected event.
    Node = DB.GetEventByIndex(ELVSelection.at(0).index);

    // Get the Model for the selected event.
    Model = Node->hypocenters.first->model;

    // Bring up the editor.
    EditParamFile(3, Model);
}







// ********************************************************************
// This function duplicates the selected event.
// This function is signaled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Duplicate_Event()
{
   int DbIndex;
   QString FileName, DupFile;

   // Check that database is not an index file.
   if (CurrentDbType == index_file) {
       Message = "Duplicate is not supported when database is an index file.";
       statusBar()->showMessage(Message, 3000);
       return;
   }

   // Check that database is open.
   if (AppStatus == db_closed) return;
   if (AppStatus == db_loading) {
       Message = "Can not perform this operation while database is loading.";
       statusBar()->showMessage(Message, 3000);
       return;
   }

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Exit if no rows are selected.
   if (!ELVSelection.size()) return;

   // Only one row can be selected.
   if (ELVSelection.size() > 1) {
      Message = "Multiple selections are not allowed for this operation.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Get file name for this event.
   DbIndex = ELVSelection.at(0).index;
   DB.GetFileByIndex(DbIndex, &FileName);

   // Duplicate the file.
   if (!DuplicateSfileOnDisk(FileName, &DupFile, Operator.GetConfirmedId())) {
      Message = "Failed to duplicate event file. Check S-file and database folder permissions.";
      emit WriteLog(logmodel::error, Message);
      ui->tabWidget->setCurrentIndex(1);
   } else {
      // Duplication succeeded. Load file into database.
      DbIndex = DB.LoadEventAndResort(DupFile);
      if (DbIndex >= 0) {
         // Update the event log view.
         ELVModel->InsertRow(DbIndex);
         // Update the program window title.
         SetWindowTitle();
         // Write message in status bar.
         Message = "Duplication succeeded.";
         statusBar()->showMessage(Message, 3000);
      }
      if (DbIndex == -1) {
         Message = "Failed to load duplicated event. Check contents of S-file '" + DupFile + "'.";
         emit WriteLog(logmodel::error, Message);
         ui->tabWidget->setCurrentIndex(1);
      }
   }
}




// ********************************************************************
// This function handles the 'Edit with text edir' menu item.
// This function is signaled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Edit_With_Text_Editor()
{
   QString FileName;
   int EventId, Status;
   int DbIndex, NewIndex;
   FileEditorDlg *Editor;

   // Check that database is open.
   if (AppStatus == db_closed) return;
   if (AppStatus != db_open) {
       Message = "Can not perform this operation while database is loading.";
       statusBar()->showMessage(Message, 3000);
       return;
   }

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Exit if no rows are selected.
   if (!ELVSelection.size()) return;

   // A maximum of 10 rows can be selected.
   if (ELVSelection.size() > 10) {
      Message = "A maximum of 10 rows allowed for this operation.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Get file name and event id for this event.
   DbIndex = ELVSelection.at(0).index;
   DB.GetFileByIndex(DbIndex, &FileName);
   EventId = DB.GetIdByIndex(DbIndex);

   // Set position of editor on screen.
   QPoint Position = this->pos();
   Position.setX(Position.x()+200);
   Position.setY(Position.y()+200);

   // Bring up the editor. Editor will automatically
   // be destroyed when the user closes the window.
   Editor = new FileEditorDlg(this);
   Editor->move(Position);
   Editor->setAttribute(Qt::WA_DeleteOnClose, true);
   if (Editor->InitDialog(FileName, AppName)) {
      // File was sucessfully opened. Open editor window.
      if (Editor->exec() == QDialog::Accepted) {
         // Reload event into database.
         Status = DB.ReloadEventAndResort(DbIndex);
         if (Status) {
            Message = "Failed to reload updated event file. ";
            if (Status == 1) Message.append("Cannot open file.");
            if (Status == 2) Message.append("Error in file.");
            if (Status == 3) Message.append("Out of memory.");
            if (Status == 5) Message.append("Event has invalid date.");
            emit WriteLog(logmodel::error, Message);
            ui->tabWidget->setCurrentIndex(1);
         } else {
            // Update the event list view.
            NewIndex = DB.GetIndexById(EventId);
            if (NewIndex == DbIndex) {
               ELVModel->RefreshRow(DbIndex);
            } else {
               ELVModel->RemoveRow(DbIndex);
               ELVModel->InsertRow(NewIndex);
               ELVModel->SelectRow(NewIndex, false, true, true);
            }
         }
      }
   } else {
      // File could not be opened.
      Message = "Unable to open file '" + FileName + "'.";
      QMessageBox::critical(this, AppName, Message);
   }
}




// ********************************************************************
// This function handles the 'Plot with Mulplot (P)' menu item.
// This function is signalled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Plot_With_Mulplot()
{
   Plot_With_Mulplot(true);
}

// ********************************************************************
// This function handles the 'Plot with Mulplot (Shift-P)' menu item.
// This function is signalled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Plot_With_Mulplot_Menu()
{
   Plot_With_Mulplot(false);
}

// ********************************************************************
// This function handles the 'Plot with Mulplot' menu item.
// This function is signalled from the two functions above.
// ********************************************************************
void MainWindow::Plot_With_Mulplot(bool plotdefault)
{
   sfile_nameinfo NameInfo, NewNameInfo;
   QString FileName, Mulplt, Eev, Line, Command;
   int DbIndex, NewIndex, Status, dStatus, rStatus, EventId;

   // Check that database is open.
   if (AppStatus == db_closed) return;
   if (AppStatus != db_open) {
       Message = "Can not perform this operation while database is loading.";
       statusBar()->showMessage(Message, 3000);
       return;
   }

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Exit if no rows are selected.
   if (!ELVSelection.size()) return;

   // Only one row can be selected.
   if (ELVSelection.size() > 1) {
      Message = "Multiple selections are not allowed for this operation.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Select a platform dependant shell program.
   #ifdef Q_OS_WIN32
   QString shell = "cmd.exe /c ";
   #endif
   #ifdef Q_OS_LINUX
   QString shell = "xterm -e ";
   #endif

   // Create platform dependant names for 'mulplot' and 'eev'.
   #ifdef Q_OS_WIN32
   Eev = "eev.exe";
   Mulplt = "mulplt.exe";
   #endif
   #ifdef Q_OS_LINUX
   Eev = "eev";
   Mulplt = "mulplt";
   #endif

   // Check that the 'eev' program exists.
   if (!IsFileInPath(Eev)) {
      Message = "Plot with mulplot: Can not find eev in the path.";
      emit WriteLog(logmodel::error, Message);
      ui->tabWidget->setCurrentIndex(1);
      return;
   }

   // Check that the 'mulplt' program exists.
   if (!IsFileInPath(Mulplt)) {
      Message = "Plot with mulplot: Can not find mulplt in the path.";
      emit WriteLog(logmodel::error, Message);
      ui->tabWidget->setCurrentIndex(1);
      return;
   }

   // Get index for the selected event.
   DbIndex = ELVSelection.at(0).index;

   // Run mulplt several times in an endless loop that will be
   // terminated by a return statement. The loop will end if:
   // - The selected event has no wave files.
   // - We are unable to launch the mulplt program.
   // - The file 'seisan.mes' is empty.
   // - The file 'seisan.mes' cannot be opened.
   // - User try to pass top/bottom of the event list.
   // - All events in the event list has been deleted.
   // - An error occurs.
   while (true) {
      // Check that selected event has a wavefile.
      if (!DB.GetEventByIndex(DbIndex)->wavefiles.numlines) {
         Message = "Selected event has no wave file.";
         statusBar()->showMessage(Message, 3000);
         return;
      }

      // Get filename for the selected row.
      DB.GetFileByIndex(DbIndex, &FileName);

      // Execute mulplot with the selected event.
      Command = shell + Mulplt;
      if (plotdefault) Command.append(" -plotdefault");
      Command.append(" -operator " + Operator.GetConfirmedId());
      Command.append(" -base " + DatabaseName + " -sfile " + FileName);
      if (system(Command.toAscii().data()) == -1) {
         Message = "Unable to start 'muplt'. Failed to execute '" + Command + "'.";
         emit WriteLog(logmodel::error, Message);
         ui->tabWidget->setCurrentIndex(1);
         return;
      }

      // Read a mulplt command from the file 'seisan.mes' and
      // delete the file. If file has a command, then perform
      // the command, else end the loop. The command uses the
      // first 4 characters in the file only.
      Status = ReadFileLine("seisan.mes", 1, &Line);
      Line.truncate(4); Line=Line.trimmed();
      QFile::remove("seisan.mes");
      if (Status == 0) {
         // If line is empty, end the loop here.
         if (Line.isEmpty()) return;

         // A command was read from file. Execute the command.
         if (Line == "NEXT") {
            // Run mulplt again with next event in database.
            // Move selection to next event in Event List View.
            // Return if we pass the end of the event list.
            if (DbIndex == DB.GetNumberOfEvents()-1) return;
            ELVModel->SelectRow(++DbIndex, true, true, true);
            // Update the event list view by processing events.
            QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
         }
         if (Line == "BACK") {
            // Run mulplt again with previous event in database.
            // Move selection to previous event in Event List View.
            // Return if we pass the start of the event list.
            if (!DbIndex) return;
            ELVModel->SelectRow(--DbIndex, true, true, true);
            // Update the event list view by processing events.
            QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
         }
         if (Line == "D") {
            // Delete the selected event and go to next event.
            // Ask user for confirmation.
            Message = "Are you sure you want to delete this event from the Seisan database?\n" + FileName;
            QString Title = "Delete event";
            if (QMessageBox::question(this, Title, Message, QMB_YES | QMB_NO, QMB_NO) == QMB_YES) {
               // Delete the event.
               if (DeleteEvent(DbIndex, &dStatus, false)) {
                  // Update the event list view by processing events.
                  QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
                  // Exit if there are no more event in DB.
                  if (!DB.GetNumberOfEvents()) return;
               } else {
                  // Unable to delete event. Show error message.
                  ui->tabWidget->setCurrentIndex(1);
                  return;
               }
            }
         }
         if (Line == "REG") {
            // Run eev to register this event in database.
            // Get date/time from the name of the event file.
            DB.GetFileByIndex(DbIndex, &FileName);
            GetSfileNameInfo(FileName, &NameInfo);
            EventId = DB.GetIdByIndex(DbIndex);

            // Execute the 'eev' program.
            Command = shell + Eev + " " + NameInfo.DateStr + NameInfo.TimeStr + " ";
            Command.append(DatabaseName + " -seisanexp REG " + Operator.GetConfirmedId());
            if (system(Command.toAscii().data()) == -1) return;

            // Check return value in 'seisan.mes'.
            Status = ReadFileLine("seisan.mes", 1, &Line);
            Line=Line.trimmed(); QFile::remove("seisan.mes");
            if (Status == 0) {
               // The file 'seisan.mes' was found.
               if (Line.isEmpty()) {
                  // File name has not changed, reload the file.
                  rStatus = DB.ReloadEventAndResort(DbIndex);
                  if (rStatus == 0 || rStatus == 4) {
                     // Update the event list view.
                     ELVModel->RemoveRow(DbIndex);
                     NewIndex = DB.GetIndexById(EventId);
                     ELVModel->InsertRow(NewIndex);
                     ELVModel->SelectRow(NewIndex, false, true, false);
                     QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
                  } else {
                     Message = "Failed to reload registered event file.";
                     emit WriteLog(logmodel::error, Message);
                     ui->tabWidget->setCurrentIndex(1);
                     return;
                  }
               }
               if (Line.size()) {
                  // File name or database has changed. File will be
                  // reloaded if it has not been moved to another database.
                  GetSfileNameInfo(Line, &NewNameInfo);
                  if (NewNameInfo.Database.toUpper() == DatabaseName.toUpper()) {
                     // File has not been moved to another database, but the name
                     // has changed. Unload the old file and load the new file.
                     DB.RemoveEvent(DbIndex);
                     ELVModel->RemoveRow(DbIndex);
                     NewIndex = DB.LoadEventAndResort(Line);
                     if (NewIndex >= 0) {
                        ELVModel->InsertRow(NewIndex);
                        ELVModel->SelectRow(NewIndex, false, true, false);
                        QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
                     }
                     if (NewIndex == -1) {
                        Message = "Failed to reload registered event file.";
                        emit WriteLog(logmodel::error, Message);
                        ui->tabWidget->setCurrentIndex(1);
                        return;
                     }
                  } else {
                     // Event file has moved to another database.
                     // Remove the file from the Explorer database.
                     DB.RemoveEvent(DbIndex);
                     ELVModel->RemoveRow(DbIndex);
                     return;
                  }
               }
            } else {
               if (Status == 1) {
                  // File was not found. Registration failed.
                  Message = "Registration failed or was aborted by user.";
                  statusBar()->showMessage(Message, 5000);
                  return;
               }
               if (Status == 2) {
                  // File could not be opened.
                  Message = "Unable to open file 'seisan.mes' in work directory.";
                  emit WriteLog(logmodel::error, Message);
                  ui->tabWidget->setCurrentIndex(1);
                  return;
               }
            }
         }
      } else {
         // A command was not read from file.
         if (Status == 1) {
            // File does not exist.
            return;
         }
         if (Status == 2) {
            // File could not be opened.
            Message = "Unable to open file 'seisan.mes' in work directory.";
            emit WriteLog(logmodel::error, Message);
            ui->tabWidget->setCurrentIndex(1);
         }
         return;
      }
   }
}




// ********************************************************************
// This function handles the 'Locate event' menu item.
// This function is signalled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Locate_Event()
{
   int NewIndex, DbIndex;
   QString EventFile, Command;
   QDateTime ModTime1, ModTime2;

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Exit if no rows are selected.
   if (!ELVSelection.size()) return;

   // Only one row can be selected.
   if (ELVSelection.size() > 1) {
      Message = "Multiple selections are not allowed for this operation.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Get filename for the selected event.
   DbIndex = ELVSelection.at(0).index;
   DB.GetFileByIndex(DbIndex, &EventFile);

   // Save the file modification time.
   GetFileModTime(EventFile, &ModTime1);

   // Create a platform dependant name for the 'hyp' program.
   #ifdef Q_OS_WIN32
   Command = "hyp.exe";
   #endif
   #ifdef Q_OS_LINUX
   Command = "hyp";
   #endif

   // Check that the 'hyp' program exists.
   if (!IsFileInPath(Command)) {
      Message = "Locate event: Can not find hyp in the path.";
      emit WriteLog(logmodel::error, Message);
      ui->tabWidget->setCurrentIndex(1);
      return;
   }

   // We need to open a terminal if we running linux.
   #ifdef Q_OS_LINUX
   Command.prepend("xterm -e ");
   #endif

   // Append the arguments to 'hyp'.
   Command.append(" " + EventFile + " -seisanexp -op " + Operator.GetConfirmedId());

   // Execute the 'hyp' program.
   if (system(Command.toAscii().data()) == -1) {
      Message = "Unable to start hyp. Failed to execute '" + Command + "'.";
      emit WriteLog(logmodel::error, Message);
      ui->tabWidget->setCurrentIndex(1);
      return;
   }

   // Check if event file has been updated.
   GetFileModTime(EventFile, &ModTime2);
   if (ModTime1 != ModTime2) {
      // Eventfile has been updated. Save event identity.
      int EventId = DB.GetIdByIndex(DbIndex);
      // Reload the event file in database.
      if (DB.ReloadEventAndResort(DbIndex)) {
          // Event file could not be reloaded.
          Message = "Unable to locate event. Failed to reload the updated event file.";
          emit WriteLog(logmodel::error, Message);
          ui->tabWidget->setCurrentIndex(1);
      } else {
          NewIndex = DB.GetIndexById(EventId);
          if (NewIndex == DbIndex) {
             // Refresh the selected event.
             ELVModel->RefreshRow(DbIndex);
          } else {
             // Reset the view and select updated event.
             ELVModel->RemoveRow(DbIndex);
             ELVModel->InsertRow(NewIndex);
             ELVModel->SelectRow(NewIndex, false, true, true);
          }
      }
   }
}




// ***********************************************************************
// This function merges two or more selected events files.
// ***********************************************************************
void MainWindow::ELV_Merge_Events()
{
   QListEventID FailedMerges;
   QString DestFile, MergeFile;
   bool MergeError, ReloadError, DeleteError;
   int EventIndex, uStatus, mStatus, NumSelectedRows;
   int NumMergedFiles, NumDeletedFiles, EventID, dStatus;

   // Check that database is open.
   if (AppStatus == db_closed) return;
   if (AppStatus == db_loading) {
       Message = "Can not perform this operation while database is loading.";
       statusBar()->showMessage(Message, 3000);
       return;
   }

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Get number of selected rows.
   NumSelectedRows = ELVSelection.size();

   // Exit if less than two rows are selected.
   if (NumSelectedRows < 2) return;

   // Mark the destination event with a '+' in the view.
   ELVModel->MarkRow(ELVSelection.at(0).index);

   // Get file name of destination file.
   DB.GetFileByIndex(ELVSelection.at(0).index, &DestFile);

   // Create the dialog box.
   MergeDlg *Dialog = new MergeDlg(this);
   Dialog->setAttribute(Qt::WA_DeleteOnClose, true);

   // Build the dialog box message.
   Message = "You have selected " + QString::number(NumSelectedRows-1) + " event(s)";
   Message.append(" for merging with the event in row number ");
   Message.append(QString::number(ELVSelection.at(0).index+1));
   Message.append(".\n(Row number ");
   Message.append(QString::number(ELVSelection.at(0).index+1));
   Message.append(" has been marked with a plus sign in the 'Date and Time' column)\n\n");
   Message.append("Press 'Merge' if you want to keep the ");
   Message.append(QString::number(NumSelectedRows-1));
   Message.append(" event(s) after merging.\n");
   Message.append("Press 'Merge and delete' if you want the ");
   Message.append(QString::number(NumSelectedRows-1));
   Message.append(" event(s) to be deleted after merging.");

   // Set intial state of the dialog and open it.
   Dialog->Init(Message);
   uStatus = Dialog->exec();

   // Unmark the destination event in the view.
   ELVModel->UnmarkRow();

   // Return if user pressed 'Cancel'.
   if (!uStatus) return;

   // Write informational message in the log.
   Message = "Merge status: Starting merge operation for destination file " + DestFile + "'. ";
   emit WriteLog(logmodel::info, Message);

   // Merge files.
   NumMergedFiles = NumDeletedFiles = 0;
   MergeError = ReloadError = DeleteError = false;
   for (int i=1;i<NumSelectedRows;i++) {
      // Loop trough all files that are to be merged.
      // First get event id and name of file to be merged.
      EventID = ELVSelection.at(i).id;
      DB.GetFileById(EventID, &MergeFile);

      // Merge the two files and handle the merge result.
      mStatus = MergeSfilesOnDisk(DestFile, MergeFile);
      if (!mStatus) {
         // Merge succeeded.
         NumMergedFiles++;
         Message = "Merge status: Merge succeeded for '" + MergeFile + "'. ";
         emit WriteLog(logmodel::info, Message);
         // Delete merged file if requested.
         if (uStatus == 2) {
            if (DeleteEvent(DB.GetIndexById(EventID), &dStatus, true)) {
               // File has been deleted. Update the log.
               NumDeletedFiles++;
               Message = "Merge status: Successfully deleted '" + MergeFile + "'.";
               emit WriteLog(logmodel::info, Message);
            } else {
               // Unable to delete merge file.
               DeleteError = true;
               Message = "Merge status: Unable to delete merge file '" + MergeFile + "'. ";
               if (dStatus == 1) Message.append("File was not found.");
               if (dStatus == 2) Message.append("Could not delete file.");
               if (dStatus == 3) Message.append("Failed to locate DELET database.");
               if (dStatus == 4) Message.append("Failed to copy file to DELET database.");
               emit WriteLog(logmodel::error, Message);
            }
         }
      } else {
         // Merge failed for this file.
         MergeError = true;
         FailedMerges.append(ELVSelection.at(i));
         // Write error message.
         Message = "Merge status: Merge failed for '" + MergeFile + "'. ";
         if (mStatus == 1) Message.append("Unable to find one of the event files.");
         if (mStatus == 2) Message.append("Unable to open one of the event files.");
         if (mStatus == 3) Message.append("File is more than 24 hours apart from destination file.");
         if (mStatus == 4) Message.append("File is dated the day before the destination file.");
         if (mStatus == 5) Message.append("Unable to open destination file for writing.");
         emit WriteLog(logmodel::error, Message);
      }
   }

   // All files have now been processed. Reload
   // the destination event if it has been merged.
   if (NumMergedFiles) {
      EventIndex = DB.GetIndexById(ELVSelection.at(0).id);
      mStatus = DB.ReloadEventAndResort(EventIndex);
      if (mStatus > 0 && mStatus < 4) {
         // Destination event cannot be reloaded.
         ReloadError = true;
         Message = "Unable to reload the destination event. ";
         if (mStatus == 1) Message.append("Cannot open file.");
         if (mStatus == 2) Message.append("Error in file.");
         if (mStatus == 3) Message.append("Out of memory.");
         if (mStatus == 5) Message.append("Event has invalid date.");
         emit WriteLog(logmodel::error, Message);
      }
   }

   // If events have been deleted, clear all selections. Then reselect
   // the destination event. Also reselect all events that failed to merge.
   if (NumDeletedFiles) {
      ELVModel->ClearSelection();
      if (!ReloadError) {
         EventIndex = DB.GetIndexById(ELVSelection.at(0).id);
         ELVModel->SelectRow(EventIndex, false, true, false);
      }
      for (int i=0;i<FailedMerges.size();i++) {
         EventIndex = DB.GetIndexById(FailedMerges.at(i).id);
         ELVModel->SelectRow(EventIndex, false, false, false);
      }
   }

   // Switch to log view if an error has occurred,
   // else inform user that the operation succeeded.
   if (MergeError || ReloadError || DeleteError) {
      ui->tabWidget->setCurrentIndex(1);
   } else {
      Message = "Merge operation completed successfully.";
      statusBar()->showMessage(Message, 3000);
   }
}




// ***********************************************************************
// This function copies one or more selecgted event files to outfile. This
// function is signalled from the Event List View popup menmu. Name of out
// file is 'copy.out'. Outfile is placed in the current working directory.
// ***********************************************************************
void MainWindow::ELV_Copy_Events()
{
   QString FileName;
   int NumSelectedRows;
   QStringList FileList;

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Get number of selected rows.
   NumSelectedRows = ELVSelection.size();

   // Exit if no rows are selected.
   if (!NumSelectedRows) return;

   // Add all selected event files to file list.
   for (int i=0;i<NumSelectedRows;i++) {
      // Get filename for this row.
      DB.GetFileByIndex(ELVSelection.at(i).index, &FileName);
      // Add to file list.
      FileList.append(FileName);
   }

   // Merge together all files in the file list
   // and save the result in the outfile. The
   // files in the file list are not modified.
   if (!MergeFiles(FileList, QString("copy.out"))) {
      // Merge failed. Write error message in the log.
      Message = "The 'Copy to file' function failed. Check disk space and write permissions for work directory.";
      emit WriteLog(logmodel::error, Message);
      // Switch to Log View to show message.
      ui->tabWidget->setCurrentIndex(1);
   } else {
      QString FileName = QDir::toNativeSeparators(QDir::currentPath() + "/copy.out");
      Message = "Selected events have been copied to '" + FileName + "'.";
      statusBar()->showMessage(Message, 5000);
   }
}




// ********************************************************************
// This function handles the 'Register' menu item.
// This function is signalled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Register_Event()
{
   QString FileName, Eev, Command, Line;
   sfile_nameinfo NameInfo, NewNameInfo;
   int DbIndex, NewIndex, Status, rStatus, EventId;

   // Check that database is open.
   if (AppStatus == db_closed) return;
   if (AppStatus != db_open) {
       Message = "Can not perform this operation while database is loading.";
       statusBar()->showMessage(Message, 3000);
       return;
   }

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Exit if no rows are selected.
   if (!ELVSelection.size()) return;

   // Only one row can be selected.
   if (ELVSelection.size() > 1) {
      Message = "Multiple selections are not allowed for this operation.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Select a platform dependant shell program.
   #ifdef Q_OS_WIN32
   QString shell = "cmd.exe /c ";
   #endif
   #ifdef Q_OS_LINUX
   QString shell = "xterm -e ";
   #endif

   // Create a platform dependant name for eev.
   #ifdef Q_OS_WIN32
   Eev = "eev.exe";
   #endif
   #ifdef Q_OS_LINUX
   Eev = "eev";
   #endif

   // Check that the 'eev' program exists.
   if (!IsFileInPath(Eev)) {
      Message = "Register event: Can not find eev in the path.";
      emit WriteLog(logmodel::error, Message);
      ui->tabWidget->setCurrentIndex(1);
      return;
   }

   // Get index for the selected event.
   DbIndex = ELVSelection.at(0).index;

   // Run eev to register this event in database.
   // Get date/time from the name of the event file.
   DB.GetFileByIndex(DbIndex, &FileName);
   GetSfileNameInfo(FileName, &NameInfo);
   EventId = DB.GetIdByIndex(DbIndex);

   // Execute the 'eev' program.
   Command = shell + Eev + " " + NameInfo.DateStr + NameInfo.TimeStr + " ";
   Command.append(DatabaseName + " -seisanexp REG " + Operator.GetConfirmedId());
   if (system(Command.toAscii().data()) == -1) return;

   // Check return value in 'seisan.mes'.
   Status = ReadFileLine("seisan.mes", 1, &Line);
   Line=Line.trimmed(); QFile::remove("seisan.mes");
   if (Status == 0) {
      // The file 'seisan.mes' was found.
      if (Line.isEmpty()) {
         // File name has not changed, reload the file.
         rStatus = DB.ReloadEventAndResort(DbIndex);
         if (rStatus == 0 || rStatus == 4) {
            // Update the event list view.
            ELVModel->RemoveRow(DbIndex);
            NewIndex = DB.GetIndexById(EventId);
            ELVModel->InsertRow(NewIndex);
            ELVModel->SelectRow(NewIndex, false, true, false);
            QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
         } else {
            Message = "Failed to reload event file.";
            emit WriteLog(logmodel::error, Message);
            ui->tabWidget->setCurrentIndex(1);
            return;
         }
      }
      if (Line.size()) {
         // File name or database has changed. File will be
         // reloaded if it has not been moved to another database.
         GetSfileNameInfo(Line, &NewNameInfo);
         if (NewNameInfo.Database.toUpper() == DatabaseName.toUpper()) {
            // File has not been moved to another database, but the name
            // has changed. Unload the old file and load the new file.
            DB.RemoveEvent(DbIndex);
            ELVModel->RemoveRow(DbIndex);
            NewIndex = DB.LoadEventAndResort(Line);
            if (NewIndex >= 0) {
               ELVModel->InsertRow(NewIndex);
               ELVModel->SelectRow(NewIndex, false, true, false);
               QApplication::processEvents(QEventLoop::ExcludeUserInputEvents);
            }
            if (NewIndex != -1) {
               Message = "Failed to reload registered event file.";
               emit WriteLog(logmodel::error, Message);
               ui->tabWidget->setCurrentIndex(1);
               return;
            }
         } else {
            // Event file has moved to another database.
            // Remove the file from the Explorer database.
            DB.RemoveEvent(DbIndex);
            ELVModel->RemoveRow(DbIndex);
            return;
         }
      }
   } else {
      if (Status == 1) {
         // File was not found. Registration failed.
         Message = "Registration failed or was aborted by user.";
         statusBar()->showMessage(Message, 5000);
      }
      if (Status == 2) {
         // File could not be opened.
         Message = "Unable to open file 'seisan.mes' in work directory.";
         emit WriteLog(logmodel::error, Message);
         ui->tabWidget->setCurrentIndex(1);
      }
   }
}




// ********************************************************************
// This function handles the 'Set distance indicator' menu item.
// This function is signalled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Set_Distance_Indicator()
{
   char status;
   QChar NewDI;
   event_node_ *Node;
   sortCol CurrSortColumn;
   Qt::SortOrder CurrSortOrder;
   int EventId, DbIndex, NewIndex;

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Exit if no rows are selected.
   if (!ELVSelection.size()) return;

   // Only one row can be selected.
   if (ELVSelection.size() > 1) {
      Message = "Multiple selections are not allowed for this operation.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Bring up dialog box and ask for new
   // distance indicator ('L','R' or 'D').
   bool ok, done = false;
   QString Title = "Set distance indicator";
   QString Label = "Enter distance indicator:";
   QString Line = "Valid values are: L,R,D";
   while (!done) {
      Line = QInputDialog::getText(this, Title, Label, QLineEdit::Normal, Line, &ok, Qt::Dialog);
      if (!ok) return;
      Line = Line.toUpper(); NewDI = Line.at(0);
      if (NewDI == 'L' || NewDI == 'R' || NewDI == 'D') done = true;
   }

   // Get pointer to event node.
   DbIndex = ELVSelection.at(0).index;
   Node = DB.GetEventByIndex(DbIndex);

   // Save the event id for this node
   EventId = DB.GetIdByIndex(DbIndex);

   // Change distance indicator if needed.
   if ((NewDI != GetNodeDistanseIndicator(Node))) {
      // Save full path to existing event file.
      QFile OldFile(QString(Node->metadata.sfile));

      // Update event node with new distance indicator.
      // This will change the node's file name.
      SetNodeDistanseIndicator(Node, NewDI);

      // Write the updated node to a new event file.
      if (DB.UpdateSfile(DbIndex, &status)) {
         // Get current sort criteria used by database.
         DB.GetSort(&CurrSortColumn, &CurrSortOrder);
         // Update event list view.
         if (CurrSortColumn == ColDist) {
            // Resort the database.
            DB.Resort(false);
            // Update the view.
            ELVModel->RemoveRow(DbIndex);
            NewIndex = DB.GetIndexById(EventId);
            ELVModel->InsertRow(NewIndex);
            ELVModel->SelectRow(NewIndex, false, true, true);
         } else {
            // Refresh the row in the view
            ELVModel->RefreshRow(DbIndex);
         }

         // Delete the old event file.
         if (!OldFile.remove()) {
            if (OldFile.exists()) {
               Message = "Unable delete old event file '";
               Message.append(OldFile.fileName());
               Message.append("'. Please remove the file manually.");
               emit WriteLog(logmodel::error, Message);
               // Switch to Log View to show message.
               ui->tabWidget->setCurrentIndex(1);
            }
         }
      } else {
         // Unable to write new event file.
         // Write error message to the log.
         Message = "Failed to update distance indicator for '";
         Message.append(OldFile.fileName()); Message.append("'.");
         emit WriteLog(logmodel::error, Message);
         // Switch to Log View to show message.
         ui->tabWidget->setCurrentIndex(1);
      }
   }
}




// ***********************************************************************
// This function handles the 'Accosiate events' menu item. Function is
// used to find events that are close to each other in time. The time
// interval supplied by the user specifies how close the events has to be.
// Search will go downwards from the currently selected event in event
// list view. Found events will be selected in the list.
// ***********************************************************************
void MainWindow::ELV_Associate_Events()
{
   bool ok;
   char str7[8];
   int DbIndex;
   QString label;
   sortCol CurrSortColumn;
   Qt::SortOrder CurrSortOrder;

   // Check that database is open.
   if (AppStatus != db_open) return;

   // Check current sort order. Sort order must be 'DateTime'.
   DB.GetSort(&CurrSortColumn, &CurrSortOrder);
   if (CurrSortColumn != DateTime) {
      Message = "Event list must be sorted by date and time.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Exit if no rows are selected.
   if (!ELVSelection.size()) return;

   // Only one row can be selected.
   if (ELVSelection.size() > 1) {
      Message = "Multiple selections are not allowed for this operation.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Start the search from the selected row.
   DbIndex = ELVSelection.at(0).index;

   // Bring up dialog box and ask user for time interval (in seconds).
   label.append("Search starts from the selected row.\n");
   label.append("Please enter time interval in seconds:");
   QString Title = "Associate events";
   AccTimeInt = QInputDialog::getInt(this, Title, label, AccTimeInt, 0, 3600, 10, &ok, Qt::Dialog);
   if (!ok) return;

   // Search for events. This is done by the database.
   // Database will return a list of found indexes.
   // If IndexList.size() is zero then none was found.
   QList<int> IndexList;
   DB.AssociateEvents(AccTimeInt, DbIndex, &IndexList);
   if (!IndexList.size()) {
      Message = "No associations where found.";
      statusBar()->showMessage(Message, 3000);
      return;
   } else {
      // Select the events in the event list view.
      QApplication::setOverrideCursor(QCursor(Qt::WaitCursor));
      ELVModel->SelectRows(&IndexList, true);
      QApplication::restoreOverrideCursor();
      // Write number of found events.
      sprintf(str7, "%i", IndexList.size());
      Message.clear(); Message.append(QString(str7));
      Message.append(" events was selected.");
      statusBar()->showMessage(Message, 4000);
   }
}




// ********************************************************************
// This function handles the 'Edit comment lines' menu item.
// This function is signaled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Edit_Comment_Lines()
{
   char Status;
   event_node_* Node;
   QStringList CommentList;
   GetTextLinesDlg *Editor;

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Exit if no rows are selected.
   if (!ELVSelection.size()) return;

   // Only one row can be selected.
   if (ELVSelection.size() > 1) {
      Message = "Multiple selections are not allowed for this operation.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Get pointer to event node.
   Node = DB.GetEventByIndex(ELVSelection.at(0).index);

   // Get all comment lines.
   NodeGetComments(Node, &CommentList);

   // Bring up the editor.
   Editor = new GetTextLinesDlg(this);
   Editor->setAttribute(Qt::WA_DeleteOnClose, true);
   Editor->InitDialog(&CommentList, 78, "Edit comment lines");
   if (Editor->exec() == QDialog::Accepted) {
      // Save new comments to node.
      NodeSetComments(Node, CommentList);
      // Rewrite the event file.
      if (!DB.UpdateSfile(ELVSelection.at(0).index, &Status)) {
         // Unable to write new event file.
         // Write error message to the log.
         Message = "Failed to update comments for '";
         Message.append(Node->metadata.sfile); Message.append("'.");
         emit WriteLog(logmodel::error, Message);
         // Switch to Log View to show message.
         ui->tabWidget->setCurrentIndex(1);
      }
   }
}




// ********************************************************************
// This function handles the 'Delete event' menu item. This
// function is signalled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Delete_Event()
{
   bool DelError;
   QString FileName;
   int NumRows, DbIndex, dStatus;

   // Check that database is not an index file.
   if (CurrentDbType == index_file) {
       Message = "Delete is not supported when database is an index file.";
       statusBar()->showMessage(Message, 5000);
       return;
   }

   // Check that database is open.
   if (AppStatus == db_closed) return;
   if (AppStatus == db_loading) {
       Message = "Can not perform this operation while database is loading.";
       statusBar()->showMessage(Message, 3000);
       return;
   }

   // Get the list and number of selected rows.
   ELVModel->GetSelection(&ELVSelection);
   NumRows = ELVSelection.size();

   // Exit if no rows are selected.
   if (!NumRows) return;

   // Check number of selected rows. Max 10 events can be selected.
   if (NumRows > 10) {
      Message = "A maximum of 10 selections are allowed for this operation.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Ask user for confirmation.
   if (NumRows == 1) {
      // Get filename for the one selected row.
      DB.GetFileByIndex(ELVSelection.at(0).index, &FileName);
      Message = "Are you sure you want to delete this event from the Seisan database?\n" + FileName;
   } else {
      Message = "WARNING: You have selected " + QString::number(NumRows) + " events for deletion.\n";
      Message.append("Are you sure you want to delete these events from the Seisan database?");
   }
   if (QMessageBox::question(this, "Delete", Message, QMB_YES | QMB_NO, QMB_NO) != QMB_YES) return;

   // Delete all events from database and view.
   DelError = false;
   for (int i=0;i<NumRows;i++) {
      DbIndex = DB.GetIndexById(ELVSelection.at(i).id);
      if (!DeleteEvent(DbIndex, &dStatus, false)) DelError = true;
   }

   // Show log view if any errors occurred.
   if (DelError) ui->tabWidget->setCurrentIndex(1);

}




// ********************************************************************
// This function handles the 'EEV' menu item. This
// function is signalled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Launch_EEV()
{
   int DbIndex;
   sfile_nameinfo NameInfo;
   QDateTime ModTime1, ModTime2;
   QString Eev, EventFile, Command;

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Exit if no rows are selected.
   if (!ELVSelection.size()) return;

   // Check number of selected rows. Only 1 event can be selected.
   if (ELVSelection.size() > 1) {
      Message = "Multiple selections are not allowed for this operation.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Create a platform dependant name for the 'eev' program.
   #ifdef Q_OS_WIN32
   Eev = "eev.exe";
   #endif
   #ifdef Q_OS_LINUX
   Eev = "eev";
   #endif

   // Check that the 'eev' program exists.
   if (!IsFileInPath(Eev)) {
      Message = "EEV: Can not find eev in the path.";
      emit WriteLog(logmodel::error, Message);
      ui->tabWidget->setCurrentIndex(1);
      return;
   }

   // We need to open a terminal if we running linux.
   #ifdef Q_OS_LINUX
   Eev.prepend("xterm -e ");
   #endif

   // Get date/time from the name of the event file.
   DbIndex = ELVSelection.at(0).index;
   DB.GetFileByIndex(DbIndex, &EventFile);
   GetSfileNameInfo(EventFile, &NameInfo);

   // Save the event file modification time.
   GetFileModTime(EventFile, &ModTime1);

   // Append the arguments for 'eev'.
   Command = Eev;
   Command.append(" " + NameInfo.DateStr + NameInfo.TimeStr + " ");
   Command.append(DatabaseName + " -seisanexp SAME " + Operator.GetConfirmedId());

   // Execute the 'eev' program.
   if (system(Command.toAscii().data()) == -1) {
      Message = "Unable to start eev. Failed to execute '" + Command + "'.";
      emit WriteLog(logmodel::error, Message);
      ui->tabWidget->setCurrentIndex(1);
      return;
   }

   // Check if event file has been updated.
   GetFileModTime(EventFile, &ModTime2);
   if (ModTime1 != ModTime2) {
      // Eventfile has been updated. Save event identity.
      int EventId = DB.GetIdByIndex(DbIndex);
      // Reload the event file in database.
      if (DB.ReloadEventAndResort(DbIndex)) {
          // Event file could not be reloaded.
          Message = "Unable to locate event. Failed to reload the updated event file.";
          emit WriteLog(logmodel::error, Message);
          ui->tabWidget->setCurrentIndex(1);
      } else {
          int NewIndex = DB.GetIndexById(EventId);
          if (NewIndex == DbIndex) {
             // Refresh the selected event.
             ELVModel->RefreshRow(DbIndex);
          } else {
             // Reset the view and select updated event.
             ELVModel->RemoveRow(DbIndex);
             ELVModel->InsertRow(NewIndex);
             ELVModel->SelectRow(NewIndex, false, true, true);
          }
      }
   }
}




// ********************************************************************
// This function load an event file from disk into Explorer. This
// function is signalled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Load_Event()
{
   QString File;
   int EventIndex;
   sfile_nameinfo NameInfo;

   // Check that database is open.
   if (AppStatus == db_closed) return;
   if (AppStatus == db_loading) {
       Message = "Can not perform this operation while database is loading.";
       statusBar()->showMessage(Message, 3000);
       return;
   }

   // Bring up file selection dialog box. Return if user cancels
   // the operation. User must point to an S-file.
   Message = "Please select an event file to load:";
   File = QFileDialog::getOpenFileName(this, Message, DatabasePath);

   // Return if user did not select a file.
   if (File.isNull()) return;

   // Check that the selected file is an S-file.
   if (!IsSfileNameOk(GetNameFromFilePath(File))) {
      Message = "The selected file has an invalid name.";
      QMessageBox::critical(this, "Load event file", Message);
      return;
   }

   // Get the path/folder for the selected file.
   GetSfileNameInfo(File, &NameInfo);
   NameInfo.Path = QDir::toNativeSeparators(NameInfo.Path);

   // Check that user has selected a file from the current database.
   // This test is skipped if an index file has been opened.
   if (CurrentDbType != index_file) {
      if (!NameInfo.Path.startsWith(DatabasePath)) {
         Message = "Please select an event file from the currently opened database.";
         QMessageBox::critical(this, "Load event file", Message);
         return;
      }
   }

   // Check that event file is not already loaded.
   EventIndex = DB.GetIndexByFile(File);
   if (EventIndex != -1) {
      Message = "This event file is already loaded.";
      QMessageBox::critical(this, "Load event file", Message);
      ELVModel->SelectRow(EventIndex, true, true, true);
      return;
   }

   // Load event file into Explorer.
   EventIndex = DB.LoadEventAndResort(File);
   if (EventIndex == -1) {
      Message = "Unable to load the event file. Check file for errors.";
      emit WriteLog(logmodel::error, Message);
      ui->tabWidget->setCurrentIndex(1);
      return;
   }
   if (EventIndex == -2) {
      Message = "The loaded event has been filtered from Event List view.";
      statusBar()->showMessage(Message, 5000);
      return;
   }

   // Show and select the new event in the view.
   ELVModel->InsertRow(EventIndex);
   ELVModel->SelectRow(EventIndex, true, true, true);
}




// ********************************************************************
// This function unloads an event file from disk into Explorer. This
// function is signalled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Unload_Event()
{
   QString FileName;

   // Check that database is open.
   if (AppStatus == db_closed) return;
   if (AppStatus == db_loading) {
       Message = "Can not perform this operation while database is loading.";
       statusBar()->showMessage(Message, 3000);
       return;
   }

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Exit if no rows are selected.
   if (!ELVSelection.size()) return;

   // Check number of selected rows. Only 1 event can be selected.
   if (ELVSelection.size() > 1) {
      Message = "Multiple selections are not allowed for this operation.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Get filename for the one selected row.
   DB.GetFileByIndex(ELVSelection.at(0).index, &FileName);
   Message = "Are you sure you want to remove this event from Explorer?\n";
   Message.append("The event file will not be deleted from the Seisan database.");
   if (QMessageBox::question(this, "Remove", Message, QMB_YES | QMB_NO, QMB_NO) != QMB_YES) return;

   // Unload event and update the view.
   DB.RemoveEvent(ELVSelection.at(0).index);
   ELVModel->RemoveRow(ELVSelection.at(0).index);
}




// *********************************************************
// This function shows locations for one or more
// selected events with Google Maps in a window.
// *********************************************************
void MainWindow::ELV_Show_Events_With_GM()
{
   float Lat, Lon;
   coord Location;
   event_node_ *Node;
   QListCoord LocList;

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Get number of selected rows.
   int NumSelectedRows = ELVSelection.size();

   // Exit if no rows are selected.
   if (!NumSelectedRows) return;

   // Create a list of all event locations.
   for (int i=0;i<NumSelectedRows;i++) {
      Node = DB.GetEventByIndex(ELVSelection.at(i).index);
      Lat = Node->hypocenters.first->lat;
      Lon = Node->hypocenters.first->lon;
      if (Lat != -999.0 && Lon != -999.0) {
         Location.lat = Lat;
         Location.lon = Lon;
         LocList.append(Location);
      }
   }

   // Return here if 'Locations' is empty.
   if (!LocList.size()) {
      Message = "The selected events do not have a location.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Show the map in the map view window.
   MapView->ShowMap(&LocList);

   // Notify user if some events where not shown.
   if (NumSelectedRows != LocList.size()) {
      Message = "Some events are not shown because they do not have a location.";
      statusBar()->showMessage(Message, 3000);
      return;
   }
}




// ***********************************************************************
// This function shows one or more selected events with Google Earth.
// Google Earth is NOT being started, only the KML file is created.
// ***********************************************************************
void MainWindow::ELV_Show_Events_With_GE()
{
   float Lat, Lon;
   event_node_ *Node;
   int NumSelectedRows;
   QList<event_node_*> EventList;

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Get number of selected rows.
   NumSelectedRows = ELVSelection.size();

   // Exit if no rows are selected.
   if (!NumSelectedRows) return;

   // Copy pointers to all events that have a location into 'EventList'.
   for (int i=0;i<NumSelectedRows;i++) {
      Node = DB.GetEventByIndex(ELVSelection.at(i).index);
      Lat = Node->hypocenters.first->lat;
      Lon = Node->hypocenters.first->lon;
      if (Lat != -999.0 && Lon != -999.0) EventList.append(Node);
   }

   // Return here if 'EventList' is empty.
   if (!EventList.size()) {
      Message = "The selected events do not have a location.";
      statusBar()->showMessage(Message, 3000);
      return;
   }

   // Set the URL for the map icon.
   QString IconUrl = "http://maps.google.com/mapfiles/kml/pal2/icon26.png";

   // Set the KML color string for the map icon.
   QString IconColor = "ff00ff00";

   // These variables defines the size of the map icons.
   float mSize = 0.6; float xSize = 0.2; float ySize = 0.5;

   // Write the KML file.
   QFile *KmlFile = new QFile(SEC_KML_FILE);
   if (KmlFile->open(QIODevice::WriteOnly | QIODevice::Truncate | QIODevice::Text)) {
      WriteKmlHeader(KmlFile);
      for (int i=0;i<EventList.size();i++) {
         WriteKml(EventList[i], KmlFile, IconUrl, IconColor, true, mSize, xSize, ySize);
      }
      WriteKmlTail(KmlFile);
      KmlFile->close();

      // Notify user if some events where not shown.
      if (NumSelectedRows != EventList.size()) {
         Message = "Some events are not shown because they do not have a location.";
         statusBar()->showMessage(Message, 3000);
         return;
      }
   } else {
       Message = "Unable to open file '" + QString(SEC_KML_FILE) + "' for writing.";
       emit WriteLog(logmodel::error, Message);
       ui->tabWidget->setCurrentIndex(1);
   }
}




// ***********************************************************************
// This function shows one or more selected events with
// Seismicity Viewer. Seismicity Viewer is a java program.
// ***********************************************************************
void MainWindow::ELV_Show_Events_With_SV()
{
   int NumSelectedRows;

   // Get the list of selected rows.
   ELVModel->GetSelection(&ELVSelection);

   // Get number of selected rows.
   NumSelectedRows = ELVSelection.size();

   // Exit if no rows are selected.
   if (!NumSelectedRows) return;

   // Create platform dependant names for the java program.
   #ifdef Q_OS_WIN32
   QString java = "java.exe";
   #endif
   #ifdef Q_OS_LINUX
   QString java = "java";
   #endif

   // Check that the 'java' command exists in the path.
   if (!IsFileInPath(java)) {
      Message = "Plot with Seismicity Viewer: Can not find the 'java'' command in the path.";
      emit WriteLog(logmodel::error, Message);
      ui->tabWidget->setCurrentIndex(1);
      return;
   }

   // Check that the file 'seis2viewer.jar' exists in PRO directory.
   QString seis2viewer = QDir::toNativeSeparators(SEISAN_TOP + "/PRO/seis2viewer.jar");
   if (!QFile::exists(seis2viewer)) {
      Message = "Plot with Seismicity Viewer: Can not find the file '" + seis2viewer + "'.";
      emit WriteLog(logmodel::error, Message);
      ui->tabWidget->setCurrentIndex(1);
      return;
   }

   // Create a platform dependant system command.
   #ifdef Q_OS_WIN32
   QString Command = "java -jar " + SEISAN_TOP + "\\PRO\\seis2viewer.jar copy.out";
   #endif
   #ifdef Q_OS_LINUX
   QString Command = "xterm -e java -jar " + SEISAN_TOP + "/PRO/seis2viewer.jar copy.out";
   #endif

   // Write the file 'copy.out'.
   ELV_Copy_Events();

   // Execute the command.
   if (system(Command.toAscii().data()) == -1) {
      Message = "Unable to start Seismicity Viewer. Failed to execute '" + Command + "'.";
      emit WriteLog(logmodel::error, Message);
      ui->tabWidget->setCurrentIndex(1);
      return;
   }
}




// ********************************************************************
// This function is brings up the 'Goto' dialog box. This dialog
// box is used to move current selection to a particular row number.
// The current selection is lost.
// ********************************************************************
void MainWindow::ELV_GoToRowNr()
{
   // Return if the database is closed.
   if (AppStatus == db_closed) return;

   // Get number of events in database.
   // Return if the database is empty.
   int NumEvents = DB.GetNumberOfEvents();
   if (!NumEvents) return;

   // Open the dialog box.
   GoToRowDlg *Dialog = new GoToRowDlg();  // Create the dialog.
   int RowNr = Dialog->exec();             // Open the dialog box.
   delete Dialog;                          // Destroy the dialog.
   if (RowNr > 0) {
      // Check row number.
      int NumEvents = DB.GetNumberOfEvents();
      if (RowNr > NumEvents) RowNr = NumEvents;
      // Select the new row.
      ELVModel->SelectRow(RowNr-1, true, true, true);
   }
}




// ********************************************************************
// This function moves current selection to the next row number that
// has Action = NEW. The current selection is lost.
// ********************************************************************
void MainWindow::ELV_GoToNextNewEvent()
{
   event_node_ *Node;

   // Return if the database is closed.
   if (AppStatus == db_closed) return;

   // Get number of selected rows.
   ELVModel->GetSelection(&ELVSelection);
   int NumSelectedRows = ELVSelection.size();

   // Get number of events in database.
   int NumEvents = DB.GetNumberOfEvents();

   // If only one row is selected, start then search
   // from selected row. Othervice search from top.
   int StartRow = 0;
   if (NumSelectedRows == 1) {
       StartRow = ELVSelection.at(0).index + 1;
       if (StartRow == NumEvents) StartRow = 0;
   }

   // Search for next event with Action = 'NEW' .
   for (int Index=StartRow;Index<NumEvents;Index++) {
      Node = DB.GetEventByIndex(Index);
      if (!QString::compare(QString(Node->action), "NEW", Qt::CaseInsensitive)) {
         // Select the found row and return.
         ELVModel->SelectRow(Index, true, true, true);
         return;
      }
   }
}




// ********************************************************************
// This function is signalled when the user presses one of the
// column headers to change either the sort column or sort order.
// ********************************************************************
void MainWindow::ELV_SortIndicatorChanged(int logicalIndex, Qt::SortOrder NewSortOrder)
{
   sortCol NewSortColumn;
   sortCol CurrSortColumn;
   Qt::SortOrder CurrSortOrder;

   // Get new sort column.
   NewSortColumn = (sortCol)logicalIndex;

   // Get current sort criteria used by database.
   DB.GetSort(&CurrSortColumn, &CurrSortOrder);

   // Return if nothing has changed.
   if (NewSortColumn == CurrSortColumn && NewSortOrder == CurrSortOrder) return;

   // If user has selected the row number column, or database
   // is still loading, then do not switch to new sort order.
   if (NewSortColumn == RowNr || AppStatus != db_open) {
      // Restore the previous column selection in event list view header.
      ELVHeader->setSortIndicator((int)CurrSortColumn, CurrSortOrder);
      QApplication::beep();
      return;
   }

   // Set new sort column or sort order.
   if (NewSortColumn != CurrSortColumn) {
      // Sort database by new column.
      DB.SetSortCol(NewSortColumn);
      // Update the view and select first row.
      ELVModel->Reset(); ELVModel->SelectRow(0, true, true, false);
   } else {
      // Set new database sort order.
      DB.SetSortOrd(NewSortOrder);
      // Update the view.
      ELVModel->Reset(); ELVModel->SelectRow(0, true, true, false);
   }
}




// ********************************************************************
// This function refreshes the Event List View. This function
// is signalled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Refresh_View()
{
    ELVModel->Reset();
    ELVModel->SelectRow(0, false, true, false);
}




// ********************************************************************
// This function selects all events in the Event List View.
// This function is signalled from the Event List View popup menmu.
// ********************************************************************
void MainWindow::ELV_Select_All()
{
    // Check that database is open.
    if (AppStatus == db_closed) return;
    if (AppStatus == db_loading) {
        Message = "Can not perform this operation while database is loading.";
        statusBar()->showMessage(Message, 3000);
        return;
    }

    // Select all rows.
    emit ui->EventListView->selectAll();
}




// ********************************************************************
// Brings up the Log View popup meny. This function is signalled
// from LogView widget when user right-clicks inside the vidget view.
// ********************************************************************
void MainWindow::CCMR_LogView(const QPoint &pos)
{
   QString FileName;
   QAction *menuItem;

   // Get the list of selected lines. Only one line can be selected.
   QModelIndexList ModelIndexes = ui->LogView->selectionModel()->selectedRows();

   // Create and show the popup menu if
   // the log view holds any information.
   if (LogModel->GetNumberOfLines()) {
      // Create a popup menu.
      QMenu *popupMenu = new QMenu(this);

      // Add 'Edit this event file' item to menu if needed.
      if (!ModelIndexes.isEmpty()) {
         FileName = LogModel->GetMsgFilename(ModelIndexes.at(0).row());
         if (FileName != "none") {
            menuItem = new QAction(tr("Edit this event file"), this);
            popupMenu->addAction(menuItem);
            QObject::connect(menuItem, SIGNAL(triggered(bool)), this, SLOT(LV_Edit_Event()));
            popupMenu->addSeparator();
         }
      }

      // Add 'Clear log view' item to menu.
      menuItem = new QAction(tr("Clear the Log View"), this);
      popupMenu->addAction(menuItem);
      QObject::connect(menuItem, SIGNAL(triggered(bool)), LogModel, SLOT(ClearLog()));

      // Add seperator to menu.
      popupMenu->addSeparator();

      // Add 'Write to log file' item to menu.
      menuItem = new QAction(tr("Save Log View to log file"), this);
      popupMenu->addAction(menuItem);
      QObject::connect(menuItem, SIGNAL(triggered(bool)), LogModel, SLOT(WriteLog2File()));

      // Add 'View or edit log file' item to menu.
      menuItem = new QAction(tr("View or edit log file"), this);
      popupMenu->addAction(menuItem);
      QObject::connect(menuItem, SIGNAL(triggered(bool)), this, SLOT(LV_Edit_LogFile()));

      // Add 'Clear log file' item to menu.
      menuItem = new QAction(tr("Clear log file"), this);
      popupMenu->addAction(menuItem);
      QObject::connect(menuItem, SIGNAL(triggered(bool)), LogModel, SLOT(ClearLogFile()));

      // Show popup menu at mouse position.
      popupMenu->show();
      popupMenu->move(QCursor::pos());
   }
}




// ********************************************************************
// This function handles the 'Edit event file' menu item.
// This function is signaled from the Log View popup menmu.
// ********************************************************************
void MainWindow::LV_Edit_Event()
{
   QString FileName;
   FileEditorDlg *Editor;

   // Get the list of selected lines. Only one line can be selected.
   QModelIndexList ModelIndexes = ui->LogView->selectionModel()->selectedRows();

   // Get position of explorer program on screen.
   QPoint Position = this->pos();

   // Get filename for this log line.
   FileName = LogModel->GetMsgFilename(ModelIndexes.at(0).row());

   // Open editor with this file.
   Editor = new FileEditorDlg(this);
   Editor->setAttribute(Qt::WA_DeleteOnClose, true);
   if (Editor->InitDialog(FileName, AppName)) {
      // File was sucessfylly opened.
      Position.setX(Position.x()+20);
      Position.setY(Position.y()+20);
      Editor->move(Position);
      Editor->show();
      Editor->activateWindow();
   } else {
      // File could not be opened.
      Message = "Unable to open file '";
      Message.append(FileName); Message.append("'.");
      QMessageBox::critical(this, AppName, Message);
      delete Editor;
   }
}




// ********************************************************************
// This function handles the 'View or edit log file' menu item.
// This function is signaled from the Log View popup menmu.
// ********************************************************************
void MainWindow::LV_Edit_LogFile()
{
   QString FileName;
   FileEditorDlg *Editor;

   // Set name of log file.
   FileName = "explorer.log";

   // Open editor with this file.
   Editor = new FileEditorDlg(this);
   Editor->setAttribute(Qt::WA_DeleteOnClose, true);
   if (Editor->InitDialog(FileName, AppName)) {
      // File was sucessfylly opened.
      Editor->show();
      Editor->activateWindow();
   } else {
      // File could not be opened.
      Message = "Unable to open file '";
      Message.append(FileName); Message.append("'.");
      QMessageBox::critical(this, AppName, Message);
      delete Editor;
   }
}
