#include <QSettings>
#include <QInputDialog>
#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "types.h"
#include "fortran.h"
#include "seisandb.h"
#include "fileeditordlg.h"


// ***************************************************************
// This function deletes an event from the explorer database and
// from disk. It also updates the explorer database and the event
// log view. User will not asked for permission before delete.
// ***************************************************************
bool MainWindow::DeleteEvent(int DbIndex, int *Status, bool Quiet)
{
   int status;
   db_info DbInfo;
   QString FileName;

   // Get filename for the selected row.
   DB.GetFileByIndex(DbIndex, &FileName);

   // Delete the event file from disk.
   DbInfo.DbType = CurrentDbType;
   DbInfo.DbName = DatabaseName;
   DbInfo.DbPath = DatabasePath;
   if ((status = DeleteSfileFromDisk(FileName, DbInfo))) {
      if (!Quiet) {
         // Delete operation failed. Write error message to log.
         QString Message = "Unable to delete event. ";
         if (status == 1) Message.append("Event file was not found.");
         if (status == 2) Message.append("Failed to delete event file.");
         if (status == 3) Message.append("Failed to locate DELET database.");
         if (status == 4) Message.append("Failed to copy event to DELET database.");
         emit WriteLog(logmodel::error, Message);
      }
      *Status = status; return false;
   } else {
      // Delete operation succeeded.
      // Update the explorer database.
      DB.RemoveEvent(DbIndex);
      // Update the event log view.
      ELVModel->RemoveRow(DbIndex);
      ELVModel->RefreshRow(DbIndex);
      ELVModel->SelectRow(DbIndex, true, true, false);
   }
   *Status = 0; return true;
}




// ***************************************************************
// This function turns on or off the user selectable columns based
// on the values found in the EvtListColStat table. This function
// also resizes the column headers to fit the data in the view.
// ***************************************************************
void MainWindow::SetEventListViewHeaders()
{
   ui->EventListView->setColumnHidden(ColRowNumber, !Config.EvtListColStat[0]);    // Row Nr
   ui->EventListView->setColumnHidden(ColAction, !Config.EvtListColStat[1]);       // Action
   ui->EventListView->setColumnHidden(ColModel, !Config.EvtListColStat[2]);        // Model
   ui->EventListView->setColumnHidden(ColAgency, !Config.EvtListColStat[3]);       // Agency
   ui->EventListView->setColumnHidden(ColRMS, !Config.EvtListColStat[4]);          // RMS
   ui->EventListView->setColumnHidden(ColGap, !Config.EvtListColStat[5]);          // Gap
   ui->EventListView->setColumnHidden(ColErrLat, !Config.EvtListColStat[6]);       // Latitude error
   ui->EventListView->setColumnHidden(ColErrLon, !Config.EvtListColStat[7]);       // Longitude error
   ui->EventListView->setColumnHidden(ColErrDep, !Config.EvtListColStat[8]);       // Depth error
   ui->EventListView->setColumnHidden(ColDist, !Config.EvtListColStat[9]);         // Distance Indicator
   ui->EventListView->setColumnHidden(ColType, !Config.EvtListColStat[10]);        // Event Indicator
   ui->EventListView->setColumnHidden(ColNrSta, !Config.EvtListColStat[11]);       // Number of Stations
   ui->EventListView->setColumnHidden(ColM, !Config.EvtListColStat[12]);           // Magnitude M
   ui->EventListView->setColumnHidden(ColMW, !Config.EvtListColStat[13]);          // Magnitude MW
   ui->EventListView->setColumnHidden(ColML, !Config.EvtListColStat[14]);          // Magnitude ML
   ui->EventListView->setColumnHidden(ColMC, !Config.EvtListColStat[15]);          // Magnitude MC
   ui->EventListView->setColumnHidden(ColMb, !Config.EvtListColStat[16]);          // Magnitude Mb
   ui->EventListView->setColumnHidden(ColMB, !Config.EvtListColStat[17]);          // Magnitude MB
   ui->EventListView->setColumnHidden(ColMs, !Config.EvtListColStat[18]);          // Magnitude Ms
   ui->EventListView->setColumnHidden(ColMS, !Config.EvtListColStat[19]);          // Magnitude MS
   ui->EventListView->setColumnHidden(ColLocality, !Config.EvtListColStat[20]);    // Locality
   ui->EventListView->setColumnHidden(ColSFile, !Config.EvtListColStat[21]);       // Path to S-File
}




// ********************************************************************
// This function updates the operator label in the status bar.
// This function is signaled from class 'OperatorId'.
// ********************************************************************
void MainWindow::SetStatusBarOperator()
{
   // Create operator string.
   QString Id = Operator.GetCurrentId();
   Message = "OPERATOR: ";
   if (Id.size()) {
      Message.append(Id);
   } else {
      Message.append("Not specified");
   }

   // Update operator label.
   LblOperator.setText(Message);
}




// ********************************************************************
// This function updates the interval label in the status bar.
// ********************************************************************
void MainWindow::SetStatusBarInterval()
{
   // Create operator string.
   Message = "TIME INTERVAL: ";
   Message.append(Config.StartDate.toString("yyyy.MM.dd"));
   Message.append(" - ");
   Message.append(Config.EndDate.toString("yyyy.MM.dd  "));

   // Update operator label.
   LblInterval.setText(Message);
}




// ********************************************************************
// This function is used to set focus for the tab widget's children.
// when user switches tab. Function is signaled from the tab widget.
// ********************************************************************
void MainWindow::CurrentTabChanged(int index)
{
   // Set focus to a view in the new tab.
   if (index == 0) {
      ui->EventListView->setFocus(Qt::OtherFocusReason);
      ui->EventListView->setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);
      ui->EventListView->setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
   }
   if (index == 1) {
      ui->LogView->setFocus(Qt::OtherFocusReason);
      ui->LogView->setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);
      ui->LogView->setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
   }
}




// *********************************************
// Loads program settings from registry or file.
// *********************************************
void MainWindow::LoadProgramSettings()
{
   // Create a QSettings object and read program settings.
   QSettings settings;

   // Load settings for the event list.
   settings.beginGroup("EventList");
   Config.EvtListColStat[0] = settings.value("ShowRowNumbers", true).toBool();
   Config.EvtListColStat[1] = settings.value("ShowAction", true).toBool();
   Config.EvtListColStat[2] = settings.value("ShowModel", true).toBool();
   Config.EvtListColStat[3] = settings.value("ShowAgency", true).toBool();
   Config.EvtListColStat[4] = settings.value("ShowRMS", true).toBool();
   Config.EvtListColStat[5] = settings.value("ShowGap", true).toBool();
   Config.EvtListColStat[6] = settings.value("ShowErrLat", true).toBool();
   Config.EvtListColStat[7] = settings.value("ShowErrLon", true).toBool();
   Config.EvtListColStat[8] = settings.value("ShowErrDep", true).toBool();
   Config.EvtListColStat[9] = settings.value("ShowDist", true).toBool();
   Config.EvtListColStat[10] = settings.value("ShowType", true).toBool();
   Config.EvtListColStat[11] = settings.value("ShowNrSta", true).toBool();
   Config.EvtListColStat[12] = settings.value("ShowM", true).toBool();
   Config.EvtListColStat[13] = settings.value("ShowMW", true).toBool();
   Config.EvtListColStat[14] = settings.value("ShowML", true).toBool();
   Config.EvtListColStat[15] = settings.value("ShowMC", false).toBool();
   Config.EvtListColStat[16] = settings.value("ShowMb", false).toBool();
   Config.EvtListColStat[17] = settings.value("Show_MB", false).toBool();
   Config.EvtListColStat[18] = settings.value("ShowMs", false).toBool();
   Config.EvtListColStat[19] = settings.value("Show_MS", false).toBool();
   Config.EvtListColStat[20] = settings.value("ShowLocality", false).toBool();
   Config.EvtListColStat[21] = settings.value("ShowSfile", false).toBool();
   settings.endGroup();

   // Load general configuration settings.
   settings.beginGroup("Configuration");
   Config.TimeInterval = settings.value("TimeInterval", 31).toInt();
   Config.AutoLoad = settings.value("AutoLoad", false).toBool();
   Config.LastUsedDatabaseType = settings.value("LastDbType", no_database).toInt();
   Config.LastUsedDatabaseName = settings.value("LastDbName").toString();
   Config.LastUsedDatabasePath = settings.value("LastDbPath").toString();
   Config.UseAltWorkDir = settings.value("UseAlternativeWorkDir", false).toBool();
   Config.AltWorkDir = settings.value("AlternativeWorkDir").toString();
   Config.GoogleStaticMapApiKey = settings.value("GoogleStaticMapApiKey").toString();
   Operator.SetUnconfirmedId(settings.value("OperatorId").toString());
   settings.endGroup();

   // Restore the current state of mainwindow's toolbars and dockwidgets.
   restoreState(settings.value("AppState", QByteArray()).toByteArray());
   restoreGeometry(settings.value("AppGeometry", QByteArray()).toByteArray());
}




// *******************************************
// Saves program settings to registry or file.
// *******************************************
void MainWindow::SaveProgramSettings()
{
   // Create a QSettings object and save program settings.
   QSettings settings;

   // Save settings for the event list.
   settings.beginGroup("EventList");
   settings.setValue(QString("ShowRowNumbers"), Config.EvtListColStat[0]);
   settings.setValue(QString("ShowAction"), Config.EvtListColStat[1]);
   settings.setValue(QString("ShowModel"), Config.EvtListColStat[2]);
   settings.setValue(QString("ShowAgency"), Config.EvtListColStat[3]);
   settings.setValue(QString("ShowRMS"), Config.EvtListColStat[4]);
   settings.setValue(QString("ShowGap"), Config.EvtListColStat[5]);
   settings.setValue(QString("ShowErrLat"), Config.EvtListColStat[6]);
   settings.setValue(QString("ShowErrLon"), Config.EvtListColStat[7]);
   settings.setValue(QString("ShowErrDep"), Config.EvtListColStat[8]);
   settings.setValue(QString("ShowDist"), Config.EvtListColStat[9]);
   settings.setValue(QString("ShowType"), Config.EvtListColStat[10]);
   settings.setValue(QString("ShowNrSta"), Config.EvtListColStat[11]);
   settings.setValue(QString("ShowM"), Config.EvtListColStat[12]);
   settings.setValue(QString("ShowMW"), Config.EvtListColStat[13]);
   settings.setValue(QString("ShowML"), Config.EvtListColStat[14]);
   settings.setValue(QString("ShowMC"), Config.EvtListColStat[15]);
   settings.setValue(QString("ShowMb"), Config.EvtListColStat[16]);
   settings.setValue(QString("Show_MB"), Config.EvtListColStat[17]);
   settings.setValue(QString("ShowMs"), Config.EvtListColStat[18]);
   settings.setValue(QString("Show_MS"), Config.EvtListColStat[19]);
   settings.setValue(QString("ShowLocality"), Config.EvtListColStat[20]);
   settings.setValue(QString("ShowSfile"), Config.EvtListColStat[21]);
   settings.endGroup();

   // Save general configuration settings.
   settings.beginGroup("Configuration");
   settings.setValue(QString("TimeInterval"), Config.TimeInterval);
   settings.setValue(QString("AutoLoad"), Config.AutoLoad);
   settings.setValue(QString("LastDbType"), Config.LastUsedDatabaseType);
   settings.setValue(QString("LastDbName"), Config.LastUsedDatabaseName);
   settings.setValue(QString("LastDbPath"), Config.LastUsedDatabasePath);
   settings.setValue(QString("UseAlternativeWorkDir"), Config.UseAltWorkDir);
   settings.setValue(QString("AlternativeWorkDir"), Config.AltWorkDir);
   settings.setValue(QString("OperatorId"), Operator.GetCurrentId());
   settings.endGroup();

   // Save the current state of mainwindow's toolbars and dockwidgets.
   settings.setValue(QString("AppState"), saveState());
   settings.setValue(QString("AppGeometry"), saveGeometry());
}




// *******************************************
// Sets the title of the main window and
// the text on the event list view tab.
// *******************************************
void MainWindow::SetWindowTitle()
{
   QString Title;

   // No database is open.
   if (AppStatus != db_open) {
      Title = "Seisan Explorer";
      this->setWindowTitle(Title);
      ui->tabWidget->setTabText(0, "Event List");
      return;
   }

   // A database is open.
   sprintf(str11, "%i", DB.GetNumberOfEvents());
   Title = "Seisan Explorer - Database: ";
   if (CurrentDbType == default_database) {
      Title.append(DEF_BASE).append(" (default) ");
      Title.append(str11).append(" events.");
   }
   if (CurrentDbType == normal_database) {
      Title.append(DatabaseName).append(" (");
      Title.append(DatabasePath).append(") ");
      Title.append(str11).append(" events.");
   }
   if (CurrentDbType == local_database) {
      Title.append(DatabasePath).append(" (Local Database) ");
      Title.append(str11).append(" events.");
   }
   if (CurrentDbType == index_file) {
      Title.append(DatabaseName).append(" (Index File) ");
      Title.append(str11).append(" events.");
   }
   this->setWindowTitle(Title);

   // Set title of Event List view.
   Title = "Event List";
   if (DB.isFilterApplied()) {
      Title.append(" [Filter: ");
      if (DbFilter.Description.size()) {
         Title.append(DbFilter.Description.left(60));
         if (DbFilter.Description.size() > 60) Title.append("...");
         Title.append("]");
      } else {
         Title.append("No description]");
      }
   }
   ui->tabWidget->setTabText(0, Title.simplified());
}




// ********************************************************
// Function for editing Seisan parameter files. The files
// can be located in work dir or in Seisan DAT dir. If file
// exists in work dir, then this file will be opened, else
// the file in the DAT dir will be opened.
//
// FileType: 1=SEISAN.DEF, 2=MULPLT.DEF, 3=STATIONx.HYP
// ********************************************************
void MainWindow::EditParamFile(char FileType, char Model)
{
   char model[2];
   FileEditorDlg *Editor;
   QString ID, FilePath, WorkFile, DatFile;
   bool OK, FileInWorkDir = false, FileInDatDir = false;

   // Handle SEISAN.DEF or MULPLT.DEF file.
   FileInWorkDir = FileInDatDir = false;
   if (FileType == 1 || FileType == 2) {
      // Check if file exists in the work directory.
      if (FileType == 1) WorkFile = QDir::currentPath() + "/SEISAN.DEF";
      if (FileType == 2) WorkFile = QDir::currentPath() + "/MULPLT.DEF";
      WorkFile = QDir::cleanPath(WorkFile);
      FileInWorkDir = QFile::exists(WorkFile);
      // Check if file exists in the DAT directory.
      if (FileType == 1) DatFile = SEISAN_TOP + "/DAT/SEISAN.DEF";
      if (FileType == 2) DatFile = SEISAN_TOP + "/DAT/MULPLT.DEF";
      DatFile = QDir::cleanPath(DatFile);
      FileInDatDir = QFile::exists(DatFile);
   }

   // Handle the STATIONx.HYP file.
   if (FileType == 3) {
      // Set a default model for the dialog box.
       strcpy(model, "0");
       if (Model != 0 && Model != 32) model[0] = Model;
      // Open a dialog box and get model ID for file name.
      ID = QInputDialog::getText(this, "Edit a STATIONx.HYP file.",
                                 "Please enter the file's model identity (one char):",
                                 QLineEdit::Normal, QString(model), &OK);
      // Handle return values from dialog box.
      if (!OK || ID.size() != 1 || ID[0] == ' ') {
          Message = "No valid model identity was entered.";
          statusBar()->showMessage(Message, 4000);
          return;
      }
      // Search for file in work dir.
      WorkFile = QDir::currentPath() + "/STATION" + ID + ".HYP";
      WorkFile = QDir::cleanPath(WorkFile);
      FileInWorkDir = QFile::exists(WorkFile);
      // Search for file in DAT dir.
      DatFile = SEISAN_TOP + "/DAT/STATION" + ID + ".HYP";
      DatFile = QDir::cleanPath(DatFile);
      FileInDatDir = QFile::exists(DatFile);
   }

   // Exit here if file is not found.
   if (!FileInWorkDir && !FileInDatDir) {
      Message = "File was not found in work or DAT directory.";
      statusBar()->showMessage(Message, 4000);
      return;
   }

   // Choose correct file. File in work
   // dir will be selected if it exists.
   if (FileInDatDir) FilePath = DatFile;
   if (FileInWorkDir) FilePath = WorkFile;

   // Set position of editor on screen.
   QPoint Position = this->pos();
   Position.setX(Position.x()+200);
   Position.setY(Position.y()+200);

   // Bring up the editor. Editor will automatically
   // be destroyed when the user closes the window.
   Editor = new FileEditorDlg(this);
   Editor->move(Position);
   Editor->setAttribute(Qt::WA_DeleteOnClose, true);
   FilePath = QDir::toNativeSeparators(FilePath);
   if (Editor->InitDialog(FilePath, AppName)) {
      // File was sucessfylly opened. Open editor window.
      Editor->exec();
   } else {
      // File could not be opened.
      Message = "Unable to open file '" + FilePath + "'.";
      QMessageBox::critical(this, AppName, Message);
   }
}




// ***************************************************
// Automatically set column widths in event list view.
// ***************************************************
void MainWindow::ResizeColumns() {
   ui->EventListView->resizeColumnsToContents();
}




// ********************************
// Returns the program version.
// ********************************
QString MainWindow::GetAppVersion()
{
   return AppVersion;
}




// **********************************
// Returns the program build date.
// **********************************
QString MainWindow::GetAppBuildDate()
{
   return AppBuildDate;
}




// *****************************************************
// Ths function is called just before application exits.
// *****************************************************
void MainWindow::CleanUpOnExit()
{
   // Destroy the map view instance.
   delete MapView;

   // Close and delete the console file.
   closefile_unit_six_();
   QFile::remove(CONSOLE_FILE);

   // Remove the work directory lock file.
   if (LockFileCreated) QFile::remove(WORKDIR_LOCK_FILE);

}
